/*
	$Id: base.c,v 1.33 2012/04/03 03:51:13 alex Exp alex $
	base.c--- base info
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: base.c,v 1.33 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	<time.h>
#include	"libpq-fe.h"
#include	"edb/edb.h"
#include	"edb/xml.h"

#define	XML_AN_BASE_EID			"eid"
#define	XML_AN_BASE_EOID		"eoid"
#define	XML_AN_BASE_MAPTO		XML_AN_MAPTO
#define	XML_AN_BASE_PARENT		"parent"
#define	XML_AN_BASE_MTIME		"mtime"
#define	XML_AN_BASE_OPERATOR	"operator"
#define	XML_AN_BASE_AVAIL		"avail"
#define	XML_AN_BASE_DATE_FROM	XML_AN_DATE_FROM
#define	XML_AN_BASE_DATE_TO		XML_AN_DATE_TO
#define	XML_AN_BASE_CENSOR		"censor"
#define	XML_AN_BASE_OWNER		"owner"
#define	XML_AN_BASE_READ		XML_AN_READ
#define	XML_AN_BASE_WRITE		XML_AN_WRITE
#define	XML_AN_BASE_DELETE		XML_AN_DELETE

void
edb_base_init(EdbContext ec, EdbBaseInfo eb)
{

	if(eb==NULL) return;
	bzero(eb, sizeof(struct EDB_BASE_INFO));
	BaseDateFrom(eb) = EDB_DATE_MIN;
	BaseDateTo(eb) = EDB_DATE_MAX;
}

int
edb_base_parse_xml(EdbContext ec, EdbBaseInfo eb, xmlDocPtr doc, xmlNodePtr base)
{

	if(doc==NULL) {
		edb_trace(ec, "edb_base_parse_xml: invalid DOC.\n");
		return FALSE;
	}
	if(base==NULL) {
		edb_trace(ec, "edb_base_parse_xml: invalid BASE.\n");
		return FALSE;
	}
	BaseEID(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_EID, 0, NULL);
	BaseEOID(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_EOID, 0, NULL);
	BaseMapto(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_MAPTO, 0, NULL);
	BaseMtime(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_MTIME, 0, NULL);
	BaseOperator(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_OPERATOR, 0, NULL);
	BaseAvail(eb) = xml_get_element_attr_bool(ec, doc, base, (xmlChar *)XML_AN_BASE_AVAIL, TRUE, NULL);
	BaseDateFrom(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_DATE_FROM, EDB_DATE_MIN, NULL);
	BaseDateTo(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_DATE_TO, EDB_DATE_MAX, NULL);
	BaseCensor(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_CENSOR, 0, NULL);
	BaseOwner(eb) = xml_get_element_attr_integer(ec, doc, base, (xmlChar *)XML_AN_BASE_OWNER, 0, NULL);
	BaseRead(eb) = xml_get_element_attr_permission(ec, doc, base, (xmlChar *)XML_AN_BASE_READ, EDB_AUTH_PUBLIC, NULL);
	BaseWrite(eb) = xml_get_element_attr_permission(ec, doc, base, (xmlChar *)XML_AN_BASE_WRITE, EDB_AUTH_USER, NULL);
	BaseDelete(eb) = xml_get_element_attr_permission(ec, doc, base, (xmlChar *)XML_AN_BASE_DELETE, BaseWrite(eb), NULL);

	BaseDateFrom(eb) = DATE_ROUND(BaseDateFrom(eb));
	BaseDateTo(eb) = DATE_ROUND(BaseDateTo(eb));
	return TRUE;
}

utf8_t *
edb_base_make_xml(EdbContext ec, EdbBaseInfo eb, int last)
{
	char	*date_from = "";
	char	*date_to = "";
	int	d;
	utf8_t	*s, *xml;

	if(eb == NULL) return (utf8_t *)strdup("");

	if(EDB_DATE_MIN < (d=BaseDateFrom(eb))) {
		asprintf((char **)&s, XML_AN_BASE_DATE_FROM "=\"%d\" ", DATE_ROUND(d));
		date_from = STRDUP((char *)s);
		FREE(s);
	}
	if((d=BaseDateTo(eb)) < EDB_DATE_MAX) {
		asprintf((char **)&s, XML_AN_BASE_DATE_TO "=\"%d\" ", DATE_ROUND(d));
		date_to = STRDUP((char *)s);
		FREE(s);
	}

	asprintf((char **)&xml, "<" EDB_XMLNS ":base "
		XML_AN_BASE_EID			"=\"%d\" "
		XML_AN_BASE_EOID		"=\"%d\" "
		XML_AN_BASE_MAPTO		"=\"%d\" "
		XML_AN_BASE_MTIME		"=\"%ld\" "
		XML_AN_BASE_OPERATOR	"=\"%d\" "
		XML_AN_BASE_AVAIL		"=\"%s\" "
		"%s"
		"%s"
		XML_AN_BASE_CENSOR		"=\"%d\" "
		XML_AN_BASE_OWNER		"=\"%d\" "
		XML_AN_BASE_READ		"=\"%s\" "
		XML_AN_BASE_WRITE		"=\"%s\" "
		XML_AN_BASE_DELETE		"=\"%s\"/>",
		BaseEID(eb),
		BaseEOID(eb),
		BaseMapto(eb),
		BaseMtime(eb),
		BaseOperator(eb),
		BaseAvail(eb) ? "true" : "false",
		date_from,
		date_to,
		BaseCensor(eb),
		BaseOwner(eb),
		edb_auth_xmlstring(ec, BaseRead(eb), last),
		edb_auth_xmlstring(ec, BaseWrite(eb), last),
		edb_auth_xmlstring(ec, BaseDelete(eb), last)
	);
	return xml;
}

/*
	edb_base_make_query
	edb_base_get_values
*/
int
edb_base_make_query(EdbContext ec, utf8_t **query, utf8_t *as)
{
	char	*prefix = "";

	if(STRisVALID(as)) {
		char	*s;

		asprintf(&s, "%s.", as);
		prefix = STRDUP(s);
		FREE(s);
	}
	asprintf((char **)query,
		"%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s,%s%s",
		prefix, EDB_BASE_EID,
		prefix, EDB_BASE_EOID,
		prefix, EDB_BASE_MAPTO_EID,
		prefix, EDB_BASE_PARENT_EID,
		prefix, EDB_BASE_MTIME,
		prefix, EDB_BASE_OPERATOR,
		prefix, EDB_BASE_AVAILABLE,
		prefix, EDB_BASE_DATE_FROM,
		prefix, EDB_BASE_DATE_TO,
		prefix, EDB_BASE_CENSOR,
		prefix, EDB_BASE_OWNER,
		prefix, EDB_BASE_RDL,
		prefix, EDB_BASE_WRL,
		prefix, EDB_BASE_DEL);
	return 14;	/* number of columns */
}

int
edb_base_get_values(EdbContext ec, EdbBaseInfo eb, void *resdata)
{
	int	n;
	char	*s;
	PGresult *res = resdata;

	n = 0;
	BaseEID(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseEOID(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseMapto(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseParent(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseMtime(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseOperator(eb)	= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseAvail(eb)		= (s = PQgetvalue(res, 0, n++)) ? !strcmp(s, "t") : 1 ;
	BaseDateFrom(eb)	= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : EDB_DATE_MIN ;
	BaseDateTo(eb)	= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : EDB_DATE_MAX ;
	BaseCensor(eb)	= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseOwner(eb)	= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseRead(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseWrite(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;
	BaseDelete(eb)		= (s = PQgetvalue(res, 0, n++)) ? atoi(s) : 0 ;

	BaseDateFrom(eb) = DATE_ROUND(BaseDateFrom(eb));
	BaseDateTo(eb) = DATE_ROUND(BaseDateTo(eb));
	return n;
}

Link
edb_base_get_lvalues(EdbContext ec, EdbBaseInfo eb, Link lres)
{
	char	*s;

	BaseEID(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseEOID(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseMapto(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseParent(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseMtime(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseOperator(eb)	= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseAvail(eb)		= (s = (char *)lres->d) ? !strcmp(s, "t") : 1 ; lres = lres->n;
	BaseDateFrom(eb)	= (s = (char *)lres->d) ? atoi(s) : EDB_DATE_MIN ; lres = lres->n;
	BaseDateTo(eb)	= (s = (char *)lres->d) ? atoi(s) : EDB_DATE_MAX ; lres = lres->n;
	BaseCensor(eb)	= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseOwner(eb)	= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseRead(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseWrite(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;
	BaseDelete(eb)		= (s = (char *)lres->d) ? atoi(s) : 0 ; lres = lres->n;

	BaseDateFrom(eb) = DATE_ROUND(BaseDateFrom(eb));
	BaseDateTo(eb) = DATE_ROUND(BaseDateTo(eb));
	return lres;
}

/*
	edb_base_make_insert
*/
void
edb_base_make_insert(EdbContext ec, EdbBaseInfo eb, utf8_t **column, utf8_t **value)
{
	Link	l_column = NULL, l_value = NULL;

	LinkAppendTop(link_atom_new(EDB_BASE_EID, NULL), l_column);
	LinkAppendTop(link_atom_new_eid2s(BaseEID(eb)), l_value);

	LinkAppendTop(link_atom_new(EDB_BASE_EOID, NULL), l_column);
	LinkAppendTop(link_atom_new_eoid2s(BaseEOID(eb)), l_value);

	if(BaseMapto(eb)!=0) {
		LinkAppendTop(link_atom_new(EDB_BASE_MAPTO_EID, NULL), l_column);
		LinkAppendTop(link_atom_new_eid2s(BaseMapto(eb)), l_value);
	}

	if(BaseParent(eb)!=0) {
		LinkAppendTop(link_atom_new(EDB_BASE_PARENT_EID, NULL), l_column);
		LinkAppendTop(link_atom_new_eid2s(BaseParent(eb)), l_value);
	}

	LinkAppendTop(link_atom_new(EDB_BASE_MTIME, NULL), l_column);
	LinkAppendTop(link_atom_new_i2s(BaseMtime(eb)), l_value);

	LinkAppendTop(link_atom_new(EDB_BASE_OPERATOR, NULL), l_column);
	LinkAppendTop(link_atom_new_eid2s(BaseOperator(eb) ? BaseOperator(eb) : ec->operator), l_value);

	if(!BaseAvail(eb)) {
		LinkAppendTop(link_atom_new(EDB_BASE_AVAILABLE, NULL), l_column);
		LinkAppendTop(link_atom_new_spf("'%c'", BaseAvail(eb) ? 't' : 'f'), l_value);
	}

	if(EDB_DATE_MIN < BaseDateFrom(eb)) {
		LinkAppendTop(link_atom_new(EDB_BASE_DATE_FROM, NULL), l_column);
		LinkAppendTop(link_atom_new_i2s(DATE_ROUND(BaseDateFrom(eb))), l_value);
	}

	if(BaseDateTo(eb) < EDB_DATE_MAX) {
		LinkAppendTop(link_atom_new(EDB_BASE_DATE_TO, NULL), l_column);
		LinkAppendTop(link_atom_new_i2s(DATE_ROUND(BaseDateTo(eb))), l_value);
	}

	if(BaseCensor(eb)!=0) {
		LinkAppendTop(link_atom_new(EDB_BASE_CENSOR, NULL), l_column);
		LinkAppendTop(link_atom_new_i2s(BaseCensor(eb)), l_value);
	}

	LinkAppendTop(link_atom_new(EDB_BASE_OWNER, NULL), l_column);
	LinkAppendTop(link_atom_new_eid2s(BaseOwner(eb) ? BaseOwner(eb) : ec->operator), l_value);

	if(BaseRead(eb)!=0) {
		LinkAppendTop(link_atom_new(EDB_BASE_RDL, NULL), l_column);
		LinkAppendTop(link_atom_new_i2s(BaseRead(eb)), l_value);
	}

	if(BaseWrite(eb)!=0) {
		LinkAppendTop(link_atom_new(EDB_BASE_WRL, NULL), l_column);
		LinkAppendTop(link_atom_new_i2s(BaseWrite(eb)), l_value);
	}

	if(BaseDelete(eb)!=0) {
		LinkAppendTop(link_atom_new(EDB_BASE_DEL, NULL), l_column);
		LinkAppendTop(link_atom_new_i2s(BaseDelete(eb)), l_value);
	}

	LinkReverse(l_column);
	LinkReverse(l_value);

	*column = (utf8_t *)str_concatenate(l_column, ',');
	*value = (utf8_t *)str_concatenate(l_value, ',');
	EdbLinkFree(l_column);
	EdbLinkFree(l_value);
}

void
edb_base_print_information(EdbContext ec, EdbBaseInfo eb, int mode, Doc doc)
{
	Caption	cap;
	utf8_t	*s;

	if(eb == NULL || doc==NULL) return;
	if(DocIsXML(doc)) {
		s = edb_base_make_xml(ec, eb, FALSE);
		if(s) doc_puts(doc, (char *)s);
		FREE(s);
		return;
	}

	/* EID */
	if(DocIsHTML(doc)) doc_printf(doc, "<b>EID</b>:%d, \n", BaseEID(eb));
	if(DocIsPLAIN(doc)) doc_printf(doc, "EID: %d\n", BaseEID(eb));
	if(DocIsTeX(doc)) doc_printf(doc, "\\BaseEID{%d}\n", BaseEID(eb));

	/* EOID */
	if(ec->staff || ec->Odin) {
		if(DocIsHTML(doc)) doc_printf(doc, "<b>EOID</b>:%d, \n", BaseEOID(eb));
		if(DocIsPLAIN(doc)) doc_printf(doc, "EOID: %d\n", BaseEOID(eb));
		if(DocIsTeX(doc)) doc_printf(doc, "\\BaseEOID{%d}\n", BaseEOID(eb));
	}

	/* map */
	if(DocIsHTML(doc)) doc_puts(doc, "<b>Map</b>:");
	if(DocIsPLAIN(doc)) doc_puts(doc, "Map: ");
	if(DocIsTeX(doc)) doc_puts(doc, "\\BaseMap{");
	if(BaseMapto(eb) && (cap = edb_make_caption(ec, BaseMapto(eb), 0))) {
		if((mode&(EDB_PRINT_MODE_LINK|EDB_PRINT_MODE_WWWLINK)) && EIDisVALID(BaseMapto(eb))) {
			doc_printf(doc, "<a href=\"%s?EID=%d\"", ec->browser_url, BaseMapto(eb));
			if(mode&EDB_PRINT_MODE_BLANKTARGET) doc_puts(doc, " target=\"_blank\"");
			doc_puts(doc, ">");
			doc_text_puts3(doc, "[", (char *)cap->c_main, "]</a>");
		} else {
			doc_text_puts3(doc, "[", (char *)cap->c_main, "]");
		}
		edb_caption_free(ec, cap);
	} else {
		doc_printf(doc, "%d", BaseMapto(eb));
	}
	if(DocIsHTML(doc)) doc_puts(doc, ", \n");
	if(DocIsPLAIN(doc)) doc_puts(doc, "\n");
	if(DocIsTeX(doc)) doc_puts(doc, "}\n");

	/* parent */
	if(ec->staff || ec->Odin) {
		if(DocIsHTML(doc)) 	doc_printf(doc, "<b>Parent</b>:%d, \n", BaseParent(eb));
		if(DocIsPLAIN(doc)) 	doc_printf(doc, "parent: %d\n", BaseParent(eb));
		if(DocIsTeX(doc)) doc_printf(doc, "\\BaseParent{%d}\n", BaseParent(eb));
	}

	/* mtime */
	if(DocIsHTML(doc)) {
		doc_puts(doc, "<b>LastModified</b>:");
		doc_put_datetime(doc, BaseMtime(eb));
		doc_puts(doc, ", \n");
	} else if(DocIsTeX(doc)) {
		doc_puts(doc, "\\BaseLast{");
		doc_put_datetime(doc, BaseMtime(eb));
		doc_puts(doc, "\n");
	} else if(DocIsPLAIN(doc)) {
		doc_puts(doc, "last: ");
		doc_put_datetime(doc, BaseMtime(eb));
		doc_puts(doc, "\n");
	}

	/* operator */
	if(DocIsHTML(doc)) doc_puts(doc, "<b>Operator</b>:");
	if(DocIsPLAIN(doc)) doc_puts(doc, "op: ");
	if(DocIsTeX(doc)) doc_puts(doc, "\\BaseOperator{");
	if(BaseOperator(eb) && (cap = edb_make_caption(ec, BaseOperator(eb), 0))) {
		if((mode&(EDB_PRINT_MODE_LINK|EDB_PRINT_MODE_WWWLINK)) && EIDisVALID(BaseOperator(eb))) {
			doc_printf(doc, "<a href=\"%s?EID=%d\"", ec->browser_url, BaseOperator(eb));
			if(mode&EDB_PRINT_MODE_BLANKTARGET) doc_puts(doc, " target=\"_blank\"");
			doc_puts(doc, ">");
			doc_text_puts3(doc, "[", (char *)cap->c_main, "]</a>");
		} else {
			doc_text_puts3(doc, "[", (char *)cap->c_main, "]");
		}
		edb_caption_free(ec, cap);
	} else {
		doc_printf(doc, "%d", BaseOperator(eb));
	}
	if(DocIsHTML(doc)) doc_puts(doc, ", \n");
	if(DocIsPLAIN(doc)) doc_puts(doc, "\n");
	if(DocIsTeX(doc)) doc_puts(doc, "}\n");

	/* avail */
	if(DocIsHTML(doc)) {
		doc_printf(doc, "<b>Avail</b>:%s, \n", BaseAvail(eb) ? "TRUE" : "FALSE");
	} else if(DocIsPLAIN(doc)) {
		doc_printf(doc, "avail: %s\n", BaseAvail(eb) ? "TRUE" : "FALSE");
	} else if(DocIsTeX(doc)) {
		doc_printf(doc, "\\BaseAvail{%s}\n", BaseAvail(eb) ? "TRUE" : "FALSE");
	}

	/* date.from */
	if(EDB_DATE_MIN < BaseDateFrom(eb)) {
		if(DocIsHTML(doc)) doc_printf(doc, "<span style=\"color:#800000;\"><b>DateFrom</b>:%d,</span> \n", BaseDateFrom(eb));
		if(DocIsPLAIN(doc)) doc_printf(doc, "datefrom: %d\n", BaseDateFrom(eb));
		if(DocIsTeX(doc)) doc_printf(doc, "\\DateFrom{%d}\n", BaseDateFrom(eb));
	}

	/* date.to */
	if(BaseDateTo(eb) < EDB_DATE_MAX) {
		if(DocIsHTML(doc)) doc_printf(doc, "<span style=\"color:#800000;\"><b>DateTo</b>:%d,</span> \n", BaseDateTo(eb));
		if(DocIsPLAIN(doc)) doc_printf(doc, "dateto: %d\n", BaseDateTo(eb));
		if(DocIsTeX(doc)) doc_printf(doc, "\\DateTo{%d}\n", BaseDateTo(eb));
	}

	/* censor */
	if(DocIsHTML(doc)) doc_printf(doc, "<b>Censor</b>:%d, \n", BaseCensor(eb));
	if(DocIsPLAIN(doc)) doc_printf(doc, "censor: %s\n", BaseCensor(eb));
	if(DocIsTeX(doc)) doc_printf(doc, "\\BaseCensor{%s}\n", BaseCensor(eb));

	/* owner */
	if(DocIsHTML(doc)) doc_puts(doc, "<b>Owner</b>:");
	if(DocIsPLAIN(doc)) doc_puts(doc, "own: ");
	if(DocIsTeX(doc)) doc_puts(doc, "\\BaseOwn{");
	if(BaseOwner(eb) && (cap = edb_make_caption(ec, BaseOwner(eb), 0))) {
		if((mode&(EDB_PRINT_MODE_LINK|EDB_PRINT_MODE_WWWLINK)) && EIDisVALID(BaseOwner(eb))) {
			doc_printf(doc, "<a href=\"%s?EID=%d\"", ec->browser_url, BaseOwner(eb));
			if(mode&EDB_PRINT_MODE_BLANKTARGET) doc_puts(doc, " target=\"_blank\"");
			doc_puts(doc, ">");
			doc_text_puts3(doc, "[", (char *)cap->c_main, "]</a>");
		} else {
			doc_text_puts3(doc, "[", (char *)cap->c_main, "]");
		}
		edb_caption_free(ec, cap);
	} else {
		doc_printf(doc, "%d", BaseOwner(eb));
	}
	if(DocIsHTML(doc)) doc_puts(doc, ", \n");
	if(DocIsPLAIN(doc)) doc_puts(doc, "\n");
	if(DocIsTeX(doc)) doc_puts(doc, "}\n");

	/* read level */
	if(DocIsHTML(doc)) {
		doc_puts(doc, "<b>Read</b>:");
		if(BaseRead(eb)!=EDB_AUTH_INHERIT) doc_puts(doc, "<span style=\"color:#ff4040\">");
		doc_printf(doc, "%s", edb_auth_string(ec, BaseRead(eb), FALSE));
		if(BaseRead(eb)!=EDB_AUTH_INHERIT) doc_puts(doc, "</span>");
		doc_puts(doc, ", \n");
	}
	if(DocIsPLAIN(doc)) doc_printf(doc, "rl: %s\n", edb_auth_string(ec, BaseRead(eb), FALSE));
	if(DocIsTeX(doc)) doc_printf(doc, "\\BaseRL{%s}\n", edb_auth_string(ec, BaseRead(eb), FALSE));

	/* write level */
	if(DocIsHTML(doc)) {
		doc_puts(doc, "<b>Write</b>:");
		if(BaseWrite(eb)!=EDB_AUTH_INHERIT) doc_puts(doc, "<span style=\"color:#ff4040\">");
		doc_printf(doc, "%s", edb_auth_string(ec, BaseWrite(eb), FALSE));
		if(BaseWrite(eb)!=EDB_AUTH_INHERIT) doc_puts(doc, "</span>");
		doc_puts(doc, ", \n");
	}
	if(DocIsPLAIN(doc)) doc_printf(doc, "wl: %s\n", edb_auth_string(ec, BaseWrite(eb), FALSE));
	if(DocIsTeX(doc)) doc_printf(doc, "\\BaseWL{%s}\n", edb_auth_string(ec, BaseWrite(eb), FALSE));

	/* delete level */
	if(DocIsHTML(doc)) {
		doc_puts(doc, "<b>Delete</b>:");
		if(BaseDelete(eb)!=EDB_AUTH_INHERIT) doc_puts(doc, "<span style=\"color:#ff4040\">");
		doc_printf(doc, "%s", edb_auth_string(ec, BaseDelete(eb), FALSE));
		if(BaseDelete(eb)!=EDB_AUTH_INHERIT) doc_puts(doc, "</span>");
		doc_puts(doc, ".\n");
	}
	if(DocIsPLAIN(doc)) doc_printf(doc, "de: %s\n", edb_auth_string(ec, BaseDelete(eb), FALSE));
	if(DocIsTeX(doc)) doc_printf(doc, "\\BaseDL{%s}\n", edb_auth_string(ec, BaseDelete(eb), FALSE));
}

void
edb_base_print_in_xml(EdbContext ec, EdbBaseInfo eb, int mode, Doc doc)
{
	int	d;

	if(eb == NULL || doc==NULL) return;
	doc_text_printf(doc, "<%s:%s", EDB_XMLNS, XML_EN_BASE);
	doc_text_printf(doc, " " XML_AN_BASE_EID "=\"%d\"", BaseEID(eb));
	doc_text_printf(doc, " " XML_AN_BASE_EOID "=\"%d\"", BaseEOID(eb));
	doc_text_printf(doc, " " XML_AN_BASE_MAPTO "=\"%d\"", BaseMapto(eb));
	doc_text_printf(doc, " " XML_AN_BASE_MTIME "=\"%d\"", BaseMtime(eb));
	doc_text_printf(doc, " " XML_AN_BASE_OPERATOR "=\"%d\"", BaseOperator(eb));
	doc_text_printf(doc, " " XML_AN_BASE_AVAIL "=\"%s\"", BaseAvail(eb) ? "true": "false");
	if(EDB_DATE_MIN < (d=BaseDateFrom(eb))) {
		doc_text_printf(doc, " " XML_AN_BASE_DATE_FROM "=\"%d\"", DATE_ROUND(d));
	}
	if((d=BaseDateTo(eb)) < EDB_DATE_MAX) {
		doc_text_printf(doc, " " XML_AN_BASE_DATE_TO "=\"%d\"", DATE_ROUND(d));
	}
	doc_text_printf(doc, " " XML_AN_BASE_CENSOR "=\"%s\"", BaseCensor(eb));
	doc_text_printf(doc, " " XML_AN_BASE_OWNER "=\"%d\"", BaseOwner(eb));
	doc_text_printf(doc, " " XML_AN_BASE_READ "=\"%s\"", edb_auth_xmlstring(ec, BaseRead(eb), mode));
	doc_text_printf(doc, " " XML_AN_BASE_WRITE "=\"%s\"", edb_auth_xmlstring(ec, BaseWrite(eb), mode));
	doc_text_printf(doc, " " XML_AN_BASE_DELETE "=\"%s\"", edb_auth_xmlstring(ec, BaseDelete(eb), mode));
	doc_text_puts(doc, ">");
}

int
edb_base_compare(EdbContext ec, EdbBaseInfo eb1, EdbBaseInfo eb2)
{
	int	diff;

	diff = BaseEID(eb1)-BaseEID(eb2);
	if(diff!=0) return diff;
	diff = BaseMapto(eb1)-BaseMapto(eb2);
	if(diff!=0) return diff;
	diff = BaseAvail(eb1)-BaseAvail(eb2);
	if(diff!=0) return diff;
	diff = BaseDateFrom(eb1)-BaseDateFrom(eb2);
	if(diff!=0) return diff;
	diff = BaseDateTo(eb1)-BaseDateTo(eb2);
	if(diff!=0) return diff;
	diff = BaseCensor(eb1)-BaseCensor(eb2);
	if(diff!=0) return diff;
	diff = BaseOwner(eb1)-BaseOwner(eb2);
	if(diff!=0) return diff;
	diff = BaseRead(eb1)-BaseRead(eb2);
	if(diff!=0) return diff;
	diff = BaseWrite(eb1)-BaseWrite(eb2);
	if(diff!=0) return diff;
	diff = BaseDelete(eb1)-BaseDelete(eb2);
	if(diff!=0) return diff;
	return 0;
}

