/*
	$Id: caption.c,v 1.13 2008/08/09 09:21:37 alex Exp $
	caption.c--- caption
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: caption.c,v 1.13 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"

utf8_t *
edb_caption_string(EdbContext ec, int value, int longname)
{

	if(ec->japanese) {
		switch(value) {
		case EDB_CAPTION_SHORT:		return (utf8_t *)(longname ? "利用する（短）" : "短" );
		case EDB_CAPTION_LONG:		return (utf8_t *)(longname ? "利用する（長）" : "長" );
		case EDB_CAPTION_NONE:
		default:						return (utf8_t *)(longname ? "利用しない" : "×" );
		}
	} else {
		switch(value) {
		case EDB_CAPTION_SHORT:		return (utf8_t *)(longname ? "used(short)" : "S" );
		case EDB_CAPTION_LONG:		return (utf8_t *)(longname ? "used(long)" : "L" );
		case EDB_CAPTION_NONE:
		default:						return (utf8_t *)(longname ? "not used" : "." );
		}
	}
}

/*
	caption--- new/free
*/
Caption
edb_caption_new(EdbContext ec)
{
	Caption	cap;

	cap = (Caption)malloc(sizeof(struct _CAPTION));
	if(cap==NULL) return NULL;
	bzero(cap, sizeof(struct _CAPTION));

	cap->c_main = NULL;
	cap->c_main_english = NULL;
	cap->c_main_japanese = NULL;
	cap->c_main_pronounce = NULL;
	cap->c_sub = NULL;
	return cap;
}

void
edb_caption_free(EdbContext ec, Caption cap)
{

	if(cap==NULL) return;
	FREE(cap->c_main);
	FREE(cap->c_main_english);
	FREE(cap->c_main_japanese);
	FREE(cap->c_main_pronounce);
	FREE(cap->c_sub);
	FREE(cap);
}

/*
	edb_make_caption	-- get caption
*/
Caption
edb_make_caption(EdbContext ec, eid_t eid, int mode)
{
	EdbTableInfo	ti;
	EdbTupleInfo	tuple;
	Caption	cap;

	if(eid==ec->masterUID) {
		cap = edb_caption_new(ec);
		cap->c_main = (utf8_t *)strdup("[ADMIN]");
		cap->c_main_english = (utf8_t *)strdup("[ADMIN]");
		cap->c_main_japanese = (utf8_t *)strdup("[管理者]");
		cap->c_main_pronounce = (utf8_t *)strdup("かんりしゃ");
		cap->c_sub = (utf8_t *)strdup("[ADMIN]");
		return cap;
	}
	if(edb_eid_isin_table(ec, eid)==0) {
		ti = edb_table_get(ec, eid);
		if(ti == NULL) return NULL;
		return edb_table_make_caption(ec, ti, mode);
	} else {
		tuple = edb_tuple_get(ec, eid, 0);
		if(tuple == NULL) return NULL;
		return edb_tuple_make_caption(ec, tuple, mode);
	}
}

Caption
edb_make_mapped_caption(EdbContext ec, EdbColumnInfo ci, eid_t eid, int mode)
{
	EdbTableInfo	ti;
	EdbTupleInfo	tuple;
	Caption	cap;
	eid_t		teid;

	if(eid==ec->masterUID) {
		cap = edb_caption_new(ec);
		cap->c_main = (utf8_t *)strdup("[ADMIN]");
		cap->c_main_english = (utf8_t *)strdup("[ADMIN]");
		cap->c_main_japanese = (utf8_t *)strdup("[管理者]");
		cap->c_main_pronounce = (utf8_t *)strdup("かんりしゃ");
		cap->c_sub = (utf8_t *)strdup("[ADMIN]");
		return cap;
	}

	teid = edb_eid_isin_table(ec, eid);
	if(teid==0) {
		ti = edb_table_get(ec, eid);
		if(ti == NULL) return NULL;
		return edb_table_make_caption(ec, ti, mode);
	} else {
		Link	lm;

		tuple = edb_tuple_get(ec, eid, 0);
		if(tuple == NULL) return NULL;

		ForLink(lm, ci->l_maplookup) {
			EdbMapLookup	map = lm->d;

			if(teid!=edb_maplookup_table_eid(ec, map)) continue;
			if(map->l_caption_ci==NULL) {
				cap = edb_tuple_make_caption(ec, tuple, mode);
			} else {
				cap = edb_tuple_make_specified_caption(ec, tuple, mode, map->l_caption_ci);
			}
			return cap;
		}
		cap = edb_tuple_make_caption(ec, tuple, mode);
		return cap;
	}
}

/*
	caption omit list
*/
static
void
omit_list_free(BiTree bt)
{

	bitree_free(bt, NULL);
}

static
int
omit_list_dup_obj(BiTree_obj_t obj, BiTree_hint_t hint, void *data)
{
	register BiTree	*bt = data;
	register eid_t	eid = hint;

	if(!EIDisVALID(eid)) return TRUE;
	BiTreeAdd(*bt, "", (BiTree_hint_t)eid, NULL, NULL, NULL);
	return TRUE;
}

void
edb_caption_omit_save(EdbContext ec)
{
	BiTree	bt;

	LinkPush(ec->l_caption_omit, ec->bt_caption_omit);
	bt = NULL;
	bitree_callback_hint(ec->bt_caption_omit, omit_list_dup_obj, &bt);
	ec->bt_caption_omit = bt;
}

void
edb_caption_omit_restore(EdbContext ec)
{

	omit_list_free(ec->bt_caption_omit);
	ec->bt_caption_omit = NULL;
	ec->bt_caption_omit = LinkPop(ec->l_caption_omit);
}

int
edb_caption_is_omitted(EdbContext ec, eid_t eid)
{

	return (bitree_search(ec->bt_caption_omit, "", (BiTree_hint_t)eid, NULL)!=NULL);
}

void
edb_caption_omit_clear(EdbContext ec)
{

	omit_list_free(ec->bt_caption_omit);
	ec->bt_caption_omit = NULL;
}

void
edb_caption_omit_append(EdbContext ec, eid_t eid)
{

	if(bitree_search(ec->bt_caption_omit, "", (BiTree_hint_t)eid, NULL)) return;
	BiTreeAdd(ec->bt_caption_omit, "", (BiTree_hint_t)eid, NULL, NULL, NULL);
}

void
edb_caption_omit_remove(EdbContext ec, eid_t eid)
{
	char	*robj = NULL;

	if(bitree_search(ec->bt_caption_omit, "", (BiTree_hint_t)eid, NULL)==NULL) return;
	BiTreeRemove(ec->bt_caption_omit, "", (BiTree_hint_t)eid, NULL, (void *)&robj);
	FREE(robj);
}


