/*
$Id: file.c,v 1.18 2008/08/09 09:21:37 alex Exp $
	file.c--- file I/O
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: file.c,v 1.18 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<strings.h>
#include	"edb/edb.h"

static
char *
file_load(FILE *fp)
{
	char *s;
	long int n;

	fseek(fp, 0L, 2);
	n = ftell(fp);
	rewind(fp);
	s = (char *)malloc(sizeof(char)*(n+1));
	if( s == NULL ) {
		fprintf(stderr, "file_load: can't alloc memory\n");
		return NULL;
	}
	if(fread(s, sizeof(char), n, fp) != n) {
		fprintf(stderr, "file_load: file read error\n");
		return NULL;
	}

	s[n] = '\0';
	return s;
}

int
file_update(char *tmp_name, char *html_name)
{

	if(!access(html_name, F_OK)) {
		char	*old_name;

		asprintf(&old_name, "%s.old", html_name);
		rename(html_name, old_name);
		FREE(old_name);
	} else {
		printf("	new (%s)\n", html_name);
	}
	rename(tmp_name, html_name);
	return 1;
}

/*#file_symlink--- make symbolic link */
int
file_symlink(char *dst, char *src)
{

#if 0
	if(!access(src, F_OK)) unlink(src);
	symlink(dst, src);
#endif
	return 1;
}

/*#makedir_hierarchy--- make dir hierarchy */
int
makedir_hierarchy(char *dirname)
{
	char *s;

	if(!access(dirname, F_OK)) return TRUE;
	dirname = STRDUP(dirname);
	s = dirname;
	if(*s=='/') ++s;
	do {
		s=index(s, '/');
		if(s) *s = '\0';
		if(access(dirname, F_OK) && mkdir(dirname, 0755)) {
			if(s) *s = '/';
			return FALSE;
		}
		if(s) *s++ = '/';
	} while(s);
	return TRUE;
}

/*
	Get one line from file with memory allocation, and eliminate last newline.
	Notice:
		Returned string as C-string.
		So, string will be truncated if null character is contained.
*/
char *
file_getline(FILE *fp)
{
	register char	*line, *s;
	size_t	len;

	if(feof(fp)) return NULL;

	s = fgetln(fp, &len);
	if(s==NULL) return NULL;

	line = str_alloc(len+1);
	memcpy(line, s, len);
	line[len] = '\0';

	if((s = strchr(line, '\n'))) *s = '\0';

	return line;
}

/*#file_load_as_linked_ucs4_str-- load text file as linked wstring */
Link
file_load_as_linked_ucs4_str(char name[], FILE *err_fp)
{
	ucs4_t	*u;
	utf8_t	*s;
	FILE	*fp;
	Link	lustr;
	int	n;

	fp = fopen(name, "r");
	if(fp == NULL) {
		if(err_fp) fprintf(err_fp, "file_load_as_linked_ucs4_str: %s is not found.\n", name);
		return NULL;
	}
	s = (utf8_t *)file_load(fp);
	fclose(fp);

	if(s == NULL) return NULL;
	n = convUTF8toUCS4(s, NULL);
	u = (ucs4_t *)malloc(sizeof(ucs4_t)*(n+1));
	if(u == NULL) {
		if(err_fp) fprintf(err_fp, "file_load_as_linked_ucs4_str: malloc is failed.\n");
		FREE(s);
		return NULL;
	}
	convUTF8toUCS4(s, u);
	FREE(s);

	lustr = ucs4_str_split(u, UCS4_NL);
	FREE(u);
	return lustr;
}

/*#file_load_as_linked_str-- load text file as linked string */
Link
file_load_as_linked_str(char name[], FILE *err_fp)
{
	char	*s;
	FILE	*fp;
	Link	lstr;

	fp = fopen(name, "r");
	if(fp == NULL) {
		if(err_fp) fprintf(err_fp, "file_load_as_linked_str: %s is not found.\n", name);
		return NULL;
	}
	s = file_load(fp);
	fclose(fp);

	if(s == NULL) return NULL;

	lstr = str_split(s, '\n');
	FREE(s);
	return lstr;
}

#ifdef EDB_READER
#else /* !EDB_READER */
void
file_utf8_to_jis(FILE *in, FILE *out)
{
	utf8_t	*utf8;

	while((utf8=(utf8_t *)file_getline(in))) {
		char		*jis;
		ucs4_t	*ucs4;
		int		len = strlen((char *)utf8);

		ucs4 = (ucs4_t *)malloc(sizeof(ucs4_t)*(len+1));
		jis = (char *)malloc(sizeof(char)*(len*7+1));
		convUTF8toUCS4(utf8, ucs4);
		convUCS4toJIS(ucs4, (utf8_t *)jis);
		fputs(jis, out);
		fputs("\n", out);
		FREE(utf8);
		FREE(ucs4);
		FREE(jis);
	}
}
#endif /* EDB_READER */

