/*
	$Id: index.c,v 1.27 2009/07/31 11:26:47 alex Exp $
	index.c--- index info
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: index.c,v 1.27 2009/07/31 11:26:47 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"

#define	COLUMN_INDEX_TYPE	"edb_type"
#define	COLUMN_INDEX_TEID	"edb_teid"
#define	COLUMN_INDEX_PEID	"edb_peid"
#define	COLUMN_INDEX_EID		"edb_eid"

EdbIndexInfo
edb_index_new(EdbContext ec)
{
	EdbIndexInfo	ei;

	if(_edb_delayed_free_index) {
		++edb_delayed_free_recycle_count;
		EdbDelayedFreePop(_edb_delayed_free_index, ei);
	} else {
		ei = (EdbIndexInfo)malloc(sizeof(struct EDB_INDEX_INFO));
	}
	bzero(ei, sizeof(struct EDB_INDEX_INFO));
	edb_base_init(ec, &ei->base);
	return ei;
}

void
edb_index_free(EdbContext ec, EdbIndexInfo ei)
{

#if 0
	if(ei->bt_eid) bitree_free(ei->bt_eid, NULL);
#endif
	FREE(ei->reids);
	FREE(ei->peids);
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_index, ei);
	} else {
		Edb_free(ei);
	}
}


static
int
edb_eid_make_query_for_index(EdbContext ec, utf8_t **query, utf8_t *sqlname)
{
	int	bc;
	utf8_t	*base_query;

	bc = edb_base_make_query(ec, &base_query, sqlname);
	asprintf((char **)query, "%s,%s." COLUMN_INDEX_TEID ",%s." COLUMN_INDEX_PEID ",%s." COLUMN_INDEX_EID, base_query, sqlname, sqlname, sqlname);
	FREE(base_query);
	return bc + 3;
}

static
Link
edb_eid_get_lvalues_for_index(EdbContext ec, EdbIndexInfo ei, Link lres)
{
	register Link	l, l_eid;

	lres = edb_base_get_lvalues(ec, &ei->base, lres);
	ei->teid = ATOI((char *)lres->d);	lres = lres->n;
	l_eid = edb_array_split((utf8_t *)lres->d);	lres = lres->n;

	FREE(ei->peids);
	if(l_eid) {
		register eid_t	*p;

		p = ei->peids = (eid_t *)malloc(sizeof(eid_t)*(link_atoms(l_eid)+1));
		ForLink(l, l_eid) *p++ = ATOI(l->d);
		*p = 0;
	}
	EdbLinkFree(l_eid);
	l_eid = edb_array_split((utf8_t *)lres->d);	lres = lres->n;

	FREE(ei->reids);
	if(l_eid) {
		register eid_t	*p;

		p = ei->reids = (eid_t *)malloc(sizeof(eid_t)*(link_atoms(l_eid)+1));
		ForLink(l, l_eid) *p++ = ATOI(l->d);
		*p = 0;
	}
	EdbLinkFree(l_eid);
	return lres;
}

EdbIndexInfo
edb_index_get(EdbContext ec, eid_t eid)
{
	utf8_t	*index_query;
	utf8_t	*query;
	int	ic, nf;
	Link	lr, lres;
	EdbIndexInfo	ei;

	if((ei = edb_eid_list_restore_index(ec, eid))!=NULL) return ei;

	ic = edb_eid_make_query_for_index(ec, &index_query, (utf8_t *)"t1");
	asprintf((char **)&query,
		"SELECT %s FROM %s AS t1 WHERE t1." EDB_BASE_EID " = %d",
		index_query,
		EDB_INDEX_TABLE,
		eid);
	FREE(index_query);
	lres = edb_pgsql_query_one_tuple(ec, query, &nf);
	FREE(query);
	if(lres==NULL || nf !=ic) {
		EdbLinkFree(lres)
		return NULL;
	}

	ei = edb_index_new(ec);
	lr = edb_eid_get_lvalues_for_index(ec, ei, lres);
	EdbLinkFree(lres);

	ei = edb_eid_list_store_index(ec, eid, ei);

	return ei;
}

static
char *
catalogue_to_constraint(char *as, char *sname, char *sop, EdbCatalogue ca)
{
	Link	l, l_sql = NULL;
	char *sql, *cons = NULL;

	ForCatalogueLink(l, ca) {
		EdbTupleInfo	tuple = CatalogueObject(l);

		asprintf(&sql, "%s.%s%s%d", as, sname, sop, TupleEID(tuple));
		LinkAppend(l_sql, link_atom_new(sql, free));
	}
	asprintf(&cons, "(%s)", sql=str_concatenate_str(l_sql, " OR "));
	EdbLinkFree(l_sql);
	FREE(sql);
	return cons;
}

static
utf8_t *
edb_eid_get_tuple_make_constraint(EdbContext ec, utf8_t *sqlname, va_list ap)
{
	register int op;
	utf8_t	*cons, *s;
	register Link	l_cons = NULL;
	register Link	l_str = NULL;

	while((op = va_arg(ap, int)) != 0) {
		cons = NULL;
		switch(op) {
		case EDB_INDEX_GET_ref:
			asprintf((char **)&cons, " %s." EDB_BASE_AVAILABLE " AND %s." COLUMN_INDEX_EID " *= %d", sqlname, sqlname, va_arg(ap, eid_t));
			break;
		case EDB_INDEX_GET_own:
			asprintf((char **)&cons, " %s." EDB_BASE_OWNER " = %d", sqlname, va_arg(ap, eid_t));
			break;
		case EDB_INDEX_GET_perm:
			asprintf((char **)&cons, "%s." COLUMN_INDEX_PEID " *= %d", sqlname, va_arg(ap, eid_t));
			break;
		case EDB_INDEX_GET_map:
			asprintf((char **)&cons, "%s." EDB_BASE_MAPTO_EID " = %d", sqlname, va_arg(ap, eid_t));
			break;
		case EDB_INDEX_GET_fresh:
			asprintf((char **)&cons, "datetime(%s.%s) > datetime(%lu)", sqlname, EDB_BASE_MTIME, ec->now-va_arg(ap, int));
			break;
		case EDB_INDEX_GET_table:
			asprintf((char **)&cons, "%s." COLUMN_INDEX_TEID " = %d", sqlname, va_arg(ap, eid_t));
			break;
		case EDB_INDEX_GET_concern:
			l_str = NULL;
			op = va_arg(ap, eid_t);

			asprintf((char **)&cons, "%s." EDB_BASE_AVAILABLE
				" AND (%s." EDB_BASE_EID "=%d"
				" OR %s." EDB_BASE_MAPTO_EID "=%d"
				" OR %s." EDB_BASE_PARENT_EID "=%d"
				" OR %s." EDB_BASE_OWNER "=%d"
				" OR %s." COLUMN_INDEX_PEID "*=%d"
				" OR %s." COLUMN_INDEX_EID "*=%d"
				")",
				sqlname,
				sqlname, op,
				sqlname, op,
				sqlname, op,
				sqlname, op,
				sqlname, op,
				sqlname, op);
			break;
		case EDB_INDEX_GET_catalogue_ref:
			s = (utf8_t *)catalogue_to_constraint((char *)sqlname, COLUMN_INDEX_EID, "*=", va_arg(ap, EdbCatalogue));
			asprintf((char **)&cons, " %s." EDB_BASE_AVAILABLE " AND %s", sqlname, s);
			FREE(s);
			break;
		case EDB_INDEX_GET_catalogue_own:
			cons = (utf8_t *)catalogue_to_constraint((char *)sqlname, EDB_BASE_OWNER, "=", va_arg(ap, EdbCatalogue));
			break;
		case EDB_INDEX_GET_catalogue_perm:
			cons = (utf8_t *)catalogue_to_constraint((char *)sqlname, COLUMN_INDEX_PEID, "*=", va_arg(ap, EdbCatalogue));
			break;
		case EDB_INDEX_GET_catalogue_map:
			cons = (utf8_t *)catalogue_to_constraint((char *)sqlname, EDB_BASE_MAPTO_EID, "=", va_arg(ap, EdbCatalogue));
			break;
		default:
			break;
		}
		if(STRisVALID(cons)) LinkAppend(l_cons, link_atom_new(cons, free));
	}
	cons = (utf8_t *)str_concatenate_str(l_cons, " AND ");
	EdbLinkFree(l_cons);
	return cons;
}

Link
edb_eid_get_tuple_indexes(EdbContext ec, ...)
{
	va_list	ap;
	utf8_t	*index_query;
	utf8_t	*query0, *query;
	utf8_t	*constraint;
	int	ic, ni, nf;
	Link	l, lbres, l_index;

	ic = edb_eid_make_query_for_index(ec, &index_query, (utf8_t *)"t1");

	va_start(ap, ec);
	constraint = edb_eid_get_tuple_make_constraint(ec, (utf8_t *)"t1", ap);
	va_end(ap);

	asprintf((char **)&query0,
		"SELECT %s FROM %s AS t1 WHERE t1." COLUMN_INDEX_TEID "<>0%s%s",
		index_query,
		EDB_INDEX_TABLE,
		STRisVALID(constraint) ? " AND " : "",
		constraint);
	FREE(constraint);
	FREE(index_query);
	asprintf((char **)&query, "%s ORDER BY t1.%s DESC", query0, EDB_BASE_MTIME);
	FREE(query0);
	lbres = edb_pgsql_query(ec, query, &ni, &nf);
	FREE(query);
	if(lbres==NULL || nf !=ic) {
		EdbLinkFree(lbres);
		return NULL;
	}

	l_index = NULL;
	ForLink(l, lbres) {
		EdbIndexInfo	ei;

		ei = edb_index_new(ec);
		edb_eid_get_lvalues_for_index(ec, ei, (Link)l->d);
		ei = edb_eid_list_store_index(ec, ei->base.eid, ei);
		LinkAppendTop(link_atom_new(ei, NULL), l_index);
	}
	EdbLinkFree(lbres);

	return link_reverse(l_index);
}

int
edb_eid_get_number_of_tuples(EdbContext ec, ...)
{
	va_list	ap;
	utf8_t	*query;
	int	v;
	utf8_t	*constraint;

	va_start(ap, ec);
	constraint = edb_eid_get_tuple_make_constraint(ec, (utf8_t *)"t1", ap);
	va_end(ap);
	asprintf((char **)&query,
		"SELECT COUNT(*) FROM %s AS t1 WHERE t1." COLUMN_INDEX_TEID "<>0%s%s",
		EDB_INDEX_TABLE,
		STRisVALID(constraint) ? " AND " : "",
		constraint);
	FREE(constraint);
	v = edb_pgsql_query_int(ec, query);
	FREE(query);
	return v;
}

static
int
count_tuple(EdbContext ec, Link l_res, void *data)
{
	register eid_t	teid;
	register EdbCatalogue	ca_count = data;
	register EdbIndexCount	eic;
	register int	num;

	if(l_res==NULL || link_atoms(l_res)!=3) {
		EdbLinkFree(l_res);
		return FALSE;
	}
	teid = ATOI(l_res->d);
	eic = edb_catalogue_lookup(ca_count, teid);
	if(eic==NULL) {
		eic = (EdbIndexCount)malloc(sizeof(struct EDB_INDEX_COUNT));
		bzero(eic, sizeof(struct EDB_INDEX_COUNT));
		edb_catalogue_append(ca_count, teid, eic);
	}
	num = ATOI(l_res->n->n->d);
	if(!strcmp(l_res->n->d, "t")) {
		eic->avail += num;
	} else {
		eic->unavail += num;
	}
	return TRUE;
}


EdbCatalogue
edb_index_count_of_tuples(EdbContext ec, ...)
{
	va_list	ap;
	utf8_t	*query;
	int	v;
	utf8_t	*constraint;
	EdbCatalogue	ca_count;

	va_start(ap, ec);
	constraint = edb_eid_get_tuple_make_constraint(ec, (utf8_t *)"t1", ap);
	va_end(ap);
	asprintf((char **)&query,
		"SELECT " COLUMN_INDEX_TEID "," EDB_BASE_AVAILABLE ",COUNT(*)"
			" FROM " EDB_INDEX_TABLE " AS t1"
			" WHERE t1." COLUMN_INDEX_TEID "<>0%s%s"
			" GROUP BY " COLUMN_INDEX_TEID "," EDB_BASE_AVAILABLE,
			STRisVALID(constraint) ? " AND " : "", constraint);
	FREE(constraint);
	ca_count = edb_catalogue_new();
	v = edb_pgsql_query_callback(ec, query, count_tuple, ca_count);
	FREE(query);
	return ca_count;
}

void
edb_index_count_free(EdbContext ec, EdbCatalogue ca_count)
{
	Link	l;

	ForCatalogueLink(l, ca_count) FREE(CatalogueObject(l));
	edb_catalogue_free(ca_count);
}

Link
edb_eid_get_tuple_indexes_from_table(EdbContext ec, EdbTableInfo ti)
{
	Link	l, l_index;

	if(ti->l_all_indexes==NULL) {
		utf8_t	*base_query;
		utf8_t	*orderby, *order, *query;
		int	bc, ni, nf;
		Link	lbres;

		if((order = edb_table_make_order(ec, ti, (utf8_t *)"t2"))!=NULL) {
			asprintf((char **)&orderby, " ORDER BY %s", order);
			FREE(order);
		} else {
			orderby = (utf8_t *)strdup("");
		}

		bc = edb_eid_make_query_for_index(ec, &base_query, (utf8_t *)"t1");
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1, %s AS t2 WHERE t1." COLUMN_INDEX_TEID " = %d AND t1." EDB_BASE_EOID "=t2." EDB_BASE_EOID " %s",
			base_query,
			EDB_INDEX_TABLE,
			TABLE_SN(ti),
			TABLE_EID(ti),
			orderby);
		FREE(base_query);
		FREE(orderby);

		lbres = edb_pgsql_query(ec, query, &ni, &nf);
		FREE(query);
		if(lbres==NULL || nf !=bc) {
			EdbLinkFree(lbres);
			return NULL;
		}

		l_index = NULL;
		ForLink(l, lbres) {
			EdbIndexInfo	ei;

			ei = edb_index_new(ec);
			edb_eid_get_lvalues_for_index(ec, ei, (Link)l->d);
			ei = edb_eid_list_store_index(ec, ei->base.eid, ei);
			LinkAppendTop(link_atom_new(ei, NULL), l_index);
		}
		EdbLinkFree(lbres);
		LinkReverse(l_index);
		if(ec->readonly) {
			ti->l_all_indexes = l_index;
		} else {
			return l_index;
		}
	}
	l_index = NULL;
	ForLink(l, ti->l_all_indexes) {
		LinkAppendTop(link_atom_new(l->d, NULL), l_index);
	}
	return link_reverse(l_index);
}

Link
edb_eid_get_authed_tuple_indexes_from_table(EdbContext ec, EdbTableInfo ti)
{
	utf8_t	*base_query;
	utf8_t	*query;
	utf8_t	*orderby;
	utf8_t	*order;
	int	bc, ni, nf;
	Link	l, lbres, l_index;

	if((order = edb_table_make_order(ec, ti, (utf8_t *)"t2"))!=NULL) {
		asprintf((char **)&orderby, " ORDER BY %s", order);
		FREE(order);
	} else {
		orderby = (utf8_t *)strdup("");
	}

	bc = edb_eid_make_query_for_index(ec, &base_query, (utf8_t *)"t1");
	asprintf((char **)&query,
		"SELECT %s FROM %s AS t1, %s AS t2, %s AS t3 WHERE t1." COLUMN_INDEX_TEID "=%d AND t1." EDB_BASE_EOID "=t2." EDB_BASE_EOID " AND t1." EDB_BASE_EID "=t3." EDB_BASE_EID " AND t3." EDB_BASE_AVAILABLE " %s",
		base_query,
		EDB_INDEX_TABLE,
		TABLE_SN(ti),
		EDB_AUTH_TABLE,
		TABLE_EID(ti),
		orderby);
	FREE(base_query);
	FREE(orderby);

	lbres = edb_pgsql_query(ec, query, &ni, &nf);
	FREE(query);
	if(lbres==NULL || nf !=bc) {
		EdbLinkFree(lbres);
		return NULL;
	}

	l_index = NULL;
	ForLink(l, lbres) {
		EdbIndexInfo	ei;

		ei = edb_index_new(ec);
		edb_eid_get_lvalues_for_index(ec, ei, (Link)l->d);
		ei = edb_eid_list_store_index(ec, ei->base.eid, ei);
		LinkAppendTop(link_atom_new(ei, NULL), l_index);
	}
	EdbLinkFree(lbres);

	return link_reverse(l_index);
}

int
edb_eid_get_number_of_available_tuples(EdbContext ec, EdbTableInfo ti)
{
	utf8_t	*query;
	int	v;

	asprintf((char **)&query, "SELECT COUNT(*) FROM %s WHERE " EDB_BASE_AVAILABLE " AND " COLUMN_INDEX_TEID "=%d",
		EDB_INDEX_TABLE, TABLE_EID(ti));
	v = edb_pgsql_query_int(ec, query);
	FREE(query);
	return v;
}

eid_t
edb_eid_isin_table(EdbContext ec, eid_t eid)
{
	EdbIndexInfo	ei;

	ei = edb_index_get(ec, eid);
	if(ei) return ei->teid;
	return 0;
}

eid_t
edb_eid_isin_table0(EdbContext ec, eid_t eid)
{
	EdbIndexInfo	ei;

	ei = edb_index_get(ec, eid);
	if(ei) return ei->teid;
	return -1;
}

int
edb_eid_referenced_table(EdbContext ec, eid_t target_eid, int **refs)
{
	utf8_t	*query;
	int	n;

	asprintf((char **)&query,
		"SELECT " COLUMN_INDEX_TEID " FROM %s WHERE " EDB_BASE_AVAILABLE " AND " COLUMN_INDEX_EID " *= %d GROUP BY " COLUMN_INDEX_TEID,
		EDB_INDEX_TABLE,
		target_eid);
	edb_pgsql_query_integers(ec, refs, &n, query);
	FREE(query);
	return n;
}

Link
edb_eid_is_referenced_from_table(EdbContext ec, EdbTableInfo ti, eid_t target_eid)
{
	int	n;
	utf8_t	*query;
	utf8_t	*orderby;
	utf8_t	*order;
	Link	l_eid;

	if((order = edb_table_make_order(ec, ti, (utf8_t *)"t2"))!=NULL) {
		asprintf((char **)&orderby, " ORDER BY %s", order);
		FREE(order);
	} else {
		orderby = (utf8_t *)strdup("");
	}

	asprintf((char **)&query,
		"SELECT t1." EDB_BASE_EID " FROM %s AS t1, %s AS t2 WHERE t1." EDB_BASE_AVAILABLE " AND t1." COLUMN_INDEX_TEID " = %d AND t1." COLUMN_INDEX_EID " *= %d AND t1." EDB_BASE_EOID " = t2." EDB_BASE_EOID " %s",
		EDB_INDEX_TABLE,
		TABLE_SN(ti),
		TABLE_EID(ti),
		target_eid,
		orderby);
	FREE(orderby);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}

int
edb_eid_owned_table(EdbContext ec, eid_t target_eid, int **refs)
{
	utf8_t	*query;
	int	n;

	asprintf((char **)&query,
		"SELECT " COLUMN_INDEX_TEID " FROM " EDB_INDEX_TABLE
		" WHERE " EDB_BASE_OWNER "=%d"
		" GROUP BY " COLUMN_INDEX_TEID,
		target_eid);
	edb_pgsql_query_integers(ec, refs, &n, query);
	FREE(query);
	return n;
}

Link
edb_eid_is_owned_from_table(EdbContext ec, EdbTableInfo ti, eid_t target_eid)
{
	int	n;
	utf8_t	*query;
	utf8_t	*orderby;
	utf8_t	*order;
	Link	l_eid;

	if((order = edb_table_make_order(ec, ti, (utf8_t *)"t2"))!=NULL) {
		asprintf((char **)&orderby, " ORDER BY %s", order);
		FREE(order);
	} else {
		orderby = (utf8_t *)strdup("");
	}

	asprintf((char **)&query,
		"SELECT t1." EDB_BASE_EID " FROM %s AS t1, %s AS t2 WHERE t1." COLUMN_INDEX_TEID " = %d AND t1." EDB_BASE_OWNER " = %d AND t1." EDB_BASE_EOID " = t2." EDB_BASE_EOID " %s",
		EDB_INDEX_TABLE,
		TABLE_SN(ti),
		TABLE_EID(ti),
		target_eid,
		orderby);
	FREE(orderby);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}

int
edb_eid_peid_table(EdbContext ec, eid_t target_eid, int **refs)
{
	utf8_t	*query;
	int	n;

	asprintf((char **)&query,
		"SELECT " COLUMN_INDEX_TEID " FROM " EDB_INDEX_TABLE
		" WHERE " COLUMN_INDEX_PEID "*=%d"
		" GROUP BY " COLUMN_INDEX_TEID,
		target_eid);
	edb_pgsql_query_integers(ec, refs, &n, query);
	FREE(query);
	return n;
}

Link
edb_eid_is_peid_from_table(EdbContext ec, EdbTableInfo ti, eid_t target_eid)
{
	int	n;
	utf8_t	*query;
	utf8_t	*orderby;
	utf8_t	*order;
	Link	l_eid;

	if((order = edb_table_make_order(ec, ti, (utf8_t *)"t2"))!=NULL) {
		asprintf((char **)&orderby, " ORDER BY %s", order);
		FREE(order);
	} else {
		orderby = (utf8_t *)strdup("");
	}

	asprintf((char **)&query,
		"SELECT t1." EDB_BASE_EID " FROM %s AS t1, %s AS t2 WHERE t1." COLUMN_INDEX_TEID " = %d AND t1." COLUMN_INDEX_PEID " *= %d AND t1." EDB_BASE_EOID " = t2." EDB_BASE_EOID " %s",
		EDB_INDEX_TABLE,
		TABLE_SN(ti),
		TABLE_EID(ti),
		target_eid,
		orderby);
	FREE(orderby);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}

int
edb_eid_mapto_table(EdbContext ec, eid_t target_eid, int **refs)
{
	utf8_t	*query;
	int	n;

	asprintf((char **)&query,
		"SELECT " COLUMN_INDEX_TEID " FROM %s WHERE " EDB_BASE_MAPTO_EID "=%d GROUP BY " COLUMN_INDEX_TEID,
		EDB_INDEX_TABLE,
		target_eid);
	edb_pgsql_query_integers(ec, refs, &n, query);
	FREE(query);
	return n;
}

Link
edb_eid_is_mapto_from_table(EdbContext ec, EdbTableInfo ti, eid_t target_eid)
{
	int	n;
	utf8_t	*query;
	utf8_t	*orderby;
	utf8_t	*order;
	Link	l_eid;

	if(!EIDisVALID(target_eid)) return NULL;
	if((order = edb_table_make_order(ec, ti, (utf8_t *)"t2"))!=NULL) {
		asprintf((char **)&orderby, " ORDER BY %s", order);
		FREE(order);
	} else {
		orderby = (utf8_t *)strdup("");
	}

	asprintf((char **)&query,
		"SELECT t1." EDB_BASE_EID " FROM %s AS t1, %s AS t2 WHERE t1." COLUMN_INDEX_TEID "=%d AND t1." EDB_BASE_MAPTO_EID " = %d AND t1." EDB_BASE_EOID " = t2." EDB_BASE_EOID " %s",
		EDB_INDEX_TABLE,
		TABLE_SN(ti),
		TABLE_EID(ti),
		target_eid,
		orderby);
	FREE(orderby);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}

#if defined(UNUSED)
static
Link
edb_eid_reference(EdbContext ec, char *fmt, ...)
{
	va_list	ap;
	utf8_t	*query;
	utf8_t	*constraint;
	int	n;
	Link	l_eid;

	va_start(ap, fmt);
	vasprintf((char **)&constraint, fmt, ap);
	va_end(ap);
	asprintf((char **)&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s", EDB_INDEX_TABLE, constraint);
	FREE(constraint);

	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}
#endif

int
edb_eid_is_referenced(EdbContext ec, eid_t eid)
{
	utf8_t	*query;
	int	n;

	asprintf((char **)&query,
		"SELECT COUNT(*) FROM " EDB_INDEX_TABLE
		" WHERE "
			EDB_BASE_EID "<>%d AND " EDB_BASE_AVAILABLE
			" AND ("
				EDB_BASE_MAPTO_EID "=%d OR "
				EDB_BASE_PARENT_EID "=%d OR "
				EDB_BASE_OWNER "=%d OR "
				COLUMN_INDEX_EID "*=%d"
			")",
		eid, eid, eid, eid, eid);
	n = edb_pgsql_query_int(ec, query);
	FREE(query);
	return (n!=0);
}

int
edb_eid_is_mapped(EdbContext ec, eid_t eid)
{
	utf8_t	*query;
	int	v;

	asprintf((char **)&query, "SELECT COUNT(*) FROM %s WHERE %s <> %d AND %s AND %s = %d",
		EDB_INDEX_TABLE,
		EDB_BASE_EID, eid,
		EDB_BASE_AVAILABLE,
		EDB_BASE_MAPTO_EID, eid);
	v = edb_pgsql_query_int(ec, query);
	FREE(query);
	return v;
}

int
edb_eid_is_owner(EdbContext ec, eid_t eid)
{
	utf8_t	*query;
	int	v;

	asprintf((char **)&query, "SELECT COUNT(*) FROM %s WHERE %s AND %s = %d",
		EDB_INDEX_TABLE,
		EDB_BASE_AVAILABLE,
		EDB_BASE_OWNER, eid);
	v = edb_pgsql_query_int(ec, query);
	FREE(query);
	return v;
}

int
edb_eid_is_available(EdbContext ec, eid_t eid)
{
	utf8_t	*query;
	int	n, av;
	Link	l;

	asprintf((char **)&query, "SELECT " EDB_BASE_AVAILABLE " FROM %s WHERE %s = %d",
		EDB_INDEX_TABLE,
		EDB_BASE_EID, eid);
	l = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	if(l==NULL) return FALSE;
	av = !strcmp(l->d, "t");
	EdbLinkFree(l);
	return av;
}

int
edb_eid_tuple_is_available(EdbContext ec, eid_t eid)
{
	utf8_t	*query;
	int	n, av;
	Link	l;

	asprintf((char **)&query, "SELECT " EDB_BASE_AVAILABLE " FROM %s WHERE %s = %d AND " COLUMN_INDEX_TEID "<>0",
		EDB_INDEX_TABLE,
		EDB_BASE_EID, eid);
	l = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	if(l==NULL) return FALSE;
	av = !strcmp(l->d, "t");
	EdbLinkFree(l);
	return av;
}


