/*
$Id: link.c,v 1.19 2012/04/03 03:51:13 alex Exp alex $
	link.c--- link control
	(C) Copyright 1994 by Alex
*/

#ifndef lint
static char * rcsid __attribute__((unused)) = "$Id: link.c,v 1.19 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<strings.h>
#include	<stdarg.h>
#include	"edb/link.h"

#if defined(__LIBEDB__)
extern int	_edb_delayed_free_mode;
extern void	*_edb_delayed_free_link_atom;
extern int	edb_delayed_free_recycle_count;
#define	EdbDelayedFreePush(list, obj)	{	*(void **)(obj) = (list);	(list) = (obj);	}
#define	EdbDelayedFreePop(list, obj)	{	(obj) = (list);	(list) = *(void **)(obj);	}
#endif

/*#link_atom_new--- new link atom */
Link
link_atom_new(
	register void *d,
	register void (*free_func)())
{
	register Link l;

#if defined(__LIBEDB__)
	if(_edb_delayed_free_link_atom) {
		++edb_delayed_free_recycle_count;
		EdbDelayedFreePop(_edb_delayed_free_link_atom, l);
	} else {
		l = (Link)malloc((size_t)sizeof(struct LINK));
	}
#else
	l = (Link)malloc((size_t)sizeof(struct LINK));
#endif
	if(l == NULL) return NULL;
	l->p = NULL;
	l->n = NULL;
	l->d = d;
	l->f = free_func;
	return l;
}

/*#link_atom_free--- free link atom */
void
link_atom_free(register Link l)
{

#if defined(__LIBEDB__)
	if(l->d && l->f) (*l->f)(l->d);
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_link_atom, l);
	} else {
		free(l);
	}
#else
	if(l->d && l->f) (*l->f)(l->d);
	free(l);
#endif
}

/*#link_atom_free2--- free link atom */
void
link_atom_free2(
	register void		*cdata,
	register Link		l)
{

#if defined(__LIBEDB__)
	if(l->d && l->f) (*l->f)(cdata, l->d);
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_link_atom, l);
	} else {
		free(l);
	}
#else
	if(l->d && l->f) (*l->f)(cdata, l->d);
	free(l);
#endif
}

/*#link_free--- free link */
void
link_free(register Link l)
{

	while(l) {
		register Link n;

		n = l->n;
		link_atom_free(l);
		l = n;
	}
}

/*#link_free2--- free link */
void
link_free2(
	register void		*cdata,
	register Link		l)
{

	while(l) {
		register Link n;

		n = l->n;
		link_atom_free2(cdata, l);
		l = n;
	}
}

/*#link_head--- head of link */
Link
link_head(register Link l)
{

	if(l==NULL) return l;
	while(l->p) l = l->p;
	return l;
}

/*#link_tail--- tail of link */
Link
link_tail(register Link l)
{

	if(l==NULL) return l;
	while(l->n) l = l->n;
	return l;
}

/*#link_get_obj--- get nth object */
void	*
link_get_obj(
	register Link	l,
	register int	n)
{

	while(n-->0 && l) l = l->n;
	if(l) {
		return l->d;
	} else {
		return NULL;
	}
}

/*#link_get_link--- get nth object */
Link
link_get_link(
	register Link	l,
	register int	n)
{

	while(n-->0 && l) l = l->n;
	if(l) {
		return l;
	} else {
		return NULL;
	}
}

/*#link_append--- append atom */
Link
link_append(
	register Link	ll,
	register Link	l)
{
	register Link t;

	if(ll==NULL) return l;
	if(l==NULL) return ll;
	for(t=ll; t->n; t=t->n) ;	/* t = link_tail(ll); */
	t->n = l;
	l->p = t;
	return ll;
}

/*#link_insert--- append atom */
Link
link_insert(
	register Link	ll,
	register int	nth,	/* 0.. */
	register Link	l)
{
	register Link t, n;

	if(ll==NULL) return l;
	if(l==NULL) return ll;
	if(nth==0) return link_append(l, ll);
	t = ll;
	while(nth-->0 && t && t->n) t = t->n;
	n = t->n;
	t->n = l;	l->p = t;
	if(n) {
		l = link_tail(l);
		l->n = n;
		n->p = l;
	}
	return ll;
}

/*#link_atom_move_top--- move to top */
Link
link_atom_move_top(
	register Link	ll,
	register Link	l)
{

	if(l==ll) {	/* already top */
		return ll;
	} else if(l->n == NULL) {	/* locate end */
		l->p->n = NULL;
		l->p = NULL;
		return link_append(l, ll);
	} else {
		l->p->n = l->n;
		l->n->p = l->p;
		l->p = l->n = NULL;
		return link_append(l, ll);
	}
}

/*#link_atoms--- number of atoms */
int
link_atoms(register Link l)
{
	register int n = 0;

	while(l) {
		++n;
		l = l->n;
	}
	return n;
}

/*#link_reverse--- rearrangement reverse order */
Link
link_reverse(register Link l)
{
	register Link	n;

	if(l==NULL || l->n==NULL) return l;
	while(l) {
		n = l->n;
		l->n = l->p;
		l->p = n;
		if(n==NULL) break;
		l = n;
	}
	return l;
}

/*#link_sort--- sort link */
Link
link_sort(
	register Link	ln,
	register int	(*compare_func)(),
	register void	* (*aux_func)(),
	register void	* aux_data,
	register void	(*aux_free)())
{
	register Link	l;
	register int	i, n;
	register LinkObj	objs;
	int	tobe_free = 0;

	n = link_atoms(ln);
	if(n<2) return ln;
	if(n<1024) {	// small size
		objs = (LinkObj)alloca(sizeof(struct LINK_OBJ)*n);
	} else {
		objs = (LinkObj)malloc(sizeof(struct LINK_OBJ)*n);
		tobe_free = 1;
	}
	bzero(objs, sizeof(struct LINK_OBJ)*n);
	if(aux_func) {
		for(i=0, l=ln; i<n && l; ++i, l=l->n) {
			objs[i].d = l->d;
			objs[i].idx = i;
			objs[i].f = l->f;
			objs[i].aux = (*aux_func)(l->d, aux_data);
		}
	} else {
		for(i=0, l=ln; i<n && l; ++i, l=l->n) {
			objs[i].d = l->d;
			objs[i].idx = i;
			objs[i].f = l->f;
			objs[i].aux = NULL;
		}
	}
	qsort(objs, (size_t)n, sizeof(struct LINK_OBJ), compare_func);
	if(aux_free) {
		for(i=0, l=ln; i<n && l ; ++i, l=l->n) {
			l->d = objs[i].d;
			l->f = objs[i].f;
			if(objs[i].aux) (*aux_free)(objs[i].aux);
		}
	} else {
		for(i=0, l=ln; i<n && l ; ++i, l=l->n) {
			l->d = objs[i].d;
			l->f = objs[i].f;
		}
	}
	if(tobe_free) free(objs);
	return ln;
}

/*#link_randomize_compare--- compare for qsort */
static
int
link_randomize_compare(register LinkObj o1, register LinkObj o2)
{
	register int	d1 = *((int *)o1->aux);
	register int	d2 = *((int *)o2->aux);

	return (int)(d1 - d2);
}

static
void *
link_randomize_aux(register void *o1, register void *data)
{
	int	*d;

	d = (int *)malloc(sizeof(int));
	*d = random();
	return d;
}

/*#link_randomize---  */
Link
link_randomize(register Link ln)
{

	return link_sort(ln, link_randomize_compare, link_randomize_aux, NULL, free);
}

void	*
link_pop(register Link *link)
{
	register void		*d = NULL;
	register Link		l;

	if(*link==NULL) return NULL;
#if defined(__LIBEDB__)
	l = *link;
	if((*link = l->n)) (*link)->p = NULL;
	d = l->d;
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_link_atom, l);
	} else {
		free(l);
	}
	return d;
#else
	l = *link;
	if(*link = l->n) (*link)->p = NULL;
	d = l->d;
	free(l);
	return d;
#endif
}

Link
link_atom_new_sdup(char *s)
{

	return link_atom_new(strdup(s ? s : ""), free);
}

static
char *
my_strndup(register char *s, register int n)
{
	register char *buf;

	if(s == NULL) return NULL;
	buf = (char *)malloc((n+1) * sizeof(char));
	if(buf == NULL) return NULL;
	strncpy(buf, s, n);
	buf[n] = '\0';
	return buf;
}

Link
link_atom_new_sndup(char *s, size_t len)
{

	return link_atom_new(s ? my_strndup(s, len) : strdup(""), free);
}

Link
link_atom_new_i2s(int i)
{
	char *s;

	asprintf(&s, "%d", i);
	return link_atom_new(s, free);
}

Link
link_atom_new_spf(char *fmt, ...)
{
	va_list	ap;
	char	*s;

	va_start(ap, fmt);
	vasprintf(&s, fmt, ap);
	va_end(ap);
	return link_atom_new(s ? s : strdup(""), free);
}


