/*
	$Id: misc.c,v 1.35 2012/04/03 03:51:13 alex Exp alex $
	misc.c--- miscellaneous functions
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: misc.c,v 1.35 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/html.h"

Link
edb_array_split(utf8_t *str)
{
	ucs4_t	*ustr, *w, *p;
	Link	lustr, lstr, l;
	int	quot, bs;

	ustr = UTF8toUCS4(str);
	lustr = NULL;
	w = ustr;
	if(*w == '{') ++w;
	bs = quot = FALSE;
	p = w;
	while(*w) {
		if(bs) {
			bs = FALSE;
		} else if(quot) {
			if(*w == '"') {
				*w = UCS4_NULL;
				LinkAppendTop(link_atom_new(ucs4_strdup(p), free), lustr);
				p = w+1;
				quot = FALSE;
			} else if(*w == '\\') {
				bs = TRUE;
			}
		} else if(*w == '"') {
			quot = TRUE;
			p = w+1;
#if 0
		} else if(*w == UCS4_SPACE) {
			p = w+1;
#endif
		} else if(*w == ',' || *w == '}') {
			*w = UCS4_NULL;
			if(p!=w) LinkAppendTop(link_atom_new(ucs4_strdup(p), free), lustr);
			p = w+1;
		}
		++w;
	}
	LinkReverse(lustr);
	lstr = NULL;
	ForLink(l, lustr) {
		utf8_t	*u;

		u = UCS4toUTF8((ucs4_t *)l->d);
		LinkAppendTop(link_atom_new_sdup((char *)u), lstr);
	}
	link_free(lustr);
	return link_reverse(lstr);
}

#if defined(UNUSED)
static
utf8_t *
edb_mtext_english(EdbContext ec, EdbMtext mt)
{

	if(mt->english && strlen((char *)mt->english)>0) return mt->english;
	if(mt->japanese && strlen((char *)mt->japanese)>0 && edb_text_isEnglish(ec, mt->japanese)) return mt->japanese;
	return (utf8_t *)"";
}

static
utf8_t *
edb_mtext_japanese(EdbContext ec, EdbMtext mt)
{

	if(mt->japanese && strlen((char *)mt->japanese)>0) return mt->japanese;
	if(mt->pronounce && strlen((char *)mt->pronounce)>0) return mt->pronounce;
	if(mt->english && strlen((char *)mt->english)>0) return mt->english;
	return (utf8_t *)"";
}

static
utf8_t *
edb_mtext_pronounce(EdbContext ec, EdbMtext mt)
{

	if(mt->pronounce && strlen((char *)mt->pronounce)>0) return mt->pronounce;
	if(mt->japanese && strlen((char *)mt->japanese)>0 && edb_text_isPronounce(ec, mt->japanese)) return mt->japanese;
	return (utf8_t *)"";
}
#endif

static	ucs4_t	ucs4_null_string[1] = {UCS4_NULL};
static int	edb_text_conv_table_initialized = FALSE;

static	Link		conv_list = NULL;
static	BiTree	conv_utf8_name_list = NULL;
static	BiTree	conv_ucs4_name_list = NULL;
static	BiTree	conv_source_list = NULL;
static	BiTree	conv_inner_list = NULL;
static	int		conv_max = 1;

#define	DEFAULT_CONV_TABLE	EDBDIR "/libdata/tbl/ustrconv.tbl"

/*
	conv_new
	conv_free
*/
static
TextConversion
conv_new(utf8_t *n, ucs4_t *s, ucs4_t *d, utf8_t *t)
{
	TextConversion	cnv;

	cnv = (TextConversion)malloc(sizeof(struct TEXT_CONVERSION));
	if(cnv == NULL) return NULL;

	cnv->utf8_name = (utf8_t *)(n ? strdup((char *)n) : NULL);
	if(cnv->utf8_name) {
		cnv->ucs4_name = UTF8toUCS4(cnv->utf8_name);
		cnv->ucs4_name = ucs4_strdup(cnv->ucs4_name);
	} else {
		cnv->ucs4_name = NULL;
	}
	cnv->ucs4_src = ucs4_strdup(s);
	cnv->ucs4_d = ucs4_strdup(d);
	cnv->ucs4_dlen = ucs4_strlen(cnv->ucs4_d);
	cnv->utf8_d = UCS4toUTF8(cnv->ucs4_d);
	cnv->utf8_d = (utf8_t *)strdup((char *)cnv->utf8_d);
	cnv->ucs4_tex = UTF8toUCS4(t);
	cnv->ucs4_tex = ucs4_strdup(cnv->ucs4_tex);
	cnv->utf8_tex = (utf8_t *)strdup((char *)t);
	cnv->count = 0;
	return cnv;
}

static
void
conv_free(TextConversion cnv)
{

	FREE(cnv->utf8_name);
	FREE(cnv->ucs4_name);

	FREE(cnv->ucs4_src);
	FREE(cnv->ucs4_d);
	FREE(cnv->utf8_d);

	FREE(cnv->ucs4_tex);
	FREE(cnv->utf8_tex);

	free(cnv);
}

static
void
table_unescape(register ucs4_t *s)
{
	register ucs4_t	*d, u;

	d = s;
	while((u=*s++)) {
		if(u=='\\') {
			ucs4_t	c;

			c = 0;
			while((u=*s)) {
				if(ucs4_isdigit(u)) {
					c <<= 4;
					c |= (u - '0')&0xf;
					++s;
				} else if('A'<=u && u<='F') {
					c <<= 4;
					c |= (u - 'A' + 10)&0xf;
					++s;
				} else if('a'<=u && u<='f') {
					c <<= 4;
					c |= (u - 'a' + 10)&0xf;
					++s;
				} else {
					break;
				}
			}
			*d++ = c;
		} else {
			*d++ = u;
		}
	}
	*d = UCS4_NULL;
}

static
int
conv_name_compare(register void *o1, register void *o2)
{
	register TextConversion	cnv1 = (TextConversion)o1;
	register TextConversion	cnv2 = (TextConversion)o2;

	return strcmp((char *)cnv1->utf8_name, (char *)cnv2->utf8_name);
}

static
int
conv_ucs4_name_compare(register void *o1, register void *o2)
{
	register TextConversion	cnv1 = (TextConversion)o1;
	register TextConversion	cnv2 = (TextConversion)o2;

	return ucs4_strcmp(cnv1->ucs4_name, cnv2->ucs4_name);
}

static
int
conv_name_compare_for_search(register void *o1, register void *o2)
{
	register char	*name = (char *)o1;
	register TextConversion	cnv2 = (TextConversion)o2;

	return strcmp(name, (char *)cnv2->utf8_name);
}

static
int
conv_ucs4_name_compare_for_search(register void *o1, register void *o2)
{
	register ucs4_t	*name = (ucs4_t *)o1;
	register TextConversion	cnv2 = (TextConversion)o2;

	return ucs4_strcmp(name, cnv2->ucs4_name);
}

void
edb_text_conv_initialize(char *name)
{
	Link	lustr, l;
	static ucs4_t	codestr[] = {'\\', 'C', '{', '}', UCS4_NULL};
	static ucs4_t	nonestr[] = {'.', UCS4_NULL};
	static ucs4_t	nullstr[] = {UCS4_NULL};

	if(edb_text_conv_table_initialized) return;

	if(name==NULL) name = DEFAULT_CONV_TABLE;
	lustr = file_load_as_linked_ucs4_str(name, stderr);
	ForLink(l, lustr) {
		ucs4_t	*w, *s, *d;
		utf8_t	*n, *t;
		Link	ll;
		int	r;
		TextConversion	cnv;

		w = (ucs4_t *)l->d;
		if(w[0] == '#') continue;
		if(ucs4_strlen(w)==0) continue;
		ll = ucs4_str_split(w, UCS4_TAB);
		if(link_atoms(ll) < 4) {
			fprintf(stderr, "conv_init: ERROR : '%s'\n", UCS4toUTF8(w));
			continue;
		}
		/* name */
		w = (ucs4_t *)ll->d;
		n = ucs4_strcmp(w, nonestr) ? UCS4toUTF8(w) : NULL;

		/* source */
		w = (ucs4_t *)ll->n->d;
		table_unescape(w);
		s = ucs4_strcmp(w, nonestr) ? w : nullstr;

		/* destination */
		w = (ucs4_t *)ll->n->n->d;
		table_unescape(w);
		d = w;

		/* TeX */
		w = (ucs4_t *)ll->n->n->n->d;
		t = ucs4_strcmp(w, nonestr) ? UCS4toUTF8(w) : (utf8_t *)"";

#if 0
		if(*s == '\\' && ucs4_strlen(&s[1]) == 4) {
			char *str = UCS4toUTF8(&s[1]);
			int	code;
			ucs4_t	ustr[2];

			sscanf(str, "%x", &code);
			ustr[0] = code;
			ustr[1] = UCS4_NULL;
			ll->n->d = s = ucs4_strdup(ustr);
		}
#endif
		if(!ucs4_strcmp(d, codestr) && n) {
			char	dst[BUFSIZ];

			sprintf(dst, "\\C{%s}", n);
			d = UTF8toUCS4((utf8_t *)dst);
		} else {
			table_unescape(d);
		}
		if(UCS4_STRisVALID(s)) {
			r = ucs4_strlen(d);
			if(r>conv_max) conv_max = r;
		}
		LinkAppendTop(link_atom_new((cnv=conv_new(n, s, d, t)), conv_free), conv_list);
		if(STRisVALID(cnv->utf8_name)) {
			BiTreeAdd(conv_utf8_name_list, cnv, bitree_hint_str((char *)cnv->utf8_name), conv_name_compare, NULL, NULL);
			BiTreeAdd(conv_ucs4_name_list, cnv, bitree_hint_ucs4str(cnv->ucs4_name), conv_ucs4_name_compare, NULL, NULL);
		}
		for(s=cnv->ucs4_src; *s; ++s) {
			BiTreeAdd(conv_source_list, cnv, *s, NULL, NULL, NULL);
		}
		if(cnv->ucs4_dlen==1) {
			BiTreeAdd(conv_inner_list, cnv, cnv->ucs4_d[0], NULL, NULL, NULL);
		}
		link_free(ll);
	}
	LinkReverse(conv_list);
	link_free(lustr);

	BiTreeRearrange(conv_utf8_name_list);
	BiTreeRearrange(conv_ucs4_name_list);
	BiTreeRearrange(conv_source_list);
	BiTreeRearrange(conv_inner_list);

	edb_text_conv_table_initialized = TRUE;
}

static ucs4_t text_conv_WKvm[] = {
/*		0		1		2		3		4		5		6		7		8		9		A		B		C		D		E		F		*/
/* hiragana */
/*3040*/	0,		0,		0,		0,		0,		0,		0x3094,	0,		0,		0,		0,		0x304c,	0,		0x304e,	0,		0x3050,	/*3040*/
/*3050*/	0,		0x3052,	0,		0x3054,	0,		0x3056,	0,		0x3058,	0,		0x305a,	0,		0x305c,	0,		0x305e,	0,		0x3060,	/*3050*/
/*3060*/	0,		0x3062,	0,		0,		0x3065,	0,		0x3067,	0,		0x3069,	0,		0,		0,		0,		0,		0,		0x3070,	/*3060*/
/*3070*/	0,		0,		0x3073,	0,		0,		0x3076,	0,		0,		0x3079,	0,		0,		0x307c,	0,		0,		0,		0,		/*3070*/
/*3080*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*3080*/
/*3090*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0x309e,	0,		0,		/*3090*/
/* katakana */
/*30A0*/	0,		0,		0,		0,		0,		0,		0x30f4,	0,		0,		0,		0,		0x30ac,	0,		0x30ae,	0,		0x30b0,	/*30A0*/
/*30B0*/	0,		0x30b2,	0,		0x30b4,	0,		0x30b6,	0,		0x30b8,	0,		0x30ba,	0,		0x30bc,	0,		0x30be,	0,		0x30c0,	/*30B0*/
/*30C0*/	0,		0x30c2,	0,		0,		0x30c5,	0,		0x30c7,	0,		0x30c9,	0,		0,		0,		0,		0,		0,		0x30d0,	/*30C0*/
/*30D0*/	0,		0,		0x30d3,	0,		0,		0x30d6,	0,		0,		0x30d9,	0,		0,		0x30dc,	0,		0,		0,		0,		/*30D0*/
/*30E0*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0x30f7,	/*30E0*/
/*30F0*/	0x30f8,	0x30f9,	0x30fa,	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0x30fe,	0,		0,		/*30F0*/
};

static ucs4_t text_conv_WKsvm[] = {
/*		0		1		2		3		4		5		6		7		8		9		A		B		C		D		E		F		*/
/* hiragana */
/*3040*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*3040*/
/*3050*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*3050*/
/*3060*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0x3071,	/*3060*/
/*3070*/	0,		0,		0x3074,	0,		0,		0x3077,	0,		0,		0x307a,	0,		0,		0x307d,	0,		0,		0,		0,		/*3070*/
/*3080*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*3080*/
/*3090*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*3090*/
/* katakana */
/*30A0*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*30A0*/
/*30B0*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*30B0*/
/*30C0*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0x30d1,	/*30C0*/
/*30D0*/	0,		0,		0x30d4,	0,		0,		0x30d7,	0,		0,		0x30da,	0,		0,		0x30dc,	0,		0,		0,		0,		/*30D0*/
/*30E0*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*30E0*/
/*30F0*/	0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		0,		/*30F0*/
};

ucs4_t	*
edb_text_conv(EdbContext ec, ucs4_t *src)
{
	ucs4_t	*s, *d, *dst;
	int	change = FALSE;
	TextConversion	cnv;

	edb_text_conv_initialize(NULL);
	dst = (ucs4_t *)alloca(sizeof(ucs4_t)*(ucs4_strlen(src)*conv_max+1));
	d = dst;
	for(s=src; *s; ++s) {
		if(s[0]=='\\' && s[1]=='C' && s[2]=='{') {
			ucs4_t	*p = ucs4_index(&s[3], '}');
			utf8_t	*name;
			TextConversion	cnv;

			if(p) {
				*p = UCS4_NULL;
				name = UCS4toUTF8(&s[3]);
				*p = '}';
				s = p;
			} else {
				name = UCS4toUTF8(&s[3]);
				while(*s) ++s;
			}
			cnv = (TextConversion)bitree_search(conv_utf8_name_list, name, bitree_hint_str((char *)name), conv_name_compare_for_search);
			if(cnv && UCS4_STRisVALID(cnv->ucs4_d)) {
				ucs4_strncpy(d, cnv->ucs4_d, cnv->ucs4_dlen);
				d += cnv->ucs4_dlen;
			}
			continue;
		}

		/* Hiragara and Katakana region */
		if(0x3040<s[0] && s[0]<=0x30ff) {
			ucs4_t	vmc;

			if(s[1]==UCS4_WKvm && (vmc=text_conv_WKvm[s[0]-0x3040])) {
				*d++ = vmc;
				++s;
				change = TRUE;
			} else if(s[1]==UCS4_WKsvm && (vmc=text_conv_WKsvm[s[0]-0x3040])) {
				*d++ = vmc;
				++s;
				change = TRUE;
			} else {
				*d++ = s[0];
			}
			continue;
		}

		cnv = (TextConversion)bitree_search(conv_source_list, s, *s, NULL);
		if(cnv && STRisVALID(cnv->ucs4_d)) {
			ucs4_strncpy(d, cnv->ucs4_d, cnv->ucs4_dlen);
			d += cnv->ucs4_dlen;
			++cnv->count;
			if(cnv->ucs4_dlen!=1 || *s!=cnv->ucs4_d[0]) {
				if(ec) {
					edb_trace(ec, "Modify: character (0x%x) -> (0x%x)\n", *s, cnv->ucs4_d[0]);
				} else {
					fprintf(stderr, "Modify: character (0x%x) -> (0x%x)\n", *s, cnv->ucs4_d[0]);
				}
				change = TRUE;
			}
		} else {
			*d++ = *s;
		}
	}
	*d = UCS4_NULL;
#if 0
	if(change && ec->trace) {
		char	*s1, *s2;

		s1 = UCS4toUTF8(org);
		s2 = UCS4toUTF8(dst);
		edb_trace(ec, "☆ 文字列変換: 「%s」を「%s」に変換しました．\n", s1, s2);
	}
#endif
	return ucs4_strdup(dst);
}

utf8_t *
text_codename_to_html(utf8_t *name)
{
	TextConversion	cnv;

	edb_text_conv_initialize(NULL);
	cnv = (TextConversion)bitree_search(conv_utf8_name_list, name, bitree_hint_str((char *)name), conv_name_compare_for_search);
	if(cnv && STRisVALID(cnv->utf8_d)) return cnv->utf8_d;
	return (utf8_t *)"";
}

ucs4_t *
text_ucs4_codename_to_html(ucs4_t *name)
{
	TextConversion	cnv;

	edb_text_conv_initialize(NULL);
	cnv = (TextConversion)bitree_search(conv_ucs4_name_list, name, bitree_hint_ucs4str(name), conv_ucs4_name_compare_for_search);
	if(cnv && UCS4_STRisVALID(cnv->ucs4_d)) return cnv->ucs4_d;
	return ucs4_null_string;
}

utf8_t *
edb_text_codename_to_html(EdbContext ec, utf8_t *name)
{

	return text_codename_to_html(name);
}

utf8_t *
text_codename_to_tex(utf8_t *name)
{
	TextConversion	cnv;

	edb_text_conv_initialize(NULL);
	cnv = (TextConversion)bitree_search(conv_utf8_name_list, name, bitree_hint_str((char *)name), conv_name_compare_for_search);
	if(cnv && STRisVALID(cnv->utf8_tex)) return cnv->utf8_tex;
	return (utf8_t *)"";
}

ucs4_t *
text_ucs4_codename_to_tex(ucs4_t *name)
{
	TextConversion	cnv;

	edb_text_conv_initialize(NULL);
	cnv = (TextConversion)bitree_search(conv_ucs4_name_list, name, bitree_hint_ucs4str(name), conv_ucs4_name_compare_for_search);
	if(cnv && UCS4_STRisVALID(cnv->ucs4_tex)) return cnv->ucs4_tex;
	return ucs4_null_string;
}

ucs4_t *
text_ucs4_unicode_to_tex(ucs4_t c)
{
	TextConversion	cnv;
	ucs4_t	u[2];

	edb_text_conv_initialize(NULL);
	u[0] = c;
	u[1] = UCS4_NULL;
	cnv = (TextConversion)bitree_search(conv_inner_list, u, c, NULL);
	if(cnv && UCS4_STRisVALID(cnv->ucs4_tex)) return cnv->ucs4_tex;
	return ucs4_null_string;
}

utf8_t *
text_codename_to_text(utf8_t *name)
{
	TextConversion	cnv;

	edb_text_conv_initialize(NULL);
	cnv = (TextConversion)bitree_search(conv_utf8_name_list, name, bitree_hint_str((char *)name), conv_name_compare_for_search);
	if(cnv && STRisVALID(cnv->utf8_d)) return cnv->utf8_d;
	return (utf8_t *)"";
}

ucs4_t *
text_ucs4_codename_to_text(ucs4_t *name)
{
	TextConversion	cnv;

	edb_text_conv_initialize(NULL);
	cnv = (TextConversion)bitree_search(conv_ucs4_name_list, name, bitree_hint_ucs4str(name), conv_ucs4_name_compare_for_search);
	if(cnv && UCS4_STRisVALID(cnv->ucs4_d)) return cnv->ucs4_d;
	return ucs4_null_string;
}

void
edb_text_conv_show_table(EdbContext ec, FILE *fp)
{
	Link	l;

	edb_text_conv_initialize(NULL);
	fprintf(fp, "<table border=\"2\">\n");
	ForLink(l, conv_list) {
		TextConversion	cnv = (TextConversion)l->d;
		utf8_t	*s, *d;;

		fprintf(fp, "<tr>\n");
		html_text_puts3(ec, "<td>", (char *)(cnv->utf8_name?cnv->utf8_name : (utf8_t *)"　"), "</td>\n", fp);
#if 1
		s = cnv->ucs4_src?UCS4toUTF8(cnv->ucs4_src) : (utf8_t *)"　";
		d = cnv->ucs4_d?UCS4toUTF8(cnv->ucs4_d) : (utf8_t *)"　";
		html_text_puts3(ec, "<td>", (char *)s, "</td>\n", fp);
		html_text_puts3(ec, "<td>", (char *)d, "</td>\n", fp);
#endif
		fprintf(fp, "<td>%s</td>\n", STRisVALID(cnv->utf8_d)?cnv->utf8_d : (utf8_t *)"　");
		fprintf(fp, "<td>%s</td>\n", STRisVALID(cnv->utf8_tex)?cnv->utf8_tex : (utf8_t *)"　");
		fprintf(fp, "</tr>\n");
	}
	fprintf(fp, "</table>\n");
}

void
edb_text_conv_list_callback(EdbContext ec, int (*cb) __P((TextConversion, void *)), void *d)
{
	Link	l;

	edb_text_conv_initialize(NULL);
	ForLink(l, conv_list) {
		TextConversion	cnv = l->d;
		(*cb)(cnv, d);
	}
}

void
edb_text_conv_dump_table(EdbContext ec, char *prefix, FILE *fp)
{
	Link	l;

	edb_text_conv_initialize(NULL);
	ForLink(l, conv_list) {
		TextConversion	cnv = (TextConversion)l->d;
		utf8_t	*s, *d;

		fprintf(fp, "%s", prefix);
		fprintf(fp, "%s\t", cnv->utf8_name?cnv->utf8_name : (utf8_t *)"");
		s = cnv->ucs4_src?UCS4toUTF8(cnv->ucs4_src) : (utf8_t *)"";
		d = cnv->ucs4_d?UCS4toUTF8(cnv->ucs4_d) : (utf8_t *)"";
		fprintf(fp, "%s\t", s);
		fprintf(fp, "%s\t", d);
		fprintf(fp, "%s\t", STRisVALID(cnv->utf8_d)?cnv->utf8_d : (utf8_t *)"");
		fprintf(fp, "%s\t", STRisVALID(cnv->utf8_tex)?cnv->utf8_tex : (utf8_t *)"");
		fprintf(fp, "\n");
	}
}

void
edb_text_conv_dump_table_for_variant(FILE *fp)
{
	Link	l;

	edb_text_conv_initialize(NULL);
	ForLink(l, conv_list) {
		TextConversion	cnv = (TextConversion)l->d;
		utf8_t	*s, *d;

		if(cnv->ucs4_src==NULL || cnv->ucs4_src[0]==UCS4_NULL) continue;
		if(cnv->ucs4_d==NULL || cnv->ucs4_d[0]==UCS4_NULL) continue;
		if(cnv->ucs4_d[0]==cnv->ucs4_src[0]) continue;
		s = cnv->ucs4_src?UCS4toUTF8(cnv->ucs4_src) : (utf8_t *)"";
		d = cnv->ucs4_d?UCS4toUTF8(cnv->ucs4_d) : (utf8_t *)"";
		fprintf(fp, "U+%08X	# %s\n", cnv->ucs4_d[0], d);
		fprintf(fp, "U+%08X	# %s\n", cnv->ucs4_src[0], s);
		fprintf(fp, "\n");
	}
}

char *
eid2strdup(eid_t eid)
{
	char *s;

	asprintf(&s, "%d", eid);
	return s;
}

char *
eid2refstrdup(eid_t eid)
{
	char *s;

	asprintf(&s, "\\E{%d}", eid);
	return s;
}

eid_t
refstr2eid(char *str)
{
	eid_t	eid;

	if(sscanf((char *)str, "\\E{%d}", &eid)==1) return eid;
	return 0;
}

static
int
edb_index_mtime_compare(LinkObj o1, LinkObj o2)
{
	register int	t;

	t = (int)BaseMtime((EdbBaseInfo)o2->d) - BaseMtime((EdbBaseInfo)o1->d);
	if(t!=0) return t;
	return (int)BaseEID((EdbBaseInfo)o2->d) - BaseEID((EdbBaseInfo)o1->d);
}

void
edb_index_sort_bymtime(EdbContext ec, Link l_index)
{

	link_sort(l_index, edb_index_mtime_compare, NULL, NULL, NULL);
}

static struct {
	char	*utf8;
	char	*katakana_utf8;
	char	*long_vowel_utf8;
	char	*caption_utf8;
} pronounce_ordered_table[] = {
	{ "あアぁァ",		"ア",	"ア",	"あ(ぁ)",	},
	{ "いイぃィ",		"イ",	"イ",	"い(ぃ)",	},
	{ "うウぅゥヴ",		"ウ",	"ウ",	"う(ぅ)",	},
	{ "えエぇェ",		"エ",	"エ",	"え(ぇ)",	},
	{ "おオぉォ",		"オ",	"オ",	"お(ぉ)",	},
	{ "かカがガヵヶ",	"カ",	"ア",	"か(が)",	},
	{ "きキぎギ",		"キ",	"イ",	"き(ぎ)",	},
	{ "くクぐグ",		"ク",	"ウ",	"く(ぐ)",	},
	{ "けケげゲ",		"ケ",	"エ",	"け(げ)",	},
	{ "こコごゴ",		"コ",	"オ",	"こ(ご)",	},
	{ "さサざザ",		"サ",	"ア",	"さ(ざ)",	},
	{ "しシじジ",		"シ",	"イ",	"し(じ)",	},
	{ "すスずズ",		"ス",	"ウ",	"す(ず)",	},
	{ "せセぜゼ",		"セ",	"エ",	"せ(ぜ)",	},
	{ "そソぞゾ",		"ソ",	"オ",	"そ(ぞ)",	},
	{ "たタだダ",		"タ",	"ア",	"た(だ)",	},
	{ "ちチぢヂ",		"チ",	"イ",	"ち(ぢ)",	},
	{ "つツづヅっッ",	"ツ",	"ウ",	"つ(づ)",	},
	{ "てテでデ",		"テ",	"エ",	"て(で)",	},
	{ "とトどド",		"ト",	"オ",	"と(ど)",	},
	{ "なナ",		"ナ",	"ア",	"な",	},
	{ "にニ",		"ニ",	"イ",	"に",	},
	{ "ぬヌ",		"ヌ",	"ウ",	"ぬ",	},
	{ "ねネ",		"ネ",	"エ",	"ね",	},
	{ "のノ",		"ノ",	"オ",	"の",	},
	{ "はハばバぱパ",	"ハ",	"ア",	"は(ば,ぱ,ヴァ)",	},
	{ "ひヒびビぴピ",	"ヒ",	"イ",	"ひ(び,ぴ,ヴィ)",	},
	{ "ふフぶブぷプ",	"フ",	"ウ",	"ふ(ぶ,ぷ,ヴ)",	},
	{ "へヘべベぺペ",	"ヘ",	"エ",	"へ(べ,ぺ,ヴェ)",	},
	{ "ほホぼボぽポ",	"ホ",	"オ",	"ほ(ぼ,ぽ,ヴォ)",	},
	{ "まマ",		"マ",	"ア",	"ま",	},
	{ "みミ",		"ミ",	"イ",	"み",	},
	{ "むム",		"ム",	"ウ",	"む",	},
	{ "めメ",		"メ",	"エ",	"め",	},
	{ "もモ",		"モ",	"オ",	"も",	},
	{ "やヤゃャ",		"ヤ",	"ア",	"や(ゃ)",	},
	{ "ゆユゅュ",		"ユ",	"ウ",	"ゆ(ゅ)",	},
	{ "よヨょョ",		"ヨ",	"オ",	"よ(ょ)",	},
	{ "らラ",		"ラ",	"ア",	"ら",	},
	{ "りリ",		"リ",	"イ",	"り",	},
	{ "るル",		"ル",	"ウ",	"る",	},
	{ "れレ",		"レ",	"エ",	"れ",	},
	{ "ろロ",		"ロ",	"オ",	"ろ",	},
	{ "わワゎヮ",		"ワ",	"ア",	"わ(ゎ)",	},
	{ "ゐヰ",		"ヰ",	"イ",	"ゐ",	},
	{ "ゑヱ",		"ヱ",	"エ",	"ゑ",	},
	{ "をヲ",		"ヲ",	"オ",	"を",	},
	{ "んン",		"ン",	"ー",	"ん",	},
	{ "ーー",		"ー",	"ー",	"ー",	}
};

static	ucs4_t	pronounce_30xx[256];
static	ucs4_t	pronounce_30xx_long_vowel[256];
static	ucs4_t	*pronounce_30xx_caption[256];
static	int	pronounce_30xx_initialized = FALSE;

static
void
pronounce_make_table_30xx()
{
	register int	i, no;

	if(pronounce_30xx_initialized) return;
	for(i=0; i<Ns(pronounce_30xx); ++i) {
		pronounce_30xx[i] = i;
		pronounce_30xx_caption[i] = NULL;
	}
	for(i=0; i<Ns(pronounce_30xx_long_vowel); ++i) {
		pronounce_30xx_long_vowel[i] = UCS4_WLONG_VOWEL;
	}

	for(no=0; no<Ns(pronounce_ordered_table); ++no) {
		register ucs4_t	dstcode, long_vowel_code, *w, *ustr, *caption;

		dstcode = *UTF8toUCS4((utf8_t *)pronounce_ordered_table[no].katakana_utf8);
		long_vowel_code = *UTF8toUCS4((utf8_t *)pronounce_ordered_table[no].long_vowel_utf8);
		ustr = UTF8toUCS4((utf8_t *)pronounce_ordered_table[no].utf8);
		caption = ucs4_strdup(UTF8toUCS4((utf8_t *)pronounce_ordered_table[no].caption_utf8));
		for(w=ustr; *w; ++w) {
			if(0x3000<=(*w) && (*w)<(0x3000 + Ns(pronounce_30xx))) {
				pronounce_30xx[(*w)-0x3000] = dstcode;
				pronounce_30xx_long_vowel[(*w)-0x3000] = long_vowel_code;
				pronounce_30xx_caption[(*w)-0x3000] = caption;
			}
		}
	}
	pronounce_30xx_initialized = TRUE;
}

/*#pronounce_convert_order--- convert pronounce ordered form */
void
pronounce_convert_order(ucs4_t *ustr)
{
	register ucs4_t	*w, *wd, wc, wp;

	pronounce_make_table_30xx();
	wp = UCS4_NULL;
	for(w=wd=ustr; (wc = *wd = *w); ++w, ++wd) {
		/* ヴァヴィヴヴェヴォ */
		if(wc == UCS4_WKVU) {
 			switch(w[1]) {
			case UCS4_WKxa:
				wc = UCS4_WKBA;
				++w;
				break;
			case UCS4_WKxi:
				wc = UCS4_WKBI;
				++w;
				break;
			case UCS4_WKxu:
				wc = UCS4_WKBU;
				++w;
				break;
			case UCS4_WKxe:
				wc = UCS4_WKBE;
				++w;
				break;
			case UCS4_WKxo:
				wc = UCS4_WKBO;
				++w;
				break;
			default:
				wc = UCS4_WKBU;
				break;
			}
		}
		if(wc == UCS4_WLONG_VOWEL && (0x3000<=wp && wp<(0x3000 + Ns(pronounce_30xx_long_vowel)))) {
			*wd = pronounce_30xx_long_vowel[wp-0x3000];
		} else if(0x3000<=wc && wc<(0x3000 + Ns(pronounce_30xx))) {
			*wd = pronounce_30xx[wc-0x3000];
		}
		wp = wc;
	}
}

#if defined(_EDB_MALLOC_DEBUG_)
static void *mdebug[50000];
static void *mdebug0;

void
malloc_debug_alloc()
{
	int	i;

	if(!mdebug0) mdebug0 = malloc(10000000);
	for(i=0; i<(sizeof(mdebug)/sizeof(void *)); ++i) {
		if(!mdebug[i]) mdebug[i] = malloc(16);
	}
}

void
malloc_debug_free()
{
	int	i;

	for(i=0; i<(sizeof(mdebug)/sizeof(void *)); ++i) {
		if(mdebug[i]) free(mdebug[i]);
		mdebug[i] = NULL;
	}
	if(mdebug0) free(mdebug0);
	mdebug0 = NULL;
}

void *
edb_memory_malloc(size_t siz, char *file, int line)
{
	register void *p;

	p = malloc(siz);
	fprintf(stderr, "\nMemDebug %08x A %s (%d).", p, file, line);
	return p;
}

void
edb_memory_free(void *p, char *file, int line)
{

	fprintf(stderr, "\nMemDebug %08x F %s (%d).", p, file, line);
	free(p);
}

#endif

#if defined(TEXT_VARIANT_CONVERSION)
static int	edb_text_variant_table_initialized = FALSE;
static Doc doc_for_vc = NULL;

static	BiTree	bt_variant = NULL;
static	BiTree	bt_variant_list = NULL;

#define	DEFAULT_VARIANT_TABLE	EDBDIR "/libdata/tbl/unicode-variant.tbl"

static
ucs4_t
U2unicode(utf8_t *str)
{
	ucs4_t	code = UCS4_NULL;

	str_oneline((char *)str);
	if(str[0]!='U' || str[1]!='+') return code;
	sscanf((char *)&str[2], "%X", &code);
	return code;
}

void
edb_text_variant_initialize(char *name)
{
	FILE	*fp;
	char	*line;

	if(edb_text_variant_table_initialized) return;

	if(doc_for_vc==NULL) doc_for_vc = doc_sopen(DocML_PLAIN);

	if(name==NULL) name = DEFAULT_VARIANT_TABLE;
	fp = fopen(name, "r");
	if(fp==NULL) {
		fprintf(stderr, "variant_init: ERROR : '%s' is not opened.\n", name);
		return;
	}

	while((line=file_getline(fp))) {
		utf8_t	*s;
		Link	ll;
		ucs4_t	src, dst;
		UnicodeVariant	variant;

		s = (utf8_t *)strchr(line, '#');
		if(s) *s = '\0';
		if(strlen(line)==0) continue;
		ll = str_split(line, '\t');
		if(link_atoms(ll) < 2) {
			fprintf(stderr, "variant_init: ERROR : '%s'\n", line);
		} else {
			Link	lv;
			src = U2unicode((utf8_t *)ll->n->d);
			dst = U2unicode((utf8_t *)ll->d);
			variant = (UnicodeVariant)malloc(sizeof(struct UNICODE_VARIANT));
			variant->src = src;
			variant->dst = dst;
			BiTreeAdd(bt_variant, variant, variant->src, NULL, NULL, NULL);
			lv = bitree_search(bt_variant_list, NULL, variant->dst, NULL);
			if(lv==NULL) {
				lv = link_atom_new(variant, NULL);
				BiTreeAdd(bt_variant_list, lv, variant->dst, NULL, NULL, NULL);
			} else {
				LinkAppend(lv, link_atom_new(variant, NULL));
			}
		}
		link_free(ll);
		FREE(line);
	}
	fclose(fp);
	BiTreeRearrange(bt_variant);
	BiTreeRearrange(bt_variant_list);
	edb_text_variant_table_initialized = TRUE;
}

static
int
dump_variant_table_object(BiTree_obj_t obj, void *d)
{
	UnicodeVariant v = (UnicodeVariant)obj;
	FILE	*fp = (FILE *)d;

	fprintf(fp, "U+%X\tU+%X\n", v->src, v->dst);
	return TRUE;
}

void
edb_text_variant_dump(FILE *fp)
{

	edb_text_variant_initialize(NULL);
	bitree_callback(bt_variant, dump_variant_table_object, fp);
}

void
edb_text_variant_list_callback(
	int (*cb) __P((BiTree_obj_t, void *)),
	Doc doc
) {

	edb_text_variant_initialize(NULL);
	bitree_callback(bt_variant_list, cb, doc);
}

utf8_t *
edb_text_variant_conversion(utf8_t *src, int to_plain)
{
	ucs4_t	*u, *ustr;
	utf8_t	*dst;

	if(!STRisValid(src)) return (utf8_t *)strdup("");

	edb_text_variant_initialize(NULL);
	if(to_plain) {
		doc_text_puts(doc_for_vc, (char *)src);
		ustr = doc_ucs4_sgets(doc_for_vc);
		if(ustr==NULL) return (utf8_t *)strdup("");
		ucs4_str_oneline(ustr);
	} else {
		ustr = UTF8toUCS4(src);
	}
	for(u=ustr; *u; ++u) {
		UnicodeVariant v = (UnicodeVariant)bitree_search(bt_variant, NULL, *u, NULL);
		if(v && v->dst!=UCS4_NULL) *u = v->dst;
	}
	dst = UCS4toUTF8(ustr);
	if(to_plain) FREE(ustr);
	return (utf8_t *)strdup((char *)dst);
}

/*#edb_text_kana_variant_conversion--- convert pronounce ordered form */
ucs4_t
edb_text_kana_variant_conversion(ucs4_t src)
{

	pronounce_make_table_30xx();
	if(0x3000<=src && src<(0x3000 + Ns(pronounce_30xx))) {
		return pronounce_30xx[src-0x3000];
	} else {
		return src;
	}
}
#endif	/* TEXT_VARIANT_CONVERSION */

