/*
$Id: pki.c,v 1.17 2011/04/18 09:45:28 alex Exp $
	pki.c--- PKI utility
*/

#ifndef lint
static char * rcsid __attribute__((unused)) = "$Id: pki.c,v 1.17 2011/04/18 09:45:28 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<sys/types.h>
#include	<sys/file.h>
#include	<ctype.h>
#include	<string.h>

#include	<openssl/x509_vfy.h>
#include	<openssl/x509v3.h>
#include	<openssl/err.h>
#include	<openssl/pem.h>
#include	<openssl/opensslv.h>

#include	"edb/edb.h"
#include	"edb/pki.h"
#include	"edb/base64.h"

#define	EDB_PKI_PERSON_CN_PREFIX				"S"
#define	EDB_PKI_PERSONIFICATION_CN_PREFIX		"S"
#define	EDB_PKI_PERSON_CN_FMT			EDB_PKI_PERSON_CN_PREFIX "%d"
#define	EDB_PKI_PERSONIFICATION_CN_FMT	EDB_PKI_PERSONIFICATION_CN_PREFIX "%d"

#if defined(EDB_MANAGE_PKI)

struct CA_CONTEXT {
	char	*root_cn;
	char	*ssl_config;
	char	*newkeys_dir;
	char	*newreqs_dir;
	char	*newcerts_dir;
	char	*ca_cert;
	char	*ca_pass;
	char	*root_crt;
	char	*root_crt_dir;
	char	*root_crl;
	char	*root_crl_dir;
	char	*cert_serial_file;
	char	*cert_index_file;
	int	rsa_bits;
	int	cert_days;
	char	*cert_ext_tag_user;
	char	*cert_ext_tag_client;
	char	*cert_ext_tag_server;
	char	*cert_ext_file;
};

#if 0
#define	G1_SSL_CONFIG		EDB_PKI_CA_DIR "/openssl.cnf"
#define	G1_NEWKEYS_DIR		EDB_PKI_CA_DIR "/newkeys"
#define	G1_NEWREQS_DIR		EDB_PKI_CA_DIR "/newreqs"
#define	G1_NEWCERTS_DIR		EDB_PKI_CA_DIR "/newcerts"
#define	G1_CA_PASS			EDB_PKI_CA_DIR "/private/pass"
#define	G1_CERT_SERIAL_FILE	EDB_PKI_CA_DIR "/serial"			// point to next certificate serial.
#endif

typedef	struct CA_CONTEXT	*CACtxt;

static struct CA_CONTEXT ca_ctxt_1st = {
	EDB_PKI_ROOT_CN,
	EDB_PKI_CA_DIR "/openssl.cnf",
	EDB_PKI_CA_DIR "/newkeys",
	EDB_PKI_CA_DIR "/newreqs",
	EDB_PKI_CA_DIR "/newcerts",
	EDB_PKI_CA_CRT,
	EDB_PKI_CA_DIR "/private/pass",
	EDB_PKI_ROOT_CRT,
	EDB_PKI_ROOT_CRT_DIR,
	EDB_PKI_ROOT_CRL,
	EDB_PKI_ROOT_CRL_DIR,
	EDB_PKI_CA_DIR "/serial",			// point to next certificate serial.
	EDB_PKI_CA_CERT_INDEX,
	EDB_PKI_RSA_BITs,
	EDB_PKI_CERTIFICATE_DAYs,
	EDB_PKI_CERTIFICATE_EXT_TAG_USER,
	EDB_PKI_CERTIFICATE_EXT_TAG_CLIENT,
	EDB_PKI_CERTIFICATE_EXT_TAG_SERVER,
	EDB_PKI_CERTIFICATE_EXT_FILE,
};

static struct CA_CONTEXT ca_ctxt_2nd = {
	EDB_PKI2_ROOT_CN,
	EDB_PKI2_CA_DIR "/openssl.cnf",
	EDB_PKI2_CA_DIR "/newkeys",
	EDB_PKI2_CA_DIR "/newreqs",
	EDB_PKI2_CA_DIR "/newcerts",
	EDB_PKI2_CA_CRT,
	EDB_PKI2_CA_DIR "/private/pass",
	EDB_PKI2_ROOT_CRT,
	EDB_PKI2_ROOT_CRT_DIR,
	EDB_PKI2_ROOT_CRL,
	EDB_PKI2_ROOT_CRL_DIR,
	EDB_PKI2_CA_DIR "/serial",		// point to next certificate serial.
	EDB_PKI2_CA_CERT_INDEX,
	EDB_PKI2_RSA_BITs,
	EDB_PKI2_CERTIFICATE_DAYs,
	EDB_PKI2_CERTIFICATE_EXT_TAG_USER,
	EDB_PKI2_CERTIFICATE_EXT_TAG_CLIENT,
	EDB_PKI2_CERTIFICATE_EXT_TAG_SERVER,
	EDB_PKI2_CERTIFICATE_EXT_FILE,
};


#endif

#if defined(EDB_MANAGE_PKI)
int
edb_pki_IamCA(EdbContext ec)
{

	// Check both 1st and 2nd generation.
	return (access(EDB_PKI_CA_LOCKFILE, R_OK|W_OK)==0) && (access(EDB_PKI2_CA_LOCKFILE, R_OK|W_OK)==0);
}

static	int	ca_lock_fd = -1;
static	int	ca2_lock_fd = -1;

// lock/unlock both 1st and 2nd CA.
int
edb_pki_CA_lock(EdbContext ec)
{

	if(ca_lock_fd>=0 || ca2_lock_fd>=0) return FALSE;

	ca_lock_fd = open(EDB_PKI_CA_LOCKFILE, O_RDWR);
	if(ca_lock_fd<0) return FALSE;
	if(flock(ca_lock_fd, LOCK_EX)) return FALSE;

	ca2_lock_fd = open(EDB_PKI2_CA_LOCKFILE, O_RDWR);
	if(ca2_lock_fd<0) return FALSE;
	if(flock(ca2_lock_fd, LOCK_EX)) return FALSE;

	return TRUE;
}

int
edb_pki_CA_unlock(EdbContext ec)
{

	if(ca2_lock_fd<0) return FALSE;
	if(flock(ca2_lock_fd, LOCK_UN)) return FALSE;
	close(ca2_lock_fd);
	ca2_lock_fd = -1;

	if(ca_lock_fd<0) return FALSE;
	if(flock(ca_lock_fd, LOCK_UN)) return FALSE;
	close(ca_lock_fd);
	ca_lock_fd = -1;
	return TRUE;
}

static
int
edb_pki_ctxt_CA_save_certificate_request(EdbContext ec, CACtxt ca_ctxt, utf8_t *serial, utf8_t *content, int length)
{
	FILE	*fp, *pfp;
	char	*name, *cmd;
	Link	l_req;

	//
	// save certificate request.
	//
	asprintf(&name, "%s/%s.pem", ca_ctxt->newreqs_dir, serial);
	fp = fopen(name, "w");
	if(fp==NULL) return FALSE;
	if(fwrite(content, sizeof(utf8_t), length, fp)!=length) return FALSE;
	fclose(fp);

	//
	// check certificate request.
	//
	asprintf(&cmd, EDB_PKI_OPENSSL_PATH " req -in %s/%s.pem", ca_ctxt->newreqs_dir, serial);
	pfp = popen(cmd, "r");
	FREE(cmd);
	l_req = NULL;
	while(!feof(pfp)) {
		utf8_t *s = (utf8_t *)file_getline(pfp);

		if(s==NULL) break;
		LinkAppend(l_req, link_atom_new(s, free));
	}
	if(pclose(pfp)!=0 || l_req==NULL) return FALSE;
	link_free(l_req);
	return TRUE;
}

int
edb_pki_CA_save_certificate_request(EdbContext ec, int generation, utf8_t *serial, utf8_t *content, int length)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	return edb_pki_ctxt_CA_save_certificate_request(ec, &ca_ctxt_1st, serial, content, length);
	case EDB_PKI2_GENERATION:	return edb_pki_ctxt_CA_save_certificate_request(ec, &ca_ctxt_2nd, serial, content, length);
	default:					return FALSE;
	}
}

static
int
edb_pki_ctxt_CA_make_certificate_request(EdbContext ec, CACtxt ca_ctxt, utf8_t *serial, utf8_t *passphrase)
{
	FILE	*pfp, *tfp;
	char	*cmd;

	tfp = tmpfile();
	fputs((char *)passphrase, tfp);
	rewind(tfp);
	asprintf(&cmd,
		EDB_PKI_OPENSSL_PATH " req -config %s"
		" -batch -new"
		" -newkey rsa:%d -days %d"
		" -keyout %s/%s.pem -keyform PEM"
		" -passout fd:%d"
		" -out %s/%s.pem -outform PEM > /dev/null",

		ca_ctxt->ssl_config,
		ca_ctxt->rsa_bits, ca_ctxt->cert_days,
		ca_ctxt->newkeys_dir, serial,
		fileno(tfp),
		ca_ctxt->newreqs_dir, serial);
	pfp = popen(cmd, "r");
	FREE(cmd);
	if(pclose(pfp)!=0) {
		fclose(tfp);
		return FALSE;
	}
	fclose(tfp);
	return TRUE;
}

int
edb_pki_CA_make_certificate_request(EdbContext ec, int generation, utf8_t *serial, utf8_t *passphrase)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	return edb_pki_ctxt_CA_make_certificate_request(ec, &ca_ctxt_1st, serial, passphrase);
	case EDB_PKI2_GENERATION:	return edb_pki_ctxt_CA_make_certificate_request(ec, &ca_ctxt_2nd, serial, passphrase);
	default:					return FALSE;
	}
}

static
utf8_t *
edb_pki_ctxt_CA_make_certificate(EdbContext ec, CACtxt ca_ctxt, utf8_t *serial, EdbPkiCertificateInfo req, int ext)
{
	FILE	*pfp;
	char	*cmd;
	utf8_t *certificate;
	Link	l_cert;
	char	*ext_tag = "";
	char	*ext_file = ca_ctxt->cert_ext_file;

	switch(ext) {
	case EDB_PKI_CERTIFICATE_MODE_USER:	ext_tag = ca_ctxt->cert_ext_tag_user;	break;
	case EDB_PKI_CERTIFICATE_MODE_CLIENT:	ext_tag = ca_ctxt->cert_ext_tag_client;	break;
	case EDB_PKI_CERTIFICATE_MODE_SERVER:	ext_tag = ca_ctxt->cert_ext_tag_server;	break;
	}

	{
		Link l_lines = file_load_as_linked_str(ca_ctxt->cert_ext_file, NULL);
		char *s;
		asprintf(&s, "%s/%s.ext", ca_ctxt->newcerts_dir, serial);
		ext_file = STRDUP(s);	FREE(s);
		FILE *fp = fopen(ext_file, "w");
		if(fp) {
			Link	l;
			ForLink(l, l_lines) fprintf(fp, "%s\n", (char *)l->d);
			switch(ext) {
			case EDB_PKI_CERTIFICATE_MODE_USER:	break;
			case EDB_PKI_CERTIFICATE_MODE_CLIENT:	break;
			case EDB_PKI_CERTIFICATE_MODE_SERVER:
				fprintf(fp, "DNS.1=%s\n", req->S.DN.CN);
				break;
			}
			fclose(fp);
		}
		link_free(l_lines);
	}

	asprintf(&cmd,
		EDB_PKI_OPENSSL_PATH " ca -config %s"
		" -batch"
		" -passin file:%s"
		" -policy policy_anything"
		" -in %s/%s.pem"
		" -subj '%s'"
		" -days %d"
		" -extensions %s -extfile %s> /dev/null",
		ca_ctxt->ssl_config,
		ca_ctxt->ca_pass,
		ca_ctxt->newreqs_dir, serial,
		(char *)req->S.DirName,
		ca_ctxt->cert_days,
		ext_tag, ext_file);
	pfp = popen(cmd, "r");
	FREE(cmd);
	if(pclose(pfp)!=0) return NULL;

	asprintf(&cmd, EDB_PKI_OPENSSL_PATH " x509 -in %s/%s.pem", ca_ctxt->newcerts_dir, serial);
	l_cert = NULL;
	pfp = popen(cmd, "r");
	FREE(cmd);
	while(!feof(pfp)) {
		utf8_t	*s = (utf8_t *)file_getline(pfp);

		if(s==NULL) break;
		LinkAppend(l_cert, link_atom_new(s, free));
	}
	if(pclose(pfp)!=0) return NULL;

	certificate = (utf8_t *)str_concatenate(l_cert, '\n');
	link_free(l_cert);
	return certificate;
}

utf8_t *
edb_pki_CA_make_certificate(EdbContext ec, int generation, utf8_t *serial, EdbPkiCertificateInfo req, int ext)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	return edb_pki_ctxt_CA_make_certificate(ec, &ca_ctxt_1st, serial, req, ext);
	case EDB_PKI2_GENERATION:	return edb_pki_ctxt_CA_make_certificate(ec, &ca_ctxt_2nd, serial, req, ext);
	default:					return NULL;
	}
}

static
void
edb_pki_ctxt_CA_output_PKCS12(EdbContext ec, CACtxt ca_ctxt, utf8_t *serial, utf8_t *passphrase, FILE *fp)
{
	FILE	*tfp1, *tfp2;
	char	*cmd;

	tfp1 = tmpfile();	fputs((char *)passphrase, tfp1);	rewind(tfp1);
	tfp2 = tmpfile();	fputs((char *)passphrase, tfp2);	rewind(tfp2);
	asprintf(&cmd, EDB_PKI_OPENSSL_PATH " pkcs12 -export"
		" -in %s/%s.pem"
		" -inkey %s/%s.pem"
		" -passin fd:%d"
		" -passout fd:%d"
		" -certfile %s",

		ca_ctxt->newcerts_dir, serial,
		ca_ctxt->newkeys_dir, serial,
		fileno(tfp1),
		fileno(tfp2), 
		ca_ctxt->ca_cert);
	system(cmd);
	FREE(cmd);
	fclose(tfp1);
	fclose(tfp2);
}

void
edb_pki_CA_output_PKCS12(EdbContext ec, int generation, utf8_t *serial, utf8_t *passphrase, FILE *fp)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	edb_pki_ctxt_CA_output_PKCS12(ec, &ca_ctxt_1st, serial, passphrase, fp);	break;
	case EDB_PKI2_GENERATION:	edb_pki_ctxt_CA_output_PKCS12(ec, &ca_ctxt_2nd, serial, passphrase, fp);	break;
	default:					break;
	}
}

static
void
edb_pki_ctxt_CA_unlink_keyfile(EdbContext ec, CACtxt ca_ctxt, utf8_t *serial)
{
	char	*name;

	asprintf(&name, "%s/%s.pem", ca_ctxt->newkeys_dir, serial);
	unlink(name);
	FREE(name);
}

void
edb_pki_CA_unlink_keyfile(EdbContext ec, int generation, utf8_t *serial)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	edb_pki_ctxt_CA_unlink_keyfile(ec, &ca_ctxt_1st, serial);	break;
	case EDB_PKI2_GENERATION:	edb_pki_ctxt_CA_unlink_keyfile(ec, &ca_ctxt_2nd, serial);	break;
	default:					break;
	}
}

static
int
edb_pki_ctxt_CA_revoke_certificate(EdbContext ec, CACtxt ca_ctxt, utf8_t *serial)
{
	FILE	*pfp;
	char	*cmd;

	if(!STRisValid(serial)) return TRUE;
	asprintf(&cmd, EDB_PKI_OPENSSL_PATH " ca -config %s"
		" -batch"
		" -revoke %s/%s.pem"
		" -passin file:%s > /dev/null",

		ca_ctxt->ssl_config,
		ca_ctxt->newcerts_dir, serial,
		ca_ctxt->ca_pass);
	pfp = popen(cmd, "r");
	FREE(cmd);
	return (pclose(pfp)==0);
}

int
edb_pki_CA_revoke_certificate(EdbContext ec, int generation, utf8_t *serial)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	return edb_pki_ctxt_CA_revoke_certificate(ec, &ca_ctxt_1st, serial);
	case EDB_PKI2_GENERATION:	return edb_pki_ctxt_CA_revoke_certificate(ec, &ca_ctxt_2nd, serial);
	default:					return TRUE;
	}
}

static
char *
edb_pki_ctxt_CA_certificate_serial(EdbContext ec, CACtxt ca_ctxt)
{
	FILE	*fp;
	char	*s;

	fp = fopen(ca_ctxt->cert_serial_file, "r");
	if(fp==NULL) return NULL;
	s = file_getline(fp);
	fclose(fp);
	if(STRisValid(s)) {
		str_oneline(s);
	} else {
		FREE(s);
	}
	return s;
}

char *
edb_pki_CA_certificate_serial(EdbContext ec, int generation)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	return edb_pki_ctxt_CA_certificate_serial(ec, &ca_ctxt_1st);
	case EDB_PKI2_GENERATION:	return edb_pki_ctxt_CA_certificate_serial(ec, &ca_ctxt_2nd);
	default:					return NULL;
	}
}

static
void
certificate_index_free(struct EDB_PKI_CERT_INDEX	*idx)
{

	if(idx==NULL) return;
	FREE(idx->type);
	FREE(idx->expire_date);
	FREE(idx->revoke_date);
	FREE(idx->serial);
	FREE(idx->file);
	FREE(idx->DN);
	FREE(idx);
}

static
Link
edb_pki_ctxt_CA_certificate_index(EdbContext ec, CACtxt ca_ctxt)
{
	Link	l_index = NULL, l_lines, l;

	l_lines = file_load_as_linked_str(ca_ctxt->cert_index_file, NULL);
	if(l_lines==NULL) return NULL;
	ForLink(l, l_lines) {
		Link	l_field;
		struct EDB_PKI_CERT_INDEX	*idx;

		l_field = str_split(l->d, '\t');
		if(link_atoms(l_field)<6) continue;
		idx = (struct EDB_PKI_CERT_INDEX *)malloc(sizeof(struct EDB_PKI_CERT_INDEX));
		if(idx==NULL) continue;
		idx->type = (utf8_t *)strdup(link_get_obj(l_field, 0));
		idx->expire_date = (utf8_t *)strdup(link_get_obj(l_field, 1));
		idx->revoke_date = (utf8_t *)strdup(link_get_obj(l_field, 2));
		idx->serial = (utf8_t *)strdup(link_get_obj(l_field, 3));
		idx->file = (utf8_t *)strdup(link_get_obj(l_field, 4));
		idx->DN = (utf8_t *)strdup(link_get_obj(l_field, 5));
		LinkAppend(l_index, link_atom_new(idx, certificate_index_free));
	}
	link_free(l_lines);
	return l_index;
}

Link
edb_pki_CA_certificate_index(EdbContext ec, int generation)
{

	switch(generation) {
	case EDB_PKI_GENERATION:	return edb_pki_ctxt_CA_certificate_index(ec, &ca_ctxt_1st);
	case EDB_PKI2_GENERATION:	return edb_pki_ctxt_CA_certificate_index(ec, &ca_ctxt_2nd);
	default:					return NULL;
	}
}
#endif

static
utf8_t *
certificate_to_pem(utf8_t *certificate)
{
	utf8_t	*pem;
	utf8_t	*s;

	if(!STRisValid(certificate)) return NULL;
	pem = (utf8_t *)STRDUP((char *)certificate);
	str_oneline((char *)pem);

	if((s=(utf8_t *)strstr((char *)pem, "----- "))) s[5] = '\n';
	while(s && (s=(utf8_t *)strchr((char *)s, ' '))) {
		*s='\n';
		if(s[1]=='-') break;
	}
	if(!STRisValid(pem)) return NULL;
	return (utf8_t *)strdup((char *)pem);
}

utf8_t *
edb_pki_certificate_to_pem(utf8_t *certificate)
{

	return certificate_to_pem(certificate);
}

int
edb_pki_print_certificate_S_DN(EdbContext ec, utf8_t *certificate, Doc doc)
{
	X509		*cert;
	BIO		*in, *out;
	utf8_t	*pem;
	char		buf[BUFSIZ];

	pem = (utf8_t *)certificate_to_pem(certificate);
	if(!STRisValid(pem)) return FALSE;

	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	in = BIO_new_mem_buf(pem, -1);
	BIO_set_close(in, BIO_NOCLOSE);
	if(!(cert=PEM_read_bio_X509(in, NULL, NULL, NULL))) {
		edb_trace(ec, "Error parsing certificate\n");
		if(ec->trace) ERR_print_errors_fp(ec->trace);
		return FALSE;
	}
	BIO_free(in);
	FREE(pem);

	out = BIO_new(BIO_s_mem());
	X509_print_ex(out, cert, 0, ~X509_FLAG_NO_SUBJECT);
	while(BIO_gets(out, buf, sizeof(buf))) doc_text_puts(doc, buf);
	BIO_free(out);
	X509_free(cert);
	return TRUE;
}

int
edb_pki_print_certificate_ex(EdbContext ec, utf8_t *certificate, Doc doc)
{
	X509		*cert;
	BIO		*in, *out;
	utf8_t	*pem;
	char		buf[BUFSIZ];

	pem = (utf8_t *)certificate_to_pem(certificate);
	if(!STRisValid(pem)) return FALSE;

	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	in = BIO_new_mem_buf(pem, -1);
	BIO_set_close(in, BIO_NOCLOSE);
	if(!(cert=PEM_read_bio_X509(in, NULL, NULL, NULL))) {
		edb_trace(ec, "Error parsing certificate\n");
		if(ec->trace) ERR_print_errors_fp(ec->trace);
		return FALSE;
	}
	BIO_free(in);
	FREE(pem);

	out = BIO_new(BIO_s_mem());
	X509_print_ex(out, cert, 0, 0);
	while(BIO_gets(out, buf, sizeof(buf))) doc_text_puts(doc, buf);
	BIO_free(out);
	X509_free(cert);
	return TRUE;
}

int
edb_pki_print_certificate_info(EdbContext ec, EdbPkiCertificateInfo info, Doc doc)
{

	doc_puts(doc, "<table border=\"0\" align=\"center\" style=\"font-size:smaller\"><tr valign=\"top\">\n");
	doc_puts(doc, "<td>\n");
	doc_puts(doc, "<u>Management:</u><br />\n");
	doc_printf(doc, "Version = \"<span style=\"font-weight:bold\">%d (0x%x)</span>\"<br />\n", info->M.version+1, info->M.version);
	doc_printf(doc, "Serial = \"<span style=\"font-weight:bold\">%d (0x%x)</span>\"\n", info->M.serial, info->M.serial);
	doc_puts(doc, "</td>\n");
	doc_puts(doc, "<td>\n");
	doc_puts(doc, "<u>Issuer:</u><br />\n");
	doc_text_puts3(doc, "countryName (C) = \"<span style=\"font-weight:bold\">", (char *)info->I.DN.C, "</span>\"<br />\n");
	doc_text_puts3(doc, "stateOrProvinceName (ST) = \"<span style=\"font-weight:bold\">", (char *)info->I.DN.ST, "</span>\"<br />\n");
	doc_text_puts3(doc, "localityName (L) = \"<span style=\"font-weight:bold\">", (char *)info->I.DN.L, "</span>\"<br />\n");
	doc_text_puts3(doc, "organizationName (O) = \"<span style=\"font-weight:bold\">", (char *)info->I.DN.O, "</span>\"<br />\n");
	doc_text_puts3(doc, "organizationalUnitName (OU) = \"<span style=\"font-weight:bold\">", (char *)info->I.DN.OU, "</span>\"<br />\n");
	doc_text_puts3(doc, "commonName (CN) = \"<span style=\"font-weight:bold\">", (char *)info->I.DN.CN, "</span>\"<br />\n");
	doc_text_puts3(doc, "emalAddress = \"<span style=\"font-weight:bold\">", (char *)info->I.DN.Email, "</span>\"\n");
	doc_puts(doc, "</td>\n");
	doc_puts(doc, "<td>\n");
	doc_puts(doc, "<u>Subject:</u><br />\n");
	doc_text_puts3(doc, "countryName (C) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.C, "</span>\"<br />\n");
	doc_text_puts3(doc, "stateOrProvinceName (ST) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.ST, "</span>\"<br />\n");
	doc_text_puts3(doc, "localityName (L) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.L, "</span>\"<br />\n");
	doc_text_puts3(doc, "organizationName (O) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.O, "</span>\"<br />\n");
	doc_text_puts3(doc, "organizationalUnitName (OU) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.OU, "</span>\"<br />\n");
	doc_text_puts3(doc, "commonName (CN) = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.CN, "</span>\"<br />\n");
	doc_text_puts3(doc, "emalAddress = \"<span style=\"font-weight:bold\">", (char *)info->S.DN.Email, "</span>\"\n");
	doc_puts(doc, "</td>\n");
	doc_puts(doc, "</tr></table>\n");
	return TRUE;
}

int
edb_pki_print_certificate(EdbContext ec, utf8_t *certificate, Doc doc)
{
	X509		*cert;
	BIO		*in, *out;
	utf8_t	*pem;
	char		buf[BUFSIZ];

	pem = (utf8_t *)certificate_to_pem(certificate);
	if(!STRisValid(pem)) return FALSE;

	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	in = BIO_new_mem_buf(pem, -1);
	BIO_set_close(in, BIO_NOCLOSE);
	if(!(cert=PEM_read_bio_X509(in, NULL, NULL, NULL))) {
		edb_trace(ec, "Error parsing certificate\n");
		if(ec->trace) ERR_print_errors_fp(ec->trace);
		return FALSE;
	}
	BIO_free(in);
	FREE(pem);

	out = BIO_new(BIO_s_mem());
	PEM_write_bio_X509(out, cert);
	while(BIO_gets(out, buf, sizeof(buf))) doc_text_puts(doc, buf);
	BIO_free(out);
	X509_free(cert);
	return TRUE;
}

int
edb_pki_check_certificate(EdbContext ec, utf8_t *certificate)
{
	X509		*cert;
	BIO		*in;
	utf8_t		*pem;

	pem = (utf8_t *)certificate_to_pem(certificate);
	if(!STRisValid(pem)) return FALSE;

	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	in = BIO_new_mem_buf(pem, -1);
	BIO_set_close(in, BIO_NOCLOSE);
	if(!(cert=PEM_read_bio_X509(in, NULL, NULL, NULL))) {
		edb_trace(ec, "Error parsing certificate\n");
		if(ec->trace) ERR_print_errors_fp(ec->trace);
		return FALSE;
	}
	BIO_free(in);
	FREE(pem);
	X509_free(cert);
	return TRUE;
}

EdbColumnInfo
edb_pki_tuple_certifiable(EdbContext ec, EdbTupleInfo tuple)
{
	Link	l;

	if(tuple == NULL) return NULL;
	ForLink(l, TupleTable(tuple)->l_column) {
		EdbColumnInfo	ci = l->d;

		if(ci->typeInfo->typeID == EDB_TYPE_CERTIFICATE) return ci;
	}
	return NULL;
}

#if defined(EDB_MANAGE_PKI)

EdbPkiCertificateInfo
edb_pki_tuple_certificate_request_info(EdbContext ec, EdbTupleInfo tuple)
{
	EdbPkiCertificateInfo	info;
	void	* (*request_func) EdbPkiTupleCertificateRequestInfo_P;
	utf8_t	*cn;
	int	mode;

	if(tuple==NULL) return NULL;

	if(edb_table_module_load(ec, TupleTable(tuple)) && (request_func = TupleTable(tuple)->tuple_certificate_request_info)) {
		return (*request_func)(ec, tuple);
	}

	if(TupleIsPerson(tuple)) {
		asprintf((char **)&cn, EDB_PKI_PERSON_CN_FMT, TupleEID(tuple));
		mode = EDB_PKI_CERTIFICATE_MODE_USER;
	} else if(TupleIsPersonification(tuple)) {
		asprintf((char **)&cn, EDB_PKI_PERSONIFICATION_CN_FMT, TupleEID(tuple));
		mode = EDB_PKI_CERTIFICATE_MODE_USER;
	} else {
		edb_trace(ec, "edb_pki_tuple_certificate_request_info: unsupported info.\n");
		return NULL;
	}

	info = (struct EDB_PKI_CERTIFICATE_INFO *)malloc(sizeof(struct EDB_PKI_CERTIFICATE_INFO));
	bzero(info, sizeof(struct EDB_PKI_CERTIFICATE_INFO));

	info->S.DN.C = (utf8_t *)strdup(EDB_PKI_DN_C);
	info->S.DN.ST = (utf8_t *)strdup(EDB_PKI_DN_ST);
	info->S.DN.L = (utf8_t *)strdup(EDB_PKI_DN_L);
	info->S.DN.O = (utf8_t *)strdup(EDB_PKI_DN_O);
	info->S.DN.OU = (utf8_t *)strdup(EDB_PKI_DN_OU);
	info->S.DN.CN = cn;
	info->S.DN.Email = (utf8_t *)edb_person_email(ec, tuple);
	info->mode = mode;
	asprintf((char **)&info->S.DirName,
		"/C=%s/ST=%s/L=%s/O=%s/OU=%s/CN=%s%s%s",
		EDB_PKI_DN_C,
		EDB_PKI_DN_ST,
		EDB_PKI_DN_L,
		EDB_PKI_DN_O,
		EDB_PKI_DN_OU,
		info->S.DN.CN,
		STRisValid(info->S.DN.Email) ? "/emailAddress=":"",
		STRisValid(info->S.DN.Email) ? info->S.DN.Email:(utf8_t *)"");
	return info;
}
#endif

utf8_t	*
edb_pki_tuple_certificate(EdbContext ec, EdbTupleInfo tuple)
{
	EdbColumnInfo ci;
	EdbTupleColumn tc;
	EdbDatum	datum;

	if(tuple==NULL) return NULL;
	ci = edb_pki_tuple_certifiable(ec, tuple);
	if(ci==NULL) return NULL;
	tc = edb_tuple_seek_column(ec, tuple, ColumnXN(ci));
	if(tc!=NULL && (datum=TCDatumFirst(tc)) && DatumEnglishIsValid(datum)) return (utf8_t *)strdup((char *)DatumEnglish(datum));
	return NULL;
}

static
utf8_t *
get_text_by_NID(EdbContext ec, X509_NAME *name, int nid)
{
	char	buf[BUFSIZ];

	if(X509_NAME_get_text_by_NID(name, nid, buf, sizeof(buf))>0) {
		buf[sizeof(buf)-1] = '\0';
		return (utf8_t *)strdup(buf);
	}
	return NULL;
}

void
edb_pki_certificate_info_free(EdbContext ec, EdbPkiCertificateInfo info)
{

	if(info==NULL) return;
	FREE(info->I.DirName);
	FREE(info->I.DN.C);
	FREE(info->I.DN.ST);
	FREE(info->I.DN.L);
	FREE(info->I.DN.O);
	FREE(info->I.DN.OU);
	FREE(info->I.DN.CN);
	FREE(info->I.DN.Email);
	FREE(info->S.DirName);
	FREE(info->S.DN.C);
	FREE(info->S.DN.ST);
	FREE(info->S.DN.L);
	FREE(info->S.DN.O);
	FREE(info->S.DN.OU);
	FREE(info->S.DN.CN);
	FREE(info->S.DN.Email);

	if(info->l_subjectAltNames) link_free(info->l_subjectAltNames);
	info->l_subjectAltNames = NULL;

	FREE(info);
}

EdbPkiCertificateInfo
edb_pki_certificate_info(EdbContext ec, utf8_t *certificate)
{
	EdbPkiCertificateInfo	info;
	X509		*cert;
	X509_NAME		*issuer, *subject;
	BIO		*in;
	utf8_t		*pem;
	char		buf[BUFSIZ];

	pem = (utf8_t *)certificate_to_pem(certificate);
	if(!STRisValid(pem)) {
		edb_trace(ec, "Error parsing certificate\n");
		return NULL;
	}

	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	in = BIO_new_mem_buf(pem, -1);
	BIO_set_close(in, BIO_NOCLOSE);
	cert=PEM_read_bio_X509(in, NULL, NULL, NULL);
	BIO_free(in);
	FREE(pem);
	if(cert==NULL) {
		edb_trace(ec, "Error parsing certificate\n");
		if(ec->trace) ERR_print_errors_fp(ec->trace);
		return NULL;
	}

	info = (struct EDB_PKI_CERTIFICATE_INFO *)malloc(sizeof(struct EDB_PKI_CERTIFICATE_INFO));
	bzero(info, sizeof(struct EDB_PKI_CERTIFICATE_INFO));

	info->M.version = X509_get_version(cert);
	info->M.serial = ASN1_INTEGER_get(X509_get_serialNumber(cert));
	issuer = X509_get_issuer_name(cert);
	if(issuer) {
		info->I.DN.C = get_text_by_NID(ec, issuer, NID_countryName);
		info->I.DN.ST = get_text_by_NID(ec, issuer, NID_stateOrProvinceName);
		info->I.DN.L = get_text_by_NID(ec, issuer, NID_localityName);
		info->I.DN.O = get_text_by_NID(ec, issuer, NID_organizationName);
		info->I.DN.OU = get_text_by_NID(ec, issuer, NID_organizationalUnitName);
		info->I.DN.CN = get_text_by_NID(ec, issuer, NID_commonName);
		info->I.DN.Email = get_text_by_NID(ec, issuer, NID_pkcs9_emailAddress);
#if 1
		X509_NAME_oneline(issuer, buf, sizeof(buf));
		if(STRisValid(buf)) info->I.DirName = (utf8_t *)strdup(buf);
#endif
	}
	subject = X509_get_subject_name(cert);
	if(subject) {
		info->S.DN.C = get_text_by_NID(ec, subject, NID_countryName);
		info->S.DN.ST = get_text_by_NID(ec, subject, NID_stateOrProvinceName);
		info->S.DN.L = get_text_by_NID(ec, subject, NID_localityName);
		info->S.DN.O = get_text_by_NID(ec, subject, NID_organizationName);
		info->S.DN.OU = get_text_by_NID(ec, subject, NID_organizationalUnitName);
		info->S.DN.CN = get_text_by_NID(ec, subject, NID_commonName);
		info->S.DN.Email = get_text_by_NID(ec, subject, NID_pkcs9_emailAddress);
#if 1
		X509_NAME_oneline(subject, buf, sizeof(buf));
		if(STRisValid(buf)) info->S.DirName = (utf8_t *)strdup(buf);
#endif
	}
	{
		STACK_OF(GENERAL_NAME) *san_names = NULL;
		san_names = X509_get_ext_d2i((X509 *)cert, NID_subject_alt_name, NULL, NULL);
		if(san_names) {
			int i, san_names_nb = sk_GENERAL_NAME_num(san_names);
			for(i=0; i<san_names_nb; ++i) {
				GENERAL_NAME *gname = sk_GENERAL_NAME_value(san_names, i);
				char *s = NULL;
				switch(gname->type) {
				case GEN_EMAIL:
#if OPENSSL_VERSION_NUMBER < 0x10100000L
					asprintf(&s, "email:%s", (char *)ASN1_STRING_data(gname->d.rfc822Name));
#else
					asprintf(&s, "email:%s", (char *)ASN1_STRING_get0_data(gname->d.rfc822Name));
#endif
					break;
				case GEN_DNS:
#if OPENSSL_VERSION_NUMBER < 0x10100000L
					asprintf(&s, "DNS:%s", (char *)ASN1_STRING_data(gname->d.dNSName));
#else
					asprintf(&s, "DNS:%s", (char *)ASN1_STRING_get0_data(gname->d.dNSName));
#endif
					break;
				case GEN_URI:
#if OPENSSL_VERSION_NUMBER < 0x10100000L
					asprintf(&s, "URI:%s", (char *)ASN1_STRING_data(gname->d.uniformResourceIdentifier));
#else
					asprintf(&s, "URI:%s", (char *)ASN1_STRING_get0_data(gname->d.uniformResourceIdentifier));
#endif
					break;
				}
				if(s) LinkAppend(info->l_subjectAltNames, link_atom_new(s, free));
			}
			sk_GENERAL_NAME_pop_free(san_names, GENERAL_NAME_free);
		}
	}
	X509_free(cert);
	return info;
}

EdbPkiCertificateInfo
edb_pki_SSL_env_info(EdbContext ec)
{
	char	*__env, *s;
	EdbPkiCertificateInfo	ssl;

	ssl = (struct EDB_PKI_CERTIFICATE_INFO *)malloc(sizeof(struct EDB_PKI_CERTIFICATE_INFO));
	bzero(ssl, sizeof(struct EDB_PKI_CERTIFICATE_INFO));

#define	GETENV_STRDUP(env)	(utf8_t *)((__env=getenv((env))) ? strdup(__env) : NULL)

	s = getenv("SSL_CLIENT_M_VERSION");
	if(STRisValid(s)) {
		sscanf(s, "%x", &ssl->M.version);
		--ssl->M.version;
	}
	s = getenv("SSL_CLIENT_M_SERIAL");
	if(STRisValid(s)) sscanf(s, "%x", &ssl->M.serial);

	ssl->I.DirName = GETENV_STRDUP("SSL_CLIENT_I_DN");
	ssl->I.DN.C = GETENV_STRDUP("SSL_CLIENT_I_DN_C");
	ssl->I.DN.ST = GETENV_STRDUP("SSL_CLIENT_I_DN_ST");
	ssl->I.DN.L = GETENV_STRDUP("SSL_CLIENT_I_DN_L");
	ssl->I.DN.O = GETENV_STRDUP("SSL_CLIENT_I_DN_O");
	ssl->I.DN.OU = GETENV_STRDUP("SSL_CLIENT_I_DN_OU");
	ssl->I.DN.CN = GETENV_STRDUP("SSL_CLIENT_I_DN_CN");
	ssl->I.DN.Email = GETENV_STRDUP("SSL_CLIENT_I_DN_Email");

	ssl->S.DirName = GETENV_STRDUP("SSL_CLIENT_S_DN");
	ssl->S.DN.C = GETENV_STRDUP("SSL_CLIENT_S_DN_C");
	ssl->S.DN.ST = GETENV_STRDUP("SSL_CLIENT_S_DN_ST");
	ssl->S.DN.L = GETENV_STRDUP("SSL_CLIENT_S_DN_L");
	ssl->S.DN.O = GETENV_STRDUP("SSL_CLIENT_S_DN_O");
	ssl->S.DN.OU = GETENV_STRDUP("SSL_CLIENT_S_DN_OU");
	ssl->S.DN.CN = GETENV_STRDUP("SSL_CLIENT_S_DN_CN");
	ssl->S.DN.Email = GETENV_STRDUP("SSL_CLIENT_S_DN_Email");

	return ssl;
}

int
edb_pki_certificate_info_equal(EdbContext ec, EdbPkiCertificateInfo info1, EdbPkiCertificateInfo info2)
{

#define	DIFFER(s1, s2)	if(!((!STRisValid((s1)) && !STRisValid((s2))) || (STRisValid((s1)) && STRisValid((s2)) && !strcmp((char *)(s1), (char *)(s2)))))

	if(info1->M.version != info2->M.version) return FALSE;
	if(info1->M.serial != info2->M.serial) return FALSE;

	DIFFER(info1->I.DN.C, info2->I.DN.C) return FALSE;
	DIFFER(info1->I.DN.ST, info2->I.DN.ST) return FALSE;
	DIFFER(info1->I.DN.L, info2->I.DN.L) return FALSE;
	DIFFER(info1->I.DN.O, info2->I.DN.O) return FALSE;
	DIFFER(info1->I.DN.OU, info2->I.DN.OU) return FALSE;
	DIFFER(info1->I.DN.CN, info2->I.DN.CN) return FALSE;
	DIFFER(info1->I.DN.Email, info2->I.DN.Email) return FALSE;

	DIFFER(info1->S.DN.C, info2->S.DN.C) return FALSE;
	DIFFER(info1->S.DN.ST, info2->S.DN.ST) return FALSE;
	DIFFER(info1->S.DN.L, info2->S.DN.L) return FALSE;
	DIFFER(info1->S.DN.O, info2->S.DN.O) return FALSE;
	DIFFER(info1->S.DN.OU, info2->S.DN.OU) return FALSE;
	DIFFER(info1->S.DN.CN, info2->S.DN.CN) return FALSE;
	DIFFER(info1->S.DN.Email, info2->S.DN.Email) return FALSE;
	return TRUE;
}

eid_t
edb_pki_person_CN2EID(EdbContext ec, utf8_t *cn)
{
	int	len = strlen(EDB_PKI_PERSON_CN_PREFIX);

	if(!strncmp((char *)cn, EDB_PKI_PERSON_CN_PREFIX, len) && str_isdigit((char *)&cn[len])) return ATOI((char *)&cn[len]);
	return 0;
}

eid_t
edb_pki_personification_CN2EID(EdbContext ec, utf8_t *cn)
{
	int	len = strlen(EDB_PKI_PERSONIFICATION_CN_PREFIX);

	if(!strncmp((char *)cn, EDB_PKI_PERSONIFICATION_CN_PREFIX, len) && str_isdigit((char *)&cn[len])) return ATOI((char *)&cn[len]);
	return 0;
}

utf8_t *
edb_pki_person_EID2CN(EdbContext ec, eid_t eid)
{
	utf8_t 	*cn;

	asprintf((char **)&cn, EDB_PKI_PERSON_CN_FMT, eid);
	return cn;
}

utf8_t *
edb_pki_personification_EID2CN(EdbContext ec, eid_t eid)
{
	utf8_t 	*cn;

	asprintf((char **)&cn, EDB_PKI_PERSONIFICATION_CN_FMT, eid);
	return cn;
}

#define	ENC_HEADER_RSA_OAEP	"{RSA-OAEP}"

utf8_t *
edb_pki_rsa_encrypt(EdbContext ec, utf8_t *dat, int len, char *cert)
{
	char	*certpath;
	FILE	*pfp, *fp;
	char	*cmd;
	char	*tmp, *str;
	utf8_t	*enc = NULL;
	int	fd;

	if(!STRisValid(cert)) return NULL;

	asprintf(&certpath, EDB_PKI_DIR "/secret-keeper/%s", cert);
	if(access(certpath, R_OK)!=0) {
		edb_trace(ec, "edb_pki_rsa_encrypt: cannot read certificate file.\n");
		FREE(certpath);
		return NULL;
	}
	asprintf(&tmp, "/tmp/edb.pki.%d.XXXXXXXXXX", getpid());
	fd = mkstemp(tmp);
	close(fd);

	asprintf(&cmd, EDB_PKI_OPENSSL_PATH " rsautl -encrypt -oaep"
		" -inkey %s -certin"
		" > %s",
		certpath, tmp);
	pfp = popen(cmd, "w");
	FREE(cmd);
	FREE(certpath);
	if(STRisValid(dat)) fwrite(dat, sizeof(utf8_t), (size_t)len, pfp);
	pclose(pfp);

	fp = fopen(tmp, "r");
	if(fp) {
		int	len;
		utf8_t	buf[BUFSIZ];

		len = fread(buf, sizeof(utf8_t), (size_t)BUFSIZ, fp);
		str = (char *)base64_encode((utf8_t *)buf, len);
		asprintf((char **)&enc, "%s%s", ENC_HEADER_RSA_OAEP, str);
		FREE(str);
		str_oneline((char *)enc);
		fclose(fp);
	}
	unlink(tmp);
	FREE(tmp);
	return enc;
}

int
edb_pki_rsa_decrypt(EdbContext ec, utf8_t *enc, utf8_t **dat, int *len, char *key, char *pass)
{
	char	*keypath, *passpath;
	FILE	*pfp;
	char	*cmd;
	char	*tmp;
	utf8_t	*str, *text;
	int	fd, ln, status;

	if(!STRisValid(key) || !STRisValid(pass)) return FALSE;

	if(strncmp((char *)enc, (char *)ENC_HEADER_RSA_OAEP, strlen(ENC_HEADER_RSA_OAEP))) return FALSE;

	asprintf(&keypath, EDB_PKI_DIR "/secret-keeper/%s", key);
	if(access(keypath, R_OK)!=0) {
		edb_trace(ec, "edb_pki_rsa_decrypt: cannot read key file.\n");
		FREE(keypath);
		return FALSE;
	}
	asprintf(&passpath, EDB_PKI_DIR "/secret-keeper/%s", pass);
	if(access(passpath, R_OK)!=0) {
		edb_trace(ec, "edb_pki_rsa_decrypt: cannot read password file.\n");
		FREE(keypath);
		FREE(passpath);
		return FALSE;
	}
	asprintf(&tmp, "/tmp/edb.pki.%d.XXXXXXXXXX", getpid());
	fd = mkstemp(tmp);
	ln = strlen(ENC_HEADER_RSA_OAEP);
	str = base64_decode((utf8_t *)&enc[ln], &ln);
	write(fd, str, ln);
	close(fd);
	asprintf(&cmd, EDB_PKI_OPENSSL_PATH " rsautl -decrypt -oaep"
		" -inkey %s -passin file:%s"
		" -in %s",
		keypath, passpath, tmp);
	pfp = popen(cmd, "r");
	FREE(cmd);
	FREE(keypath);
	FREE(passpath);

	text = (utf8_t *)malloc(sizeof(utf8_t)*BUFSIZ);
	text[0] = '\0';
	ln = fread(text, sizeof(utf8_t), (size_t)(BUFSIZ-1), pfp);
	text[ln] = '\0';
	pclose(pfp);
	if(STRisValid(text)) {
		*dat = (utf8_t *)text;
		*len = ln;
		status = TRUE;
	} else {
		FREE(text);
		status = FALSE;
	}
	unlink(tmp);
	FREE(tmp);
	return status;
}

static
int
verify_callback(int ok, X509_STORE_CTX *stor)
{
	int err = X509_STORE_CTX_get_error(stor);

//// commented by alex (X509_STORE_CTX is deprecated)
////	if(!ok) fprintf(stderr, "Error: %s\n", X509_verify_cert_error_string(stor->error));
////	if(!ok) fprintf(stderr, "Error: pki.c : verify_callback : something wrong\n");
	if(!ok) fprintf(stderr, "Error: %s\n", X509_verify_cert_error_string(err));
	return ok;
}

static
int
edb_pki_ctxt_verify_certificate(EdbContext ec, CACtxt ca_ctxt, utf8_t *certificate)
{
	X509		*cert;
	BIO		*in;
	utf8_t		*pem;
	X509_STORE	*store;
	X509_LOOKUP	*lookup;
	X509_STORE_CTX		*verify_ctx = NULL;
	int	status = TRUE;

	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	pem = (utf8_t *)certificate_to_pem(certificate);
	if(!STRisValid(pem)) return FALSE;

	in = BIO_new_mem_buf(pem, -1);
	BIO_set_close(in, BIO_NOCLOSE);
	if(!(cert=PEM_read_bio_X509(in, NULL, NULL, NULL))) {
		edb_trace(ec, "edb_pki_verify_certificate: Error parsing certificate\n");
		if(ec->trace) ERR_print_errors_fp(ec->trace);
		BIO_free(in);
		FREE(pem);
		return FALSE;
	}
	BIO_free(in);

	if(status && !(store=X509_STORE_new())) {
		edb_trace(ec, "edb_pki_verify_certificate: Error Creating X509_STORE_CTX object\n");
		status = FALSE;
	}
	X509_STORE_set_verify_cb_func(store, verify_callback);

	if(status && X509_STORE_load_locations(store, ca_ctxt->root_crt, ca_ctxt->root_crt_dir)!=1) {
		edb_trace(ec, "edb_pki_verify_certificate: Error Loading the CA file or directory\n");
		status = FALSE;
	}
	if(status && X509_STORE_set_default_paths(store)!=1) {
		edb_trace(ec, "edb_pki_verify_certificate: Error Loading the system-wide CA certificates\n");
		status = FALSE;
	}
	if(status && !(lookup=X509_STORE_add_lookup(store, X509_LOOKUP_file()))) {
		edb_trace(ec, "edb_pki_verify_certificate: Error X509_STORE_add_lookup\n");
		status = FALSE;
	}
	if(status && X509_load_crl_file(lookup, ca_ctxt->root_crl, X509_FILETYPE_PEM)!=1) {
		edb_trace(ec, "edb_pki_verify_certificate: Error X509_load_crl_file\n");
		status = FALSE;
	}

	X509_STORE_set_flags(store, X509_V_FLAG_CRL_CHECK | X509_V_FLAG_CRL_CHECK_ALL);

	if(status && !(verify_ctx = X509_STORE_CTX_new())) {
		edb_trace(ec, "edb_pki_verify_certificate: Error X509_STORE_CTX_new\n");
		status = FALSE;
	}
	if(status && X509_STORE_CTX_init(verify_ctx, store, cert, NULL)!=1) {
		edb_trace(ec, "edb_pki_verify_certificate: X509_STORE_CTX_init\n");
		status = FALSE;
	}

	if(status && X509_verify_cert(verify_ctx)!=1) {
#if 0
		edb_trace(ec, "edb_pki_verify_certificate: X509_verify_cert\n");
#endif
		status = FALSE;
	}

	FREE(pem);
	if(cert) X509_free(cert);
	if(store) X509_STORE_free(store);
	if(verify_ctx) X509_STORE_CTX_free(verify_ctx);
	return status;
}

int
edb_pki_certificate_generation(EdbContext ec, utf8_t *certificate)
{
	int gen = 0;

	EdbPkiCertificateInfo info = edb_pki_certificate_info(ec, certificate);
	if(info==NULL) return 0;

	if(info->I.DN.CN!=NULL) {
		if(!strcmp((char *)info->I.DN.CN, (&ca_ctxt_1st)->root_cn)) {
			gen = EDB_PKI_GENERATION;
		} else if(!strcmp((char *)info->I.DN.CN, (&ca_ctxt_2nd)->root_cn)) {
			gen = EDB_PKI2_GENERATION;
		}
	}
	edb_pki_certificate_info_free(ec, info);
	return gen;
}

int
edb_pki_verify_certificate(EdbContext ec, utf8_t *certificate)
{

	switch(edb_pki_certificate_generation(ec, certificate)) {
	case EDB_PKI_GENERATION:	return edb_pki_ctxt_verify_certificate(ec, &ca_ctxt_1st, certificate);
	case EDB_PKI2_GENERATION:	return edb_pki_ctxt_verify_certificate(ec, &ca_ctxt_2nd, certificate);
	default:					return FALSE;
	}
}

