/*
	$Id: tcdatum.c,v 1.4 2008/08/09 09:21:37 alex Exp $
	tcdatum.c--- tuple column datum manipulation
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: tcdatum.c,v 1.4 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/tcdatum.h"

void
edb_tcdatum_free(EdbContext ec, EdbTupleColumn tc)
{
	register EdbDatum	datum, n;

	if(tc==NULL) return;
	for(datum=tc->datum; datum; datum=n) {
		n = datum->n;
		edb_datum_free(ec, datum);
	}
	tc->datum = NULL;
}

int
edb_tcdatum_num(EdbContext ec, EdbTupleColumn tc)
{
	register int	num;
	register EdbDatum	datum;

	if(tc==NULL) return 0;
	num = 0;
	for(datum=tc->datum; datum; datum=datum->n) ++num;
	return num;
}

EdbDatum
edb_tcdatum_get_datum(EdbContext ec, EdbTupleColumn tc, int nth)
{
	register EdbDatum	datum;

	if(tc==NULL) return NULL;
	datum = tc->datum;
	while(nth-- > 0 && datum) datum = datum->n;
	return datum;
}

void
edb_tcdatum_sort(
	register EdbContext ec,
	register EdbTupleColumn tc,
	register int	(*compare_func)(),
	register void	* (*aux_func)(),
	register void	* aux_data,
	register void	(*aux_free)())
{
	register int	i, n;
	register LinkObj	objs;
	register EdbDatum	datum;

	if(tc==NULL) return;

	n = edb_tcdatum_num(ec, tc);
	if(n<2) return;
	objs = (LinkObj)alloca(sizeof(struct LINK_OBJ)*n);
	for(i=0, datum=tc->datum; i<n && datum; ++i, datum=datum->n) {
		objs[i].d = datum;
		objs[i].idx = i;
		objs[i].f = NULL;
	}
	if(aux_func) for(i=0; i<n; ++i) objs[i].aux = (*aux_func)(objs[i].d, aux_data);
	qsort(objs, (size_t)n, sizeof(struct LINK_OBJ), compare_func);
	tc->datum = objs[0].d;
	for(i=0; i<n ; ++i) {
		datum = objs[i].d;
		datum->n = (i+1<n) ? objs[i+1].d : NULL;
	}
	if(aux_free) for(i=0; i<n; ++i) if(objs[i].aux) (*aux_free)(objs[i].aux);
}

void
edb_tcdatum_reverse(EdbContext ec, EdbTupleColumn tc)
{
	register EdbDatum	p, datum, n;

	if(tc==NULL) return;
	p = NULL;
	for(datum=tc->datum; datum; datum=n) {
		n = datum->n;
		datum->n = p;
		p = datum;
	}
	tc->datum = p;
}

EdbDatum *
edb_tcdatum_to_array(EdbContext ec, EdbTupleColumn tc, EdbDatum data[], int *num)
{
	register EdbDatum	*d, datum;

	*num = edb_tcdatum_num(ec, tc);
	if(tc==NULL) return NULL;
	if(data==NULL) data = (EdbDatum *)malloc(sizeof(EdbDatum)*(*num));
	d = data;
	ForTCDatum(datum, tc) *d++ = datum;
	return data;
}

void
edb_tcdatum_from_array(EdbContext ec, EdbTupleColumn tc, EdbDatum data[], int num)
{
	register int	i;

	if(tc==NULL) return;
	if(num==0) {
		tc->datum = NULL;
	} else {
		tc->datum = data[0];
		for(i=0; i<num-1; ++i) data[i]->n = data[i+1];
		data[num-1]->n = NULL;
	}
}

void
edb_tcdatum_append(EdbContext ec, EdbTupleColumn tc, EdbDatum datum)
{
	register EdbDatum	d;

	if(tc==NULL) return;
	if(tc->datum==NULL) {
		tc->datum = datum;
		return;
	}
	ForTCDatum(d, tc) {
		if(d->n==NULL) {
			d->n = datum;
			break;
		}
	}
}

