/*
	$Id: text.c,v 1.18 2011/04/18 09:45:28 alex Exp $
	misc.c--- miscellaneous functions
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: text.c,v 1.18 2011/04/18 09:45:28 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"

/*
	edb_text_encode-- encode text
	edb_text_decode-- decode text
	edb_mtext_concatenate--- concatenate mtext
	edb_mtext_split--- split mtext
*/
utf8_t	*
edb_text_encode(EdbContext ec, utf8_t *src)
{
	utf8_t	*dst, *d, *s, c;

	if(src==NULL) return (utf8_t *)strdup("");
	dst = (utf8_t *)alloca(strlen((char *)src)*2+1);
	for(s=src, d=dst; (c = *s); ++s) {
		if(c=='\\' || c=='|') *d++ = '\\';
		*d++ = c;
	}
	*d = '\0';
	return (utf8_t *)strdup((char *)dst);
}

utf8_t	*
edb_text_decode(EdbContext ec, utf8_t *src)
{
	utf8_t	*dst, *d, *s, c;
	int	bs;

	if(src==NULL) return (utf8_t *)strdup("");
	dst = (utf8_t *)alloca(strlen((char *)src)*2+1);
	bs = FALSE;
	for(s=src, d=dst; (c = *s); ++s) {
		if(bs) {
			*d++ = c;
			bs = FALSE;
		} else if(c=='\\') {
			bs = TRUE;
		} else {
			*d++ = c;
		}
	}
	*d = '\0';
	return (utf8_t *)strdup((char *)dst);
}

utf8_t *
edb_mtext_concatenate(EdbContext ec, eid_t mapto, utf8_t *en, utf8_t *jp, utf8_t *pr)
{
	utf8_t	*dst;

	en = en ? edb_text_encode(ec, en) : (utf8_t *)strdup("");
	jp = jp ? edb_text_encode(ec, jp) : (utf8_t *)strdup("");
	pr = pr ? edb_text_encode(ec, pr) : (utf8_t *)strdup("");
	asprintf((char **)&dst, "%d|%s|%s|%s", mapto, en, jp, pr);
	FREE(en);
	FREE(jp);
	FREE(pr);
	return dst;
}

utf8_t *
edb_mtext_concatenate_variant(EdbContext ec, eid_t mapto, utf8_t *en, utf8_t *jp, utf8_t *pr)
{
	utf8_t	*dst, *s;

	en = edb_text_encode(ec, s = edb_text_variant_conversion(en, TRUE));	FREE(s);
	jp = edb_text_encode(ec, s = edb_text_variant_conversion(jp, TRUE));	FREE(s);
	pr = edb_text_encode(ec, s = edb_text_variant_conversion(pr, TRUE));	FREE(s);
	asprintf((char **)&dst, "%d|%s|%s|%s", mapto, en, jp, pr);
	FREE(en);
	FREE(jp);
	FREE(pr);
	return dst;
}

static
utf8_t *
edb_mtext_split_get_token(EdbContext ec, utf8_t **src)
{
	register int	bs = FALSE, len;
	register utf8_t	c, *s, *token;

	for(s=*src; (c=*s); ++s) {
		if(bs) {
			bs = FALSE;
		} else if(c=='\\') {
			bs = TRUE;
		} else if(c=='|') {
			break;
		}
	}
	len = sizeof(utf8_t)*(s - *src);
	token = (utf8_t *)malloc(len+1);
	strncpy((char *)token, (char *)*src, len);
	token[len] = '\0';
	if(*s) ++s;
	*src = s;
	return token;
}

void
edb_mtext_split(EdbContext ec, eid_t *mapto, utf8_t **en, utf8_t **jp, utf8_t **pr, utf8_t *src)
{
	utf8_t	*s;

	s = edb_mtext_split_get_token(ec, &src);
	*mapto = ATOI((char *)s);
	FREE(s);
	*en = edb_mtext_split_get_token(ec, &src);
	*jp = edb_mtext_split_get_token(ec, &src);
	*pr = edb_mtext_split_get_token(ec, &src);
}

utf8_t *
edb_text_for_pgsql(EdbContext ec, utf8_t *str)
{
	register utf8_t	c, *d, *s;
	utf8_t *dst;

	if(str==NULL) return (utf8_t *)strdup("");
	dst = (utf8_t *)alloca(sizeof(utf8_t)*(strlen((char *)str)*2+1));
	for(d = dst, s = str; (c=*s); ++s, ++d) {
		if(c=='\\') {
			*d++ = '\\';
			*d = '\\';
		} else if(c=='\'') {
			*d++ = '\'';
			*d = '\'';
		} else {
			*d = *s;
		}
	}
	*d = '\0';
	return (utf8_t *)strdup((char *)dst);
}

utf8_t *
edb_text_for_pgsql_literal(EdbContext ec, utf8_t *str)
{
	utf8_t *text, *lit;

	text = edb_text_for_pgsql(ec, str);
	asprintf((char **)&lit, "%s'%s'", (PG_SV_cmp(ec, 8, 1, 0) >= 0) ? "E":"", text);
	FREE(text);
	return lit;
}

#if	defined(TEXT_VARIANT_CONVERSION)
utf8_t *
edb_text_for_pgsql_variant(EdbContext ec, utf8_t *str)
{
	register utf8_t	c, *d, *s;
	utf8_t *dst;

	if(str==NULL) return (utf8_t *)strdup("");
	dst = (utf8_t *)alloca(sizeof(utf8_t)*(strlen((char *)str)*2+1));
	for(d = dst, s = str; (c=*s); ++s, ++d) {
		if(c=='\\') {
			*d++ = '\\';
			*d = '\\';
		} else if(c=='\'') {
			*d++ = '\'';
			*d = '\'';
		} else {
			*d = *s;
		}
	}
	*d = '\0';
	if(STRisValid(dst)) {
		return edb_text_variant_conversion(dst, TRUE);
	}
	return (utf8_t *)strdup((char *)dst);
}

utf8_t *
edb_text_for_pgsql_literal_variant(EdbContext ec, utf8_t *str)
{
	utf8_t *text, *lit;

	text = edb_text_for_pgsql_variant(ec, str);
	asprintf((char **)&lit, "%s'%s'", (PG_SV_cmp(ec, 8, 1, 0) >= 0) ? "E":"", text);
	FREE(text);
	return lit;
}

#endif	/* TEXT_VARIANT_CONVERSION */

utf8_t *
edb_text_for_pgsql_array(EdbContext ec, utf8_t *str)
{
	register utf8_t	c, *d, *s;
	utf8_t *dst;

	if(str==NULL) return (utf8_t *)strdup("");
	dst = (utf8_t *)alloca(sizeof(utf8_t)*(strlen((char *)str)*4+1));

	for(d = dst, s = str; (c=*s); ++s, ++d) {
		if(c=='\\') {
			*d++ = '\\';
			*d++ = '\\';
			*d++ = '\\';
			*d = '\\';
		} else if(c=='\'') {
			*d++ = '\'';
			*d = '\'';
		} else if(c=='"') {
			*d++ = '\\';
			*d++ = '\\';
			*d = '"';
		} else {
			*d = *s;
		}
	}
	*d = '\0';
	return (utf8_t *)strdup((char *)dst);
}

int
edb_text_get_reference(EdbContext ec, utf8_t *s, utf8_t *str)
{
	register int	len;

	len = strlen((char *)s);
	if(len>4 && !strncmp((char *)s, "\\E{", 3) && s[len-1]=='}') {
		strcpy((char *)str, (char *)&s[3]);
		str[len-4] = '\0';
		return TRUE;
	}
	return FALSE;
}

int
edb_text_parse(
	register utf8_t **sp,
	register EdbTextParseCallback callback,
	register void *context)
{
	register utf8_t	*s = *sp, sc;
	utf8_t	*e;
	register int	found;

	sc = s[0];
	if(sc=='\0') return FALSE;
	if(sc=='\\' || sc=='{' || sc=='}' || sc=='L') {
		if(callback) (*callback)(context, sc, NULL);
		*sp = ++s;
		return TRUE;
	}
	if(!(strchr("_^AMCPNDI", sc) && s[1]=='{')) {
		if(callback) (*callback)(context, sc, NULL);
		*sp = ++s;
		return FALSE;
	}
	++*sp;
	++*sp;
	e = s = (utf8_t *)STRDUP((char *)*sp);
	/* find corresponding {} */
	found = FALSE;
	while(*e) {
		if(*e=='\\') {
			++e;
			edb_text_parse(&e, NULL, NULL);
			continue;
		} else if(*e=='}') {
			*e++ = '\0';
			found = TRUE;
			break;
		}
		++e;
	}
	*sp += (int)(e-s);
	if(callback) (*callback)(context, sc, s);
	return found;
}

int
edb_text_ucs4_parse(
	register ucs4_t **sp,
	register EdbTextUCS4ParseCallback callback,
	register void *context)
{
	register ucs4_t	*s = *sp, sc;
	ucs4_t	*e;
	register int	found;

	sc = s[0];
 	if(sc==UCS4_NULL) return FALSE;
	if(sc=='\\' || sc=='{' || sc=='}' || sc=='L') {
		if(callback) (*callback)(context, sc, NULL);
		*sp = ++s;
		return TRUE;
	}
	if(!(strchr("_^AMCPNDI", sc) && s[1]=='{')) {
		if(callback) (*callback)(context, sc, NULL);
		*sp = ++s;
		return FALSE;
	}
	++*sp;
	++*sp;
	e = s = UCS4_STRDUP(*sp);
	/* find corresponding {} */
	found = FALSE;
	while(*e) {
		if(*e=='\\') {
			++e;
			edb_text_ucs4_parse(&e, NULL, NULL);
			continue;
		} else if(*e=='}') {
			*e++ = '\0';
			found = TRUE;
			break;
		}
		++e;
	}
	*sp += (int)(e-s);
	if(callback) (*callback)(context, sc, s);
	return found;
}

