/*
	$Id: tuple.c,v 1.111 2012/04/03 03:51:13 alex Exp alex $
	tuple.c--- tuple info
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: tuple.c,v 1.111 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"

EdbTupleColumn
edb_tuple_column_new(
	register EdbContext		ec,
	register EdbTupleInfo		super_tuple,
	register EdbDatum		super_datum,
	register EdbColumnInfo	ci)
{
	register EdbTupleColumn	tc;

	if(_edb_delayed_free_tuple_column) {
		++edb_delayed_free_recycle_count;
		EdbDelayedFreePop(_edb_delayed_free_tuple_column, tc);
	} else {
		tc = (EdbTupleColumn)Edb_malloc(sizeof(struct EDB_TUPLE_COLUMN));
	}
	bzero(tc, sizeof(struct EDB_TUPLE_COLUMN));
	tc->super = (ci->super) ? (void *)super_datum : (void *)super_tuple;
	tc->ci = ci;
	return tc;
}

void
edb_tuple_column_free(
	register EdbContext		ec,
	register EdbTupleColumn	tc)
{

	if(tc==NULL) return;
	edb_tcdatum_free(ec, tc);
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_tuple_column, tc);
	} else {
		Edb_free(tc);
	}
}

EdbDatum
edb_tuple_TCD_dup(
	register EdbContext		ec,
	register EdbTupleInfo		super_tuple,
	register EdbTupleColumn	super_tc,
	register EdbDatum		src)
{
	register Link	l;
	register EdbDatum	new;

	new = edb_datum_dup(ec, super_tc, src);
	ForLink(l, src->l_child) {
		LinkAppend(new->l_child, link_atom_new(edb_tuple_column_dup(ec, super_tuple, new, l->d), edb_tuple_column_free));
	}
	return new;
}

EdbTupleColumn
edb_tuple_column_dup(
	register EdbContext		ec,
	register EdbTupleInfo		super_tuple,
	register EdbDatum		super_datum,
	register EdbTupleColumn	src_tc)
{
	register EdbTupleColumn	tc;
	EdbDatum	src_datum;

	tc = edb_tuple_column_new(ec, super_tuple, super_datum, src_tc->ci);
	bcopy(src_tc, tc, sizeof(struct EDB_TUPLE_COLUMN));
	tc->super = (tc->ci->super) ? (void *)super_datum : (void *)super_tuple;
	tc->datum = NULL;
	ForTCDatum(src_datum, src_tc) {
		EdbDatum	new;

		new = edb_datum_dup(ec, tc, src_datum);
		if(src_datum->l_child) {
			Link	lc;

			ForLink(lc, src_datum->l_child) {
				LinkAppend(new->l_child, link_atom_new(edb_tuple_column_dup(ec, super_tuple, new, (EdbTupleColumn)lc->d), edb_tuple_column_free));
			}
		}
		new->n = tc->datum;
		tc->datum = new;
	}
	edb_tcdatum_reverse(ec, tc);
	return tc;
}

static
int
edb_tuple_TCD_callback_sub(
	EdbContext		ec,
	EdbTupleInfo		tuple,
	int				layer,
	Link				l_column,
	TCD_callback_func_t	callback,
	EdbTupleCBData	cbd)
{
	register Link	l;
	register int	cont = TRUE;

	ForLink(l, l_column) {
		register EdbTupleColumn	tc = (EdbTupleColumn)l->d;
		register EdbDatum	datum;

		ForTCDatum(datum, tc) {
			if(callback) {
				cont = (*callback)(ec, tuple, layer, tc, datum, cbd);
				if(!cont) break;
			}
			if(datum->l_child) {
				cont = edb_tuple_TCD_callback_sub(ec, tuple, layer+1, datum->l_child, callback, cbd);
				if(!cont) break;
			}
		}
		if(!cont) break;
	}
	return cont;
}

void
edb_tuple_TCD_callback(
	EdbContext		ec,
	EdbTupleInfo		tuple,
	EdbTupleColumn	tc,
	TCD_callback_func_t	callback,
	void				*data)
{
	register	int	cont = TRUE;
	register	int	layer = 0;
	struct EDB_TUPLE_CB_DATA	cb_data;
	EdbTupleCBData			cbd = &cb_data;

	if(tuple==NULL) return;
	bzero(cbd, sizeof(struct EDB_TUPLE_CB_DATA));
	cbd->user_data = data;
	if(tc) {
		register EdbDatum	datum;

		ForTCDatum(datum, tc) {
			if(callback) {
				cont = (*callback)(ec, tuple, layer, tc, datum, cbd);
				if(!cont) break;
			}
			if(datum->l_child) {
				cont = edb_tuple_TCD_callback_sub(ec, tuple, layer+1, datum->l_child, callback, cbd);
				if(!cont) break;
			}
		}
	} else {
		edb_tuple_TCD_callback_sub(ec, tuple, 0, tuple->l_column, callback, cbd);
	}
}

static
int
edb_tuple_TC_callback_sub(
	EdbContext		ec,
	EdbTupleInfo		tuple,
	int				layer,
	Link				l_column,
	TC_callback_func_t	callback,
	EdbTupleCBData	cbd)
{
	register Link	l;
	register int	cont = TRUE;

	ForLink(l, l_column) {
		register EdbTupleColumn	tc = (EdbTupleColumn)l->d;
		register EdbDatum	datum;

		if(callback) {
			cont = (*callback)(ec, tuple, layer, tc, cbd);
			if(!cont) break;
		}
		ForTCDatum(datum, tc) {
			if(datum->l_child) {
				cont = edb_tuple_TC_callback_sub(ec, tuple, layer+1, datum->l_child, callback, cbd);
				if(!cont) break;
			}
		}
		if(!cont) break;
	}
	return cont;
}

void
edb_tuple_TC_callback(
	EdbContext		ec,
	EdbTupleInfo		tuple,
	EdbTupleColumn	tc,
	TC_callback_func_t	callback,
	void				*data)
{
	register int	cont = TRUE;
	register int	layer = 0;
	struct EDB_TUPLE_CB_DATA	cb_data;
	EdbTupleCBData			cbd = &cb_data;

	if(tuple==NULL) return;
	bzero(cbd, sizeof(struct EDB_TUPLE_CB_DATA));
	cbd->user_data = data;
	if(tc) {
		register EdbDatum	datum;

		ForTCDatum(datum, tc) {
			if(datum->l_child) {
				cont = edb_tuple_TC_callback_sub(ec, tuple, layer+1, datum->l_child, callback, cbd);
				if(!cont) break;
			}
		}
	} else {
		edb_tuple_TC_callback_sub(ec, tuple, 0, tuple->l_column, callback, cbd);
	}
}

EdbTupleInfo
edb_tuple_new(EdbContext ec, EdbTableInfo ti)
{
	register EdbTupleInfo	tuple;

	tuple = (EdbTupleInfo)malloc(sizeof(struct EDB_TUPLE_INFO));
	if(tuple == NULL) return NULL;
	bzero(tuple, sizeof(struct EDB_TUPLE_INFO));
	edb_base_init(ec, TupleBase(tuple));

	TupleTable(tuple) = ti;

	TupleEID(tuple) = 0;
	TupleOperator(tuple) = ec->operator;;
	TupleRead(tuple) = EDB_AUTH_INHERIT;
	TupleWrite(tuple) = EDB_AUTH_INHERIT;
	TupleDelete(tuple) = EDB_AUTH_INHERIT;
	TupleOwner(tuple) = ec->operator;
	return tuple;
}

void
edb_tuple_free(EdbContext ec, EdbTupleInfo tuple)
{

	if(ec->delayed_free) {
		EdbDelayedFreePush(ec->delayed_free_tuple, tuple);
	} else {
		edb_tuple_free_exec(ec, tuple);
	}
}

void
edb_tuple_free_exec(EdbContext ec, EdbTupleInfo tuple)
{

	EdbLinkFree2(ec, tuple->l_column);
	EdbLinkFree(tuple->expand_map);
	EdbLinkFree(tuple->l_required_cxn);
	EdbLinkFree(tuple->l_preferred_cxn);
	EdbLinkFree(tuple->l_optional_cxn);
	EdbLinkFree(tuple->l_unused_cxn);
	if(tuple->ca_expand) edb_catalogue_free(tuple->ca_expand);
	FREE(tuple);
}

static
Link
edb_tuple_datum_parse_xml(EdbContext ec, EdbTupleInfo tuple, EdbDatum super_datum, Link l_column, xmlDocPtr doc, xmlNodePtr cur)
{
	EdbTableInfo	ti = TupleTable(tuple);
	Link	l, l_tc = NULL;

	ForLink(l, l_column) {
		EdbColumnInfo		ci = l->d;
		EdbTupleColumn	tc;
		Link	ln, l_node;
		int	elements, xmlelements = edb_table_column_xml_elements(ec, ti, ci->super, ci);

		elements = 0;
		l_node = xml_find_element_node_list(ec, doc, cur, ColumnXN(ci), NULL);
		if(l_node==NULL) continue;
		tc = edb_tuple_column_new(ec, tuple, super_datum, ci);
		if(tc==NULL) {
			edb_trace(ec, "edb_tuple_datum_parse_xml: can't allocate TC.\n");
			EdbLinkFree(l_node);
			continue;
		}
		tc->datum = NULL;
		ForLink(ln, l_node) {
			EdbDatum	datum = NULL;
			int	ex = FALSE;

			if(ci->typeInfo) {
				datum = edb_datum_parse_xml_node(ec, tc, ColumnCType(ci), doc, (xmlNodePtr)ln->d);
			} else {
				continue;
			}
			if(datum==NULL) continue;
			if(!edb_datum_is_empty(ec, datum)) ex = TRUE;
			if(ci->l_child) {
				datum->l_child = edb_tuple_datum_parse_xml(ec, tuple, datum, ci->l_child, doc, ((xmlNodePtr)ln->d)->xmlChildrenNode);
				if(datum->l_child) ex = TRUE;
			}
			if(ex) {
				datum->n = tc->datum;
				tc->datum = datum;
				++elements;
			} else {
				edb_datum_free(ec, datum);
			}
			if(xmlelements>0 && elements>=xmlelements) {
				if(ln->n) edb_trace(ec, "edb_tuple_datum_parse_xml: EID(%d) EOID(%d) (%s) has excess datum.\n", TupleEID(tuple), TupleEOID(tuple), ColumnXN(ci));
				break;
			}
		}
		EdbLinkFree(l_node);
		if(tc->datum==NULL && !tc->nothing) {
			edb_tuple_column_free(ec, tc);
			continue;
		}
		edb_tcdatum_reverse(ec, tc);
		LinkAppend(l_tc, link_atom_new(tc, edb_tuple_column_free));
		if(ci->chronological && edb_tcdatum_num(ec, tc)>1 && ci->elements!=0) {
			if(ec->chronological_order) {
				edb_tcdatum_sort(ec, tc, edb_datum_period_compare_for_sort_ascend, NULL, NULL, NULL);
			} else {
				edb_tcdatum_sort(ec, tc, edb_datum_period_compare_for_sort_descend, NULL, NULL, NULL);
			}
		}
	}
	return l_tc;
}

EdbTupleInfo
edb_tuple_parse_xml(EdbContext ec, EdbTableInfo ti, utf8_t *xmlstr)
{
	xmlDocPtr	doc;
	xmlNodePtr	root, base;
	EdbTupleInfo	tuple;

	doc = xmlParseMemory((char *)xmlstr, xmlStrlen(xmlstr));
	if(doc==NULL) return NULL;

	root = xmlDocGetRootElement(doc);

	tuple = edb_tuple_new(ec, ti);

	base = xml_find_element_node(ec, doc, root, TABLE_XN(ti), XML_EN_BASE, NULL);
	edb_base_parse_xml(ec, TupleBase(tuple), doc, base);

	tuple->l_column = edb_tuple_datum_parse_xml(ec, tuple, NULL, ti->l_column, doc, root->xmlChildrenNode);

	xmlFreeDoc(doc);
	return tuple;
}

utf8_t *
edb_tuple_datum_make_xml(EdbContext ec, EdbTupleInfo tuple, Link l_column, int completion)
{
//	EdbTableInfo	ti = TupleTable(tuple);
	Link	l, l_xml;
	utf8_t	*xml = NULL;

	l_xml = NULL;
	ForLink(l, l_column) {
		EdbTupleColumn	tc = (EdbTupleColumn)l->d;
		EdbColumnInfo		ci = tc->ci;
		EdbDatum	datum;

		if(!completion && tc->mapped) continue;
		ForTCDatum(datum, tc) {
			utf8_t	*datum_xml=NULL, *child_xml=NULL, *m_attr = NULL, *p_attr = NULL;
			Caption	cap;

			xml = NULL;
			if(datum->l_child) child_xml = edb_tuple_datum_make_xml(ec, tuple, datum->l_child, completion);
			if(completion && DatumEIDisValid(datum) && ci->typeInfo->multilingual && (cap = edb_make_mapped_caption(ec, ci, DatumEID(datum), CAP_norecurse))) {
				EdbDatum	d = edb_datum_dup(ec, DatumSuper(datum), datum);

				edb_datum_set(ec, d, DatumEID(d), cap->c_main_english, cap->c_main_japanese, cap->c_main_pronounce);
				if(ci->typeInfo) datum_xml = edb_datum_make_xml(ec, ColumnCType(ci), d, &m_attr, &p_attr, completion);
				edb_datum_free(ec, d);
				edb_caption_free(ec, cap);
			} else {
				if(ci->typeInfo) datum_xml = edb_datum_make_xml(ec, ColumnCType(ci), datum, &m_attr, &p_attr, completion);
			}

			if(child_xml==NULL && datum_xml==NULL) {
				if(m_attr) asprintf((char **)&xml, "<%s:%s %s%s%s/>", EDB_XMLNS, ColumnXN(ci), m_attr, p_attr?" ":"", p_attr ? p_attr : (utf8_t *)"");
			} else {
				asprintf((char **)&xml, "<%s:%s%s%s%s%s>%s%s</%s:%s>", EDB_XMLNS, ColumnXN(ci),
					m_attr?" ":"", m_attr ? m_attr : (utf8_t *)"",
					p_attr?" ":"", p_attr ? p_attr : (utf8_t *)"",
					datum_xml ? datum_xml : (utf8_t *)"",
					child_xml ? child_xml : (utf8_t *)"",
					EDB_XMLNS, ColumnXN(ci));
			}
			FREE(child_xml);
			FREE(datum_xml);
			FREE(m_attr);
			FREE(p_attr);
			if(xml) LinkAppend(l_xml, link_atom_new(xml, free));
		}
		if(TCDatumFirst(tc)==NULL && TCDatumNothing(tc)) {
			asprintf((char **)&xml, "<%s:%s %s/>", EDB_XMLNS, ColumnXN(ci), XML_AN_ASSERT "=\"" XML_AV_ASSERT_NOTHING "\"");
			if(xml) LinkAppend(l_xml, link_atom_new(xml, free));
		}
	}
	xml = (utf8_t *)str_concatenate0(l_xml);
	EdbLinkFree(l_xml);
	if(STRisVALID(xml)) return xml;
	FREE(xml);
	return NULL;
}

utf8_t *
edb_tuple_make_xml(EdbContext ec, EdbTupleInfo tuple, int completion)
{
	register utf8_t	*base_xml, *body_xml;
	utf8_t	*xml;

	base_xml = edb_base_make_xml(ec, TupleBase(tuple), FALSE);
	body_xml = edb_tuple_datum_make_xml(ec, tuple, tuple->l_column, completion);

	asprintf((char **)&xml, "<?xml version=\"%s\" encoding=\"utf-8\"?><!DOCTYPE %s:%s SYSTEM \"%s%s.dtd\"><%s:%s xmlns:%s=\"%s\">%s%s</%s:%s>",
		XML_DEFAULT_VERSION,
		EDB_XMLNS, TABLE_XN(TupleTable(tuple)), EDB_DTD_URI, TABLE_XN(TupleTable(tuple)),
		EDB_XMLNS, TABLE_XN(TupleTable(tuple)), EDB_XMLNS, EDB_XMLNS_URI,
		base_xml, body_xml ? body_xml : (utf8_t *)"",
		EDB_XMLNS, TABLE_XN(TupleTable(tuple)));
	FREE(base_xml);
	FREE(body_xml);
	return xml;
}

EdbTupleInfo
edb_tuple_duplicate(EdbContext ec, EdbTupleInfo tuple)
{
	EdbTupleInfo	new = NULL;
	utf8_t	*xml;

	xml = edb_tuple_make_xml(ec, tuple, FALSE);
	if(STRisVALID(xml)) new = edb_tuple_parse_xml(ec, TupleTable(tuple), xml);
	FREE(xml);
	return new;
}

EdbTupleInfo
edb_tuple_get_XMLDB(EdbContext ec, EdbTableInfo ti, eid_t eid, eoid_t eoid)
{
	EdbTupleInfo	tuple = NULL;
	EdbXmldbInfo	xi = edb_xmldb_get(ec, eid, eoid);

	if(xi==NULL) return NULL;
	if(ti==NULL) ti = edb_table_get(ec, XmldbTEID(xi));
	if(ti) tuple = edb_tuple_parse_xml(ec, ti, XmldbXML(xi));
	edb_xmldb_free(ec, xi);
	return tuple;
}

eid_t
edb_tuple_get_teid_XMLDB(EdbContext ec, eid_t eid, eoid_t eoid)
{
	utf8_t *query=NULL, *select;
	utf8_t *latest = NULL;
	register eid_t	teid;
	utf8_t	*text;

#ifdef EDB_READER
	if(eoid) {
		asprintf((char **)&query, "SELECT " EDB_XMLDB_COLUMN_TEID " FROM " EDB_XMLDB_TABLE
			" WHERE " EDB_XMLDB_COLUMN_EOID "=%d", eoid);
	} else {
		asprintf((char **)&query, "SELECT " EDB_XMLDB_COLUMN_TEID " FROM " EDB_XMLDB_TABLE
			" WHERE " EDB_XMLDB_COLUMN_EID "=%d", eid);
	}
#else /* !EDB_READER */
	select = (utf8_t *)"SELECT " EDB_XMLDB_COLUMN_TEID " FROM " EDB_XMLDB_TABLE " AS t1 WHERE ";
	if(eoid) {
		asprintf((char **)&query, "%s t1.%s=%d", select, EDB_XMLDB_COLUMN_EOID, eoid);
	} else {
		latest = edb_create_latest_phrase(ec, (utf8_t *)EDB_XMLDB_TABLE, (utf8_t *)"t1", (utf8_t *)"t2", NULL, eid);
		asprintf((char **)&query, "%s t1.%s=%d AND %s", select, EDB_XMLDB_COLUMN_EID, eid, latest);
		FREE(latest);
	}
#endif /* EDB_READER */
#if 1
	teid = -1;
	text = edb_pgsql_query_one_text(ec, query);
	FREE(query);
	if(STRisVALID(text)) teid = ATOI((char *)text);
	FREE(text);
	return teid;
#else
	teid = edb_pgsql_query_int(ec, query);
	FREE(query);
	return teid;
#endif
}

Link
edb_tuple_get_XMLDBs(EdbContext ec, EdbTableInfo ti)
{
	utf8_t *query;
	utf8_t *latest;
	Link	l, l_xml, l_tuple;
	int	n;

#ifdef EDB_READER
	asprintf((char **)&query, "SELECT " EDB_XMLDB_COLUMN_XML " FROM " EDB_XMLDB_TABLE
		" WHERE " EDB_XMLDB_COLUMN_TEID "=%d", TABLE_EID(ti));
#else /* !EDB_READER */
	if(edb_reldb_is_usable(ec)) {
		asprintf((char **)&query, "SELECT t1." EDB_XMLDB_COLUMN_XML
			" FROM " EDB_XMLDB_TABLE " AS t1, %s AS t2"
			" WHERE t1." EDB_XMLDB_COLUMN_EOID "=t2." EDB_BASE_EOID,
			TABLE_SN(ti));
	} else {
		latest = edb_create_latest_phrase(ec, (utf8_t *)EDB_XMLDB_TABLE, (utf8_t *)"t1", (utf8_t *)"t2", NULL, 0);
		asprintf((char **)&query, "SELECT t1." EDB_XMLDB_COLUMN_XML
			" FROM " EDB_XMLDB_TABLE " AS t1"
			" WHERE t1." EDB_XMLDB_COLUMN_TEID "=%d AND %s",
			TABLE_EID(ti), latest);
		FREE(latest);
	}
#endif /* EDB_READER */
	l_xml = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_xml) {
		EdbTupleInfo	tuple;

		tuple = edb_tuple_parse_xml(ec, ti, (utf8_t *)l->d);
		LinkAppendTop(link_atom_new(tuple, edb_tuple_free), l_tuple);
	}
	LinkReverse(l_tuple);
	EdbLinkFree(l_xml);
	return l_tuple;
}

#define	PERIOD_OVERLAP(f1, t1, f2, t2)	(((f1) <= (t2)) && ((f2) <= (t1)))
static
int
edb_tuple_check_datum_period_overlap(EdbContext ec, EdbDatum data[], int n, int th, int acc, int f, int t)
{
	register int	i;

	for(i=0; i<n; ++i) {
		register int	f1, t1;

		if( !data[i] ) continue;
		f1 = DatumDateFrom(data[i]);
		t1 = DatumDateTo(data[i]);
		if(f1<=t1 && PERIOD_OVERLAP(f, t, f1, t1)) {
			register EdbDatum	save;

			if(acc+1>th) return TRUE;
			save = data[i];
			data[i] = NULL;
			if(edb_tuple_check_datum_period_overlap(ec, data, n, th, acc+1, max(f, f1), min(t, t1))) return TRUE;
			data[i] = save;
		}
	}
	return FALSE;
}

struct TCD_CHECK {
	int	status;
};

static
int
edb_tuple_check_datum(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbDatum datum, EdbTupleCBData cbd)
{
	struct TCD_CHECK	*tcd = (struct TCD_CHECK *)cbd->user_data;
	register eid_t	mapto;

	if(datum==NULL) return TRUE;
	mapto = DatumEID(datum);
	if(EIDisVALID(mapto)) {
		register eid_t	teid;
		register Link	lm;

		if(TupleAvail(tuple) && !edb_eid_is_available(ec, mapto)) {
			edb_trace(ec, "edb_tuple_check_datum: Not exist EID : Tuple(%d).%s.mapto=%d.\n", TupleEID(tuple), ColumnXN(tc->ci), mapto);
			tcd->status = FALSE;
		}
		teid = edb_eid_isin_table(ec, mapto);
		ForLink(lm, tc->ci->l_maplookup) {
			EdbMapLookup	map = (EdbMapLookup)lm->d;

			if(teid == edb_maplookup_table_eid(ec, map)) break;
		}
		if(lm==NULL) {
			edb_trace(ec, "edb_tuple_check_datum: Tuple(%d).%s.mapto=%d is in unspecified table(%s).\n", TupleEID(tuple), ColumnXN(tc->ci), mapto, edb_table_get(ec, teid)->xmlname);
			tcd->status = FALSE;
		}
	} else if(mapto != 0) {
		edb_trace(ec, "edb_tuple_check_datum: Invalid EID : Tuple(%d).%s.mapto=%d.\n", TupleEID(tuple), ColumnXN(tc->ci), mapto);
		tcd->status = FALSE;
	}
	if(!edb_datum_check(ec, ColumnCType(tc->ci), datum)) {
		edb_trace(ec, "edb_tuple_check_datum: Invalid data : Tuple(%d).%s.\n", TupleEID(tuple), ColumnXN(tc->ci));
		tcd->status = FALSE;
	}

	if(tc->ci->caption == EDB_CAPTION_SHORT && DatumRead(datum)!=EDB_AUTH_INHERIT) {
		edb_trace(ec, "edb_tuple_check_datum: WARNING : read permission of short caption is restricted: Tuple(%d).%s.\n", TupleEID(tuple), ColumnXN(tc->ci));
	}
#if 1
	if(tc->ci->chronological) {
		if(tc->ci->elements==0) {
		} else if(tc->ci->elements==1) {
			EdbDatum	d;

			ForTCDatum(d, tc) {
				if(datum==d) continue;
				if(edb_datum_period_overlapped(ec, datum, d)) {
					edb_trace(ec, "edb_tuple_check_datum: period overlapped : Tuple(%d).%s.\n", TupleEID(tuple), ColumnXN(tc->ci));
					tcd->status = FALSE;
				}
			}
		} else {	/* tc->ci->elements > 0 */
			int	n;

			n = edb_tcdatum_num(ec, tc);
			if(n<=tc->ci->elements) {
			} else {
				EdbDatum	d, *da, *data = (EdbDatum *)alloca(sizeof(EdbDatum)*n);

				da = data;
				ForTCDatum(d, tc) *da++ = d;
				if(edb_tuple_check_datum_period_overlap(ec, data, n, tc->ci->elements, 0, EDB_DATE_MIN, EDB_DATE_MAX)) {
					edb_trace(ec, "edb_tuple_check_datum: period overlapped : Tuple(%d).%s.\n", TupleEID(tuple), ColumnXN(tc->ci));
					tcd->status = FALSE;
				}
			}
		}
	}
#endif
	return TRUE;
}

int
edb_tuple_check(EdbContext ec, EdbTupleInfo tuple)
{
	struct TCD_CHECK	tcd;

	tcd.status = TRUE;
	if(TupleAvail(tuple)) {
		register eid_t	owner = TupleOwner(tuple);
		register eid_t	mapto = TupleMapto(tuple);
		register eid_t	teid;

		if(EIDisVALID(owner)) {
			if(!edb_eid_is_available(ec, owner)) {
				edb_trace(ec, "edb_tuple_check: Unavailable Owner : Tuple(%d).owner=%d.\n", TupleEID(tuple), owner);
				tcd.status = FALSE;
			} else {
				teid = edb_eid_isin_table(ec, owner);
				if(teid==0 || !(teid == EDB_TABLE_PERSON_EID || teid == EDB_TABLE_PERSONIFICATION_EID)) {
					edb_trace(ec, "edb_tuple_check: Invalid Owner : Tuple(%d).owner=%d is not person or personification.\n", TupleEID(tuple), owner);
				}
			}
		} else if(owner != ec->masterUID) {
			edb_trace(ec, "edb_tuple_check: Invalid Owner : Tuple(%d).owner=%d.\n", TupleEID(tuple), owner);
		}
		if(EIDisVALID(TupleMapto(tuple)) && !edb_eid_is_available(ec, TupleMapto(tuple))) {
			edb_trace(ec, "edb_tuple_check: Not exist EID : Tuple(%d).mapto=%d.\n", TupleEID(tuple), TupleMapto(tuple));
			tcd.status = FALSE;
		}
		if(EIDisVALID(mapto)) {
			register Link	lm;

			teid = edb_eid_isin_table(ec, mapto);
			ForLink(lm, TupleTable(tuple)->l_maplookup) {
				EdbMapLookup	map = (EdbMapLookup)lm->d;

				if(teid == edb_maplookup_table_eid(ec, map)) break;
			}
			if(lm==NULL) {
				edb_trace(ec, "edb_tuple_check: Tuple(%d).mapto=%d is in unspecified table(%s).\n", TupleEID(tuple), mapto, edb_table_get(ec, teid)->xmlname);
				tcd.status = FALSE;
			}
		} else if(mapto != 0) {
			edb_trace(ec, "edb_tuple_check: Tuple(%d).mapto=%d is invalid.\n", TupleEID(tuple), mapto);
		}
	}
	edb_tuple_TCD_callback(ec, tuple, NULL, edb_tuple_check_datum, &tcd);
	return tcd.status;
}

#if defined(EDB_READER)
#else	/* EDB_READER */
static
int
edb_tuple_check_updatable_column(EdbContext ec,
		EdbTupleInfo tuple, EdbColumnInfo ci,
		EdbTupleInfo otuple, EdbTupleColumn otc,
		EdbTupleInfo ntuple, EdbTupleColumn ntc)
{
	int onum = otc ? edb_tcdatum_num(ec, otc) : 0;
	int nnum = ntc ? edb_tcdatum_num(ec, ntc) : 0;

	if(onum==0 && nnum==0) return TRUE;
	if(onum<nnum && !edb_auth_creatable_tuple_column(ec, tuple, ci)) return FALSE;
	if(onum>nnum && !edb_auth_deletable_tuple_column(ec, tuple, ci)) return FALSE;
	int	o = (otc && TCDatumNothing(otc)) ? TRUE : FALSE;
	int	n = (ntc && TCDatumNothing(ntc)) ? TRUE : FALSE;
	if((o!=n) && !edb_auth_writable_tuple_column(ec, tuple, ci)) return FALSE;

	int	status = TRUE;
	EdbDatum odatum = otc ? TCDatumFirst(otc) : NULL;
	EdbDatum ndatum = ntc ? TCDatumFirst(ntc) : NULL;
	while(odatum && ndatum) {
		Link	l;

		if((DatumDateFrom(odatum)!=DatumDateFrom(ndatum)
				|| DatumDateTo(odatum)!=DatumDateTo(ndatum)
				|| edb_datum_diff(ec, odatum, ndatum))
			&& !edb_auth_writable_tuple_column(ec, tuple, ci)) {
			status = FALSE;
		} else {
			ForLink(l, ci->l_child) {
				EdbColumnInfo		cci = l->d;
				if(cci->obsoleted) continue;

				int st = edb_tuple_check_updatable_column(ec, tuple, cci,
					otuple, (otuple && otc && odatum) ? edb_tuple_seek_TC_by_xmlname(ec, otuple, otc, odatum, ColumnXN(cci)) : NULL,
					ntuple, (ntuple && ntc && ndatum) ? edb_tuple_seek_TC_by_xmlname(ec, ntuple, ntc, ndatum, ColumnXN(cci)) : NULL);
				if(!st) {
					edb_trace(ec, "CHECK: 編集不可の項目[%s]が変更されています．\n", ColumnName_ja(cci));
					status = FALSE;
				}
			}
		}
		odatum = TCDatumNext(otc, odatum);
		ndatum = TCDatumNext(ntc, ndatum);
	}
	return status;
}

static
int
edb_tuple_check_updatable_tuple(EdbContext ec, EdbTupleInfo otuple, EdbTupleInfo ntuple)
{
	EdbTupleInfo	tuple = NULL;
	Link	l;

	if(otuple==NULL && ntuple==NULL) return TRUE;
	tuple = otuple ? otuple : ntuple;
	if(!tuple) return TRUE;
	int status = TRUE;
	ForLink(l, TupleTable(tuple)->l_column) {
		EdbColumnInfo		ci = l->d;
		if(ci->obsoleted) continue;

		int st = edb_tuple_check_updatable_column(ec, tuple, ci,
			otuple, otuple ? edb_tuple_seek_TC_by_xmlname(ec, otuple, NULL, NULL, ColumnXN(ci)) : NULL,
			ntuple, ntuple ? edb_tuple_seek_TC_by_xmlname(ec, ntuple, NULL, NULL, ColumnXN(ci)) : NULL);
		if(!st) {
			edb_trace(ec, "CHECK: 編集不可の項目[%s]が変更されています．\n", ColumnName_ja(ci));
			status = FALSE;
		}
	}
	return status;
}
#endif	/* EDB_READER */

int
edb_tuple_check_updatable(EdbContext ec, EdbTupleInfo tuple)
{
	int	(*check_func) EdbTupleCheckUpdatable_P;

	if(edb_table_module_load(ec, TupleTable(tuple)) && (check_func = TupleTable(tuple)->tuple_check_updatable)) {
		return (*check_func)(ec, tuple);
	}

#if defined(EDB_READER)
	if(!TupleAvail(tuple)) return TRUE;
	if(EIDisVALID(TupleMapto(tuple))) return TRUE;
	if(tuple->l_column==NULL) {
		if(ec->japanese) {
			edb_trace(ec, "CHECK: 項目がひとつも定義されていません．\n");
		} else {
			edb_trace(ec, "CHECK: No column is defined.\n");
		}
		return FALSE;
	}
	return TRUE;
#else
	EdbTupleInfo	ntuple = edb_tuple_duplicate(ec, tuple);
	EdbTupleInfo	otuple = NULL;
	if(EIDisValid(TupleEID(tuple))) otuple = edb_tuple_get(ec, TupleEID(tuple), 0);
	if(otuple!=NULL) otuple = edb_tuple_duplicate(ec, otuple);

	// 項目の編集権限チェックのための追加．
	int	status = TRUE;
	if(!TupleAvail(ntuple)) {
		status = edb_tuple_check_updatable_tuple(ec, otuple && TupleAvail(otuple) ? otuple : NULL, NULL);
	} else if(EIDisVALID(TupleMapto(ntuple))) {
		status = edb_tuple_check_updatable_tuple(ec, otuple && TupleAvail(otuple) ? otuple : NULL, ntuple);
	} else if(ntuple->l_column==NULL) {
		edb_trace(ec,
			ec->japanese ? "CHECK: 項目がひとつも定義されていません．\n" : "CHECK: No column is defined.\n");
		status = FALSE;
	} else {
		status = edb_tuple_check_updatable_tuple(ec, otuple && TupleAvail(otuple) ? otuple : NULL, ntuple);
	}
	if(ntuple) edb_tuple_free(ec, ntuple);
	if(otuple) edb_tuple_free(ec, otuple);

	return status;
#endif
}

int
edb_tuple_check_writable_by(EdbContext ec, EdbTupleInfo tuple, eid_t op)
{
	int	(*check_func) EdbTupleCheckWritableBy_P;

	if(edb_table_module_load(ec, TupleTable(tuple)) && (check_func = TupleTable(tuple)->tuple_check_writable_by)) {
		return (*check_func)(ec, tuple, op);
	}
	return FALSE;
}

EdbTupleColumn
edb_tuple_seek_column(EdbContext ec, EdbTupleInfo tuple, utf8_t *name)
{
	register EdbTupleColumn	tc;

	if((tc=edb_tuple_seek_column_by_xmlname(ec, tuple, name))) return tc;
	if((tc=edb_tuple_seek_column_by_sqlname(ec, tuple, name))) return tc;
	if((tc=edb_tuple_seek_column_by_name(ec, tuple, name))) return tc;
	return NULL;
}

EdbTupleColumn
edb_tuple_seek_column_by_name(EdbContext ec, EdbTupleInfo tuple, utf8_t *name)
{
	register Link	l;

	ForTupleColumn(l, tuple) {
		register EdbTupleColumn	tc = (EdbTupleColumn)l->d;

		if(!strcasecmp((char *)ColumnName_en(tc->ci), (char *)name)
			|| !strcasecmp((char *)ColumnName_ja(tc->ci), (char *)name)) {
			return tc;
		}
	}
	return NULL;
}

struct SEEK_COLUMN {
	EdbTupleColumn	tc_found;
	utf8_t	*target;
};

static
int
edb_tuple_seek_column_by_xmlname_func(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbTupleCBData cbd)
{
	struct SEEK_COLUMN	*sc = cbd->user_data;

	if(!strcmp((char *)ColumnXN(tc->ci), (char *)sc->target)) {
		sc->tc_found = tc;
		return FALSE;
	}
	return TRUE;
}

EdbTupleColumn
edb_tuple_seek_column_by_xmlname(EdbContext ec, EdbTupleInfo tuple, utf8_t *xmlname)
{
	struct SEEK_COLUMN	sc;

	if(xmlname[0]=='@' && xmlname[1]=='.') {
		char	*s;

		asprintf(&s, "%s.%s", TABLE_XN(TupleTable(tuple)), &xmlname[2]);
		xmlname = (utf8_t *)STRDUP(s);
		FREE(s);
	}
	sc.tc_found = NULL;
	sc.target = xmlname;
	edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_seek_column_by_xmlname_func, &sc);
	return sc.tc_found;
}

EdbTupleColumn
edb_tuple_seek_TC_by_xmlname(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc, EdbDatum datum, utf8_t *xmlname)
{
	struct SEEK_COLUMN	sc;
	struct EDB_TUPLE_CB_DATA	cb_data;
	EdbTupleCBData			cbd = &cb_data;

	if(tc && datum==NULL) return NULL;
	if(xmlname[0]=='@' && xmlname[1]=='.') {
		char	*s;

		asprintf(&s, "%s.%s", tc ? ColumnXN(tc->ci) : TABLE_XN(TupleTable(tuple)), &xmlname[2]);
		xmlname = (utf8_t *)STRDUP(s);
		FREE(s);
	}
	sc.tc_found = NULL;
	sc.target = xmlname;
	bzero(cbd, sizeof(struct EDB_TUPLE_CB_DATA));
	cbd->user_data = &sc;
	edb_tuple_TC_callback_sub(ec, tuple, 0, datum ? datum->l_child : tuple->l_column, edb_tuple_seek_column_by_xmlname_func, cbd);
	return sc.tc_found;
}

EdbTupleColumn
edb_tuple_seek_column_by_sqlname(EdbContext ec, EdbTupleInfo tuple, utf8_t *sqlname)
{
	register Link	l;

	ForTupleColumn(l, tuple) {
		register EdbTupleColumn	tc = (EdbTupleColumn)l->d;

		if(!strcasecmp((char *)ColumnSN(tc->ci), (char *)sqlname)) return tc;
	}
	return NULL;
}

EdbTupleColumn
edb_tuple_seek_picture_column(EdbContext ec, EdbTupleInfo tuple)
{
	register Link	l;
	EdbColumnInfo	ci = edb_table_seek_picture_column(ec, TupleTable(tuple));
	if(ci==NULL) return NULL;

	return edb_tuple_seek_column_by_xmlname(ec, tuple, ColumnXN(ci));
}

static
void
edb_tuple_mapping(EdbContext ec, EdbTupleInfo tuple, EdbTupleInfo mtuple)
{
	Link	l, l_column = NULL;

	if(mtuple==NULL) return;
	ForLink(l, TupleTable(tuple)->l_column) {
		EdbColumnInfo		ci = l->d;
		EdbTupleColumn	tc = edb_tuple_seek_TC_by_xmlname(ec, tuple, NULL, NULL, ColumnXN(ci));
		EdbTupleColumn	mtc;

		if(tc && (tc->datum || tc->nothing)) {
			LinkAppend(l_column, link_atom_new(edb_tuple_column_dup(ec, tuple, NULL, tc), edb_tuple_column_free));
			continue;
		}
		mtc = edb_tuple_seek_TC_by_xmlname(ec, mtuple, NULL, NULL, ColumnXN(ci));
		if(mtc && mtc->datum) {
			LinkAppend(l_column, link_atom_new(tc=edb_tuple_column_dup(ec, tuple, NULL, mtc), edb_tuple_column_free));
			tc->mapped = TRUE;
			continue;
		}
	}
	EdbLinkFree2(ec, tuple->l_column);
	tuple->l_column = l_column;
}

void
edb_tuple_get_mapto(EdbContext ec, EdbTupleInfo tuple)
{
	EdbTupleInfo	mtuple;

	if(!EIDisVALID(TupleMapto(tuple))) return;
	if(TupleMapto(tuple) && (mtuple=edb_tuple_get(ec, TupleMapto(tuple), 0))) {
		edb_tuple_mapping(ec, tuple, mtuple);
	}
}

static
int
edb_tuple_import_unmark_mapped(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbTupleCBData cbd)
{

	tc->mapped = FALSE;
	return TRUE;
}

void
edb_tuple_import(EdbContext ec, EdbTupleInfo tuple, eid_t mapto)
{
	EdbTupleInfo	mtuple;

	if(TupleMapto(tuple) && (mtuple=edb_tuple_get(ec, mapto, 0))) {
		edb_tuple_mapping(ec, tuple, mtuple);
		edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_import_unmark_mapped, NULL);
	}
}

void
edb_tuple_import_column(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc)
{
	Link	l_column;
	struct EDB_TUPLE_CB_DATA	cb_data;
	EdbTupleCBData			cbd = &cb_data;

	l_column = link_atom_new(tc, NULL);
	bzero(cbd, sizeof(struct EDB_TUPLE_CB_DATA));
	cbd->user_data = NULL;
	edb_tuple_TC_callback_sub(ec, tuple, 0, l_column, edb_tuple_import_unmark_mapped, cbd);
	EdbLinkFree(l_column);
}

static
EdbTupleInfo
edb_tuple_get_exec(EdbContext ec, eid_t eid, eoid_t eoid, EdbTableInfo template_ti, int permitted, int nrecurse)
{
	EdbTupleInfo	tuple, mtuple;

	tuple = edb_eid_list_restore(ec, eid);
	if(tuple == NULL) {
#if 1
		tuple = edb_tuple_get_XMLDB(ec, template_ti, eid, eoid);
		if(tuple) tuple = edb_eid_list_store(ec, eid, EDB_EID_IS_TUPLE, tuple);
#else
		if(template_ti) {
			template_ti;
		} else {
			EdbIndexInfo	ei;

			ei = edb_index_get(ec, eid);
			if(ei==NULL) return NULL;
			template_ti = edb_table_get(ec, ei->teid);
			eoid = ei->base.eoid;
		}
		if(template_ti == NULL) return NULL;
		tuple = edb_tuple_get_XMLDB(ec, template_ti, eid, eoid);
		tuple = edb_eid_list_store(ec, eid, EDB_EID_IS_TUPLE, tuple);
#endif
	}
	if(tuple==NULL) return NULL;
	if(!tuple->tuple_mapped && TupleMapto(tuple) && nrecurse>0 && (mtuple=edb_tuple_get_exec(ec, TupleMapto(tuple), 0, NULL, permitted, nrecurse-1))) {
		if(!tuple->tuple_mapped) {
			edb_tuple_mapping(ec, tuple, mtuple);
			tuple->tuple_mapped = TRUE;
		}
	}
#ifdef EDB_READER
#else /* !EDB_READER */
	/* check permission and eliminating datum */
	if(!permitted && !tuple->tuple_perm_checked) edb_tuple_check_permission(ec, tuple);
#endif /* EDB_READER */
	return tuple;
}

EdbTupleInfo
edb_tuple_get(EdbContext ec, eid_t eid, ...)
{
	va_list	ap;
	int	op;
	EdbTableInfo	template_ti = NULL;
	int	permitted = FALSE;
	int	nrecurse = ec->max_mapto_recurse;
	eoid_t	eoid = 0;
	EdbIndexInfo	ei;

	if(!EIDisVALID(eid)) return NULL;
	va_start(ap, eid);
	while((op = va_arg(ap, int)) != 0) {
		switch(op) {
		case EDB_TUPLE_GET_template_table:
			template_ti = va_arg(ap, EdbTableInfo);
			break;
		case EDB_TUPLE_GET_permitted:
			permitted = TRUE;
			break;
		case EDB_TUPLE_GET_without_mapped:
			nrecurse = 0;
			break;
		case EDB_TUPLE_GET_eoid:
			eoid = va_arg(ap, eoid_t);
			break;
		case EDB_TUPLE_GET_index:
			ei = va_arg(ap, EdbIndexInfo);
			template_ti = edb_table_get(ec, ei->teid);
			eoid = ei->base.eoid;
			break;
		default:
			break;
		}
	}
	va_end(ap);

	return edb_tuple_get_exec(ec, eid, eoid, template_ti, permitted, MAX_MAPTO_RECURSE);
}

/*
	edb_tuple_put--- put info
	edb_tuple_delete--- delete info
*/
struct TCD_REF_EID {
	Link		l_eid;
	Link		l_peid;
	BiTree	bt_eid;
};

static
int
edb_tuple_get_ref_eid_bitree2link(BiTree_obj_t obj, BiTree_hint_t hint, void *cbd)
{
	register struct TCD_REF_EID	*ref = cbd;
	register eid_t	eid = hint;

	if(!strcmp(obj, EDB_INDEX_REF_PEID_MARK)) LinkAppend(ref->l_peid, link_atom_new_eid2s(eid));
	LinkAppend(ref->l_eid, link_atom_new_eid2s(eid));
	return TRUE;
}

static
int
edb_tuple_get_ref_eid_datum(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbDatum datum, EdbTupleCBData cbd)
{
	register struct TCD_REF_EID	*ref = (struct TCD_REF_EID *)cbd->user_data;
	register Link	l, l_eid;

	if(datum==NULL) return TRUE;
	if(tc->mapped) return TRUE;

	l_eid = edb_datum_get_reference_list(ec, datum);
	if(l_eid==NULL) return TRUE;

	if(tc->ci->authorize && edb_datum_at_present(ec, datum)) {
		ForLink(l, l_eid) {
			register eid_t	eid = ATOI(l->d);
			register BiTree	node;

			node = bitree_search_node(ref->bt_eid, EDB_INDEX_REF_PEID_MARK, eid, NULL);
			if(node) {
				node->obj = EDB_INDEX_REF_PEID_MARK;
			} else {
				ref->bt_eid = BiTreeAdd(ref->bt_eid, EDB_INDEX_REF_PEID_MARK, eid, NULL, NULL, NULL);
			}
		}
	} else {
		ForLink(l, l_eid) {
			register eid_t	eid = ATOI(l->d);
			register BiTree	node;

			node = bitree_search_node(ref->bt_eid, EDB_INDEX_REF_EID_MARK, eid, NULL);
			if(!node) BiTreeAdd(ref->bt_eid, EDB_INDEX_REF_EID_MARK, eid, NULL, NULL, NULL);
		}
	}
	EdbLinkFree(l_eid);

	return TRUE;
}

void
edb_tuple_get_ref_eid(EdbContext ec, EdbTupleInfo tuple, Link *l_eid, Link *l_peid, BiTree *bt_eid)
{
	struct TCD_REF_EID ref;

	ref.l_eid = ref.l_peid = NULL;
	ref.bt_eid = NULL;
	edb_tuple_TCD_callback(ec, tuple, NULL, edb_tuple_get_ref_eid_datum, &ref);
	if(l_eid || l_peid) {
		ref.l_eid = ref.l_peid = NULL;
		bitree_callback_hint(ref.bt_eid, edb_tuple_get_ref_eid_bitree2link, (void *)&ref);
		if(l_eid) {
			*l_eid = ref.l_eid;
		} else {
			EdbLinkFree(ref.l_eid);
		}
		if(l_peid) {
			*l_peid = ref.l_peid;
		} else {
			EdbLinkFree(ref.l_peid);
		}
	}
	if(bt_eid) {
		*bt_eid = ref.bt_eid;
	} else {
		bitree_free(ref.bt_eid, NULL);
	}
}

EdbIndexInfo
edb_tuple_make_eid_index(EdbContext ec, EdbTupleInfo tuple)
{
	EdbIndexInfo	ei;
	Link	l, l_eid, l_peid;

	if(tuple==NULL) return NULL;
	ei = edb_index_new(ec);
	if(ei==NULL) return NULL;
	bcopy(TupleBase(tuple), &ei->base, sizeof(struct EDB_BASE_INFO));
	ei->teid = TABLE_EID(TupleTable(tuple));
	edb_tuple_get_ref_eid(ec, tuple, &l_eid, &l_peid, NULL);
	FREE(ei->reids);
	if(l_eid) {
		register eid_t	*p;

		p = ei->reids = (eid_t *)malloc(sizeof(eid_t)*(link_atoms(l_eid)+1));
		ForLink(l, l_eid) *p++ = ATOI(l->d);
		*p = 0;
	}
	EdbLinkFree(l_eid);
	FREE(ei->peids);
	if(l_peid) {
		register eid_t	*p;

		p = ei->peids = (eid_t *)malloc(sizeof(eid_t)*(link_atoms(l_peid)+1));
		ForLink(l, l_peid) *p++ = ATOI(l->d);
		*p = 0;
	}
	EdbLinkFree(l_peid);
	return ei;
}

int
edb_tuple_check_reference(EdbContext ec, EdbTupleInfo tuple)
{
	int	status;
	Link	l, l_peid, l_eid;

	/* check */
	if(TupleMapto(tuple) && !edb_eid_tuple_is_available(ec, TupleMapto(tuple))) return FALSE;
	edb_tuple_get_ref_eid(ec, tuple, &l_eid, &l_peid, NULL);
	status = TRUE;
	for(l=l_eid; l && status; l=l->n) {
		if(!edb_eid_tuple_is_available(ec, ATOI(l->d))) status = FALSE;
	}
	EdbLinkFree(l_eid);
	EdbLinkFree(l_peid);
	return status;
}

static
char *
edb_tuple_tc_make_xml_for_diff(EdbContext ec, EdbTupleInfo tuple, EdbTupleColumn tc, int completion)
{
//	EdbTableInfo	ti = TupleTable(tuple);
	Link	l_xml;
	char	*xml = NULL;
	EdbColumnInfo		ci;
	register EdbDatum	datum;

	l_xml = NULL;
	if(tc==NULL) return NULL;
	ci = tc->ci;
	if(!completion && tc->mapped) return NULL;
	ForTCDatum(datum, tc) {
		utf8_t	*datum_xml=NULL, *child_xml=NULL, *m_attr = NULL, *p_attr = NULL;
		Caption	cap;

		xml = NULL;
		if(datum->l_child) child_xml = edb_tuple_datum_make_xml(ec, tuple, datum->l_child, completion);
		if(completion && DatumEIDisValid(datum) && ci->typeInfo->multilingual && (cap = edb_make_mapped_caption(ec, ci, DatumEID(datum), CAP_norecurse))) {
			EdbDatum	d = edb_datum_dup(ec, DatumSuper(datum), datum);

			edb_datum_set(ec, d, DatumEID(d), cap->c_main_english, cap->c_main_japanese, cap->c_main_pronounce);
			if(ci->typeInfo) datum_xml = edb_datum_make_xml(ec, ColumnCType(ci), d, &m_attr, &p_attr, completion);
			edb_datum_free(ec, d);
			edb_caption_free(ec, cap);
		} else {
			if(ci->typeInfo) datum_xml = edb_datum_make_xml(ec, ColumnCType(ci), datum, &m_attr, &p_attr, completion);
		}

		if(child_xml==NULL && datum_xml==NULL) {
			if(m_attr) asprintf((char **)&xml, "<%s:%s %s%s%s/>", EDB_XMLNS, ColumnXN(ci), m_attr, p_attr?" ":"", p_attr ? p_attr : (utf8_t *)"");
		} else {
			asprintf((char **)&xml, "<%s:%s%s%s%s%s>%s%s</%s:%s>", EDB_XMLNS, ColumnXN(ci),
				m_attr?" ":"", m_attr ? m_attr : (utf8_t *)"",
				p_attr?" ":"", p_attr ? p_attr : (utf8_t *)"",
				datum_xml ? datum_xml : (utf8_t *)"",
				child_xml ? child_xml : (utf8_t *)"",
				EDB_XMLNS, ColumnXN(ci));
		}
		FREE(child_xml);
		FREE(datum_xml);
		FREE(m_attr);
		FREE(p_attr);
		if(xml) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(TCDatumFirst(tc)==NULL && TCDatumNothing(tc)) {
		asprintf((char **)&xml, "<%s:%s %s/>", EDB_XMLNS, ColumnXN(ci), XML_AN_ASSERT "=\"" XML_AV_ASSERT_NOTHING "\"");
		if(xml) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	xml = str_concatenate0(l_xml);
	EdbLinkFree(l_xml);
	if(STRisVALID(xml)) return xml;
	FREE(xml);
	return NULL;
}

static
int
edb_tuple_column_diff_compare(EdbContext ec, EdbTupleInfo new, EdbTupleColumn new_tc, EdbTupleInfo old, EdbTupleColumn old_tc)
{
	char	*new_xml, *old_xml;
	int	st;

	new_xml = edb_tuple_tc_make_xml_for_diff(ec, new, new_tc, FALSE);
	old_xml  = edb_tuple_tc_make_xml_for_diff(ec, old, old_tc, FALSE);

	if(new_xml==NULL && old_xml==NULL) {
		st = 0;
	} else if(new_xml==NULL) {
		st = -1;
	} else if(old_xml==NULL) {
		st = 1;
	} else {
		st = strcmp(new_xml, old_xml);
	}
	FREE(new_xml);
	FREE(old_xml);
	return st;
}

Link
edb_tuple_diff_columns(EdbContext ec, EdbTupleInfo new, EdbTupleInfo old)
{
	Link	l, l_column = NULL;
	int	differ;

	if(new == NULL || old == NULL) return NULL;
	differ = edb_base_compare(ec, TupleBase(new), TupleBase(old));
	if(differ) LinkAppend(l_column, link_atom_new_sdup(XML_EN_BASE));

#if 0
	ForTupleColumn(l, new) {
		EdbTupleColumn	new_tc = (EdbTupleColumn)l->d;
		EdbColumnInfo		ci = new_tc->ci;

		differ = edb_tuple_column_diff_compare(ec, new, new_tc, old, edb_tuple_seek_column_by_xmlname(ec, old, ColumnXN(ci)));
		if(differ) LinkAppend(l_column, link_atom_new_sdup(ColumnXN(ci)));
	}
#else
	ForLink(l, TupleTable(new)->l_column) {
		EdbColumnInfo		ci = l->d;
		EdbTupleColumn	old_tc = edb_tuple_seek_TC_by_xmlname(ec, old, NULL, NULL, ColumnXN(ci));
		EdbTupleColumn	new_tc = edb_tuple_seek_TC_by_xmlname(ec, new, NULL, NULL, ColumnXN(ci));

		differ = edb_tuple_column_diff_compare(ec, new, new_tc, old, old_tc);
		if(differ) LinkAppend(l_column, link_atom_new_sdup((char *)ColumnXN(ci)));
	}
#endif
	return l_column;
}

#define	MAX_CAPTION_RECURSE	10

/*
	edb_tuple_make_caption	-- make caption
*/
struct MAKE_CAP_CBD {
	Link	l_cap;
	Link	l_en;
	Link	l_ja;
	Link	l_pr;
	int	mode;
	int	recurse;
	Link	l_ci;
};

static
int
edb_tuple_make_caption_callback_func(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbTupleCBData cbd)
{
	struct MAKE_CAP_CBD *capd = (struct MAKE_CAP_CBD *)cbd->user_data;
	int	docap = FALSE;
	Link	l;
	EdbDatum	d;
	int	present_is_exist;

	if(capd->l_ci==NULL) {
		docap = tc && (tc->ci->caption==EDB_CAPTION_SHORT);
	} else {
		ForLink(l, capd->l_ci) {
			EdbColumnInfo	ci = l->d;

			if(tc && !strcmp((char *)ColumnXN(tc->ci), (char *)ColumnXN(ci))) docap = TRUE;
			if(docap) break;
		}
	}
	if(!docap) return TRUE;

	present_is_exist = FALSE;
	ForTCDatum(d, tc) {
		if(tc->ci->chronological && !edb_datum_at_present(ec, d)) continue;
		present_is_exist = TRUE;
		break;
	}
	ForTCDatum(d, tc) {
		if(tc->ci->chronological && !edb_datum_at_present(ec, d)) {
			if(present_is_exist) {
				if(!TCDatumIsFirst(tc, d) || !TCDatumIsLast(tc, d)) continue;
			} else if(!TCDatumIsFirst(tc, d)) continue;
		}
		if(DatumEIDisValid(d)) {
			EdbTupleInfo	rtuple;

			rtuple = edb_tuple_get(ec, DatumEID(d), 0);
			if(rtuple) {
				Caption	acap;
				Link	lm, l_mci = NULL;

				ForLink(lm, tc->ci->l_maplookup) {
					EdbMapLookup	map = lm->d;

					if(TABLE_EID(TupleTable(rtuple))==edb_maplookup_table_eid(ec, map)) {
						l_mci = map->l_caption_ci;
						break;
					}
				}

				acap = edb_tuple_make_caption_func(ec, rtuple, capd->mode, capd->recurse-1, l_mci);
				if(acap) {
					LinkAppend(capd->l_cap, link_atom_new_spf((capd->mode&CAP_nobracket) ? "%s" : "[%s]", acap->c_main));
					LinkAppend(capd->l_en, link_atom_new_sdup((char *)acap->c_main_english));
					LinkAppend(capd->l_ja, link_atom_new_sdup((char *)acap->c_main_japanese));
					LinkAppend(capd->l_pr, link_atom_new_sdup((char *)acap->c_main_pronounce));
					edb_caption_free(ec, acap);
				}
			}
		} else if(tc->ci->typeInfo->typeID == EDB_TYPE_PICTURE) {
			LinkAppend(capd->l_cap, link_atom_new_spf("\\P{%d}", TupleEID(tuple)));
		} else if(tc->ci->typeInfo->typeID == EDB_TYPE_TEXT) {
			char	*capstr = "", *s;

			if(DatumEnglishIsValid(d)) {
				if(ec->japanese) {
					asprintf(&s, "%s%s%s",
						STRisVALID(tc->ci->prefix.japanese)?tc->ci->prefix.japanese:(utf8_t *)"",
						DatumEnglish(d),
						STRisVALID(tc->ci->suffix.japanese)?tc->ci->suffix.japanese:(utf8_t *)"");
					capstr = STRDUP(s);
					FREE(s);
				} else {
					asprintf(&s, "%s%s%s",
						STRisVALID(tc->ci->prefix.english)?tc->ci->prefix.english:(utf8_t *)"",
						DatumEnglish(d),
						STRisVALID(tc->ci->suffix.english)?tc->ci->suffix.english:(utf8_t *)"");
					capstr = STRDUP(s);
					FREE(s);
				}
			}
			LinkAppend(capd->l_cap, link_atom_new_sdup(capstr));
			LinkAppend(capd->l_en, link_atom_new_sdup((char *)DatumEnglish(d)));
			LinkAppend(capd->l_ja, link_atom_new_sdup((char *)DatumEnglish(d)));
			LinkAppend(capd->l_pr, link_atom_new_sdup(""));
		} else {
			char	*capstr = "", *s;

			if(ec->japanese) {
				if(DatumJapaneseIsValid(d)) {
					asprintf(&s, "%s%s%s",
						STRisVALID(tc->ci->prefix.japanese)?tc->ci->prefix.japanese:(utf8_t *)"",
						DatumJapanese(d),
						STRisVALID(tc->ci->suffix.japanese)?tc->ci->suffix.japanese:(utf8_t *)"");
					capstr = STRDUP(s);
					FREE(s);
				} else if(DatumEnglishIsValid(d)) {
					asprintf(&s, "%s%s%s",
						STRisVALID(tc->ci->prefix.english)?tc->ci->prefix.english:(utf8_t *)"",
						DatumEnglish(d),
						STRisVALID(tc->ci->suffix.english)?tc->ci->suffix.english:(utf8_t *)"");
					capstr = STRDUP(s);
					FREE(s);
				}
			} else {
				if(DatumEnglishIsValid(d)) {
					asprintf(&s, "%s%s%s",
						STRisVALID(tc->ci->prefix.english)?tc->ci->prefix.english:(utf8_t *)"",
						DatumEnglish(d),
						STRisVALID(tc->ci->suffix.english)?tc->ci->suffix.english:(utf8_t *)"");
					capstr = STRDUP(s);
					FREE(s);
				} else if(DatumJapaneseIsValid(d)) {
					asprintf(&s, "%s%s%s",
						STRisVALID(tc->ci->prefix.japanese)?tc->ci->prefix.japanese:(utf8_t *)"",
						DatumJapanese(d),
						STRisVALID(tc->ci->suffix.japanese)?tc->ci->suffix.japanese:(utf8_t *)"");
					capstr = STRDUP(s);
					FREE(s);
				}
			}
			LinkAppend(capd->l_cap, link_atom_new_sdup(capstr));
			LinkAppend(capd->l_en, link_atom_new_sdup((char *)DatumEnglish(d)));
			LinkAppend(capd->l_ja, link_atom_new_sdup((char *)DatumJapanese(d)));
			LinkAppend(capd->l_pr, link_atom_new_sdup((char *)DatumPronounce(d)));
		}
		if(tc->ci->alias) break;
	}
	return TRUE;
}

static	eid_t	edb_tuple_hierarchyparent_anytime	__P((EdbContext ec, EdbTupleInfo tuple));

Caption
edb_tuple_make_caption_func(EdbContext ec, EdbTupleInfo tuple, int mode, int recurse, Link l_ci)
{
	char	*str;
	Link	l_cap = NULL, l_en = NULL, l_ja = NULL, l_pr = NULL;
	Caption	cap;
	struct	MAKE_CAP_CBD	capd;

	if(edb_table_module_load(ec, TupleTable(tuple)) && TupleTable(tuple)->tuple_make_caption_func) {
		cap = (*TupleTable(tuple)->tuple_make_caption_func)(ec, tuple, mode, recurse, l_ci);
		return cap;
	}

	if(!(recurse>0)) return NULL;

	cap = edb_caption_new(ec);
	if(cap==NULL) return NULL;

	if(tuple == NULL) {
		cap->c_main = (utf8_t *)strdup("(no caption)");
		cap->c_main_english = (utf8_t *)strdup("(no caption)");
		cap->c_main_japanese = (utf8_t *)strdup("(見出しなし)");
		cap->c_main_pronounce = (utf8_t *)strdup("みだしなし");
		return cap;
	}

	bzero(&capd, sizeof(struct MAKE_CAP_CBD));
	capd.mode = mode;
	capd.recurse = recurse;
	capd.l_ci = l_ci;
	edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_make_caption_callback_func, &capd);
	l_cap = capd.l_cap;
	l_en = capd.l_en;
	l_ja = capd.l_ja;
	l_pr = capd.l_pr;

	if(!ec->japanese && (capd.mode&CAP_noreverse)==0) LinkReverse(l_cap);
	if(TupleIsPerson(tuple)) {
		str = str_concatenate(l_cap, ' ');
	} else if(capd.mode&CAP_delimiter_comma) {
		str = str_concatenate_str(l_cap, ", ");
	} else {
		str = str_concatenate(l_cap, '/');
	}
	EdbLinkFree(l_cap);
	str_oneline(str);
	FREE(cap->c_main);
	cap->c_main = (utf8_t *)str;

	str = str_concatenate(l_en, ' ');
	EdbLinkFree(l_en);
	str_oneline(str);
	FREE(cap->c_main_english);
	cap->c_main_english = (utf8_t *)str;

	str = str_concatenate(l_ja, ' ');
	EdbLinkFree(l_ja);
	str_oneline(str);
	FREE(cap->c_main_japanese);
	cap->c_main_japanese = (utf8_t *)str;

	str = str_concatenate(l_pr, ' ');
	EdbLinkFree(l_pr);
	str_oneline(str);
	FREE(cap->c_main_pronounce);
	cap->c_main_pronounce = (utf8_t *)str;

	if(edb_table_is_hierarchical(ec, TupleTable(tuple)) && !(mode&CAP_norecurse) && recurse>0) {
		eid_t	parent_eid;

		parent_eid = edb_tuple_hierarchyparent_anytime(ec, tuple);
		if(EIDisVALID(parent_eid) && !edb_caption_is_omitted(ec, parent_eid)) {
			EdbTupleInfo	up_tuple;
			Caption	pcap = NULL;

			up_tuple = edb_tuple_get(ec, parent_eid, 0);
			if(up_tuple) pcap = edb_tuple_make_caption_func(ec, up_tuple, mode, recurse-1, NULL);
			if(pcap) {
				l_cap = link_atom_new_sdup((char *)pcap->c_main);
				LinkAppend(l_cap, link_atom_new(cap->c_main, free));
				cap->c_main = (utf8_t *)str_concatenate(l_cap, '.');
				EdbLinkFree(l_cap);

				l_en = link_atom_new_sdup((char *)pcap->c_main_english);
				LinkAppend(l_en, link_atom_new(cap->c_main_english, free));
				cap->c_main_english = (utf8_t *)str_concatenate(l_en, '.');
				EdbLinkFree(l_en);

				l_ja = link_atom_new_sdup((char *)pcap->c_main_japanese);
				LinkAppend(l_ja, link_atom_new(cap->c_main_japanese, free));
				cap->c_main_japanese = (utf8_t *)str_concatenate(l_ja, '.');
				EdbLinkFree(l_ja);

				l_pr = link_atom_new_sdup((char *)pcap->c_main_pronounce);
				LinkAppend(l_pr, link_atom_new(cap->c_main_pronounce, free));
				cap->c_main_pronounce = (utf8_t *)str_concatenate(l_pr, ' ');
				EdbLinkFree(l_pr);

				edb_caption_free(ec, pcap);
			}
		}
	}
	return cap;
}

static
utf8_t *
edb_tuple_make_longcaption_func(EdbContext ec, EdbTupleInfo tuple, int mode, int recurse)
{
	char	*str;
	Link	lcaption = NULL;
	Link	mlcaption = NULL;
	int	is_organ;
	Link	lc;

	if(!(recurse>0)) return (utf8_t *)strdup("");
	if(tuple == NULL) return (utf8_t *)strdup("(no caption)");

	is_organ = TupleIsOrganization(tuple);
// for multilingual caption
#if 1
	ForTupleColumn(lc, tuple) {
		EdbTupleColumn	tc = (EdbTupleColumn)lc->d;

		if(tc->ci->caption == EDB_CAPTION_SHORT && tc->ci->multilingualcaption) {
			EdbDatum	d;

			ForTCDatum(d, tc) {
				char *capstr;

				if(tc->ci->hierarchyparent && !(mode&CAP_norecurse)) continue;
				if(tc->ci->chronological && !edb_datum_at_present(ec, d)) continue;
				if(tc->ci->typeInfo->basicTypeID == EDB_TYPE_EID) {
					if(DatumEID(d)>0) {
						Caption	acap;
						EdbTupleInfo	rtuple;

						rtuple = edb_tuple_get(ec, DatumEID(d), 0);
						if(rtuple) {
							Link	lm, l_mci = NULL;

							ForLink(lm, tc->ci->l_maplookup) {
								EdbMapLookup	map = lm->d;

								if(TABLE_EID(TupleTable(rtuple))==edb_maplookup_table_eid(ec, map)) {
									l_mci = map->l_caption_ci;
									break;
								}
							}

							acap = edb_tuple_make_caption_func(ec, rtuple, mode, recurse-1, l_mci);
							if(acap) LinkAppend(lcaption, link_atom_new_spf((mode&CAP_nobracket) ? "%s" : "[%s]", ec->japanese ? acap->c_main_english : acap->c_main_japanese));
							edb_caption_free(ec, acap);
						}
					}
				} else if(tc->ci->typeInfo->basicTypeID == EDB_TYPE_NAME && DatumEID(d)>0) {
					EdbTupleInfo	rtuple;

					rtuple = edb_tuple_get(ec, DatumEID(d), 0);
					if(rtuple) {
						Caption	acap;
						Link	lm, l_mci = NULL;

						ForLink(lm, tc->ci->l_maplookup) {
							EdbMapLookup	map = lm->d;

							if(TABLE_EID(TupleTable(rtuple))==edb_maplookup_table_eid(ec, map)) {
								l_mci = map->l_caption_ci;
								break;
							}
						}

						acap = edb_tuple_make_caption_func(ec, rtuple, mode, recurse-1, l_mci);
						if(acap) LinkAppend(lcaption, link_atom_new_spf((mode&CAP_nobracket) ? "%s" : "[%s]", ec->japanese ? acap->c_main_english : acap->c_main_japanese));
						edb_caption_free(ec, acap);
					}
				} else if(tc->ci->typeInfo->typeID == EDB_TYPE_PICTURE) {
					LinkAppend(lcaption, link_atom_new_spf("\\P{%d}", TupleEID(tuple)));
				} else {
					if(!ec->japanese) {
						if(DatumJapaneseIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumJapanese(d)));
						} else if(DatumEnglishIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumEnglish(d)));
						}
					} else {
						if(DatumEnglishIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumEnglish(d)));
						} else if(DatumJapaneseIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumJapanese(d)));
						}
					}
				}
				if(tc->ci->alias) break;
			}
		}
	}
#endif
	ForTupleColumn(lc, tuple) {
		EdbTupleColumn	tc = (EdbTupleColumn)lc->d;

		if(tc->ci->caption == EDB_CAPTION_LONG) {
			EdbDatum	d;

			ForTCDatum(d, tc) {
				char *capstr;

				if(tc->ci->hierarchyparent && !(mode&CAP_norecurse)) continue;
				if(tc->ci->chronological && !edb_datum_at_present(ec, d)) continue;
				if(tc->ci->typeInfo->basicTypeID == EDB_TYPE_EID) {
					if(DatumEID(d)>0) {
						Caption	acap;
						EdbTupleInfo	rtuple;

						rtuple = edb_tuple_get(ec, DatumEID(d), 0);
						if(rtuple) {
							Link	lm, l_mci = NULL;

							ForLink(lm, tc->ci->l_maplookup) {
								EdbMapLookup	map = lm->d;

								if(TABLE_EID(TupleTable(rtuple))==edb_maplookup_table_eid(ec, map)) {
									l_mci = map->l_caption_ci;
									break;
								}
							}

							acap = edb_tuple_make_caption_func(ec, rtuple, mode, recurse-1, l_mci);
							if(acap) LinkAppend(lcaption, link_atom_new_spf((mode&CAP_nobracket) ? "%s" : "[%s]", acap->c_main));
							edb_caption_free(ec, acap);
						}
					}
				} else if(tc->ci->typeInfo->basicTypeID == EDB_TYPE_NAME && DatumEID(d)>0) {
					EdbTupleInfo	rtuple;

					rtuple = edb_tuple_get(ec, DatumEID(d), 0);
					if(rtuple) {
						Caption	acap;
						Link	lm, l_mci = NULL;

						ForLink(lm, tc->ci->l_maplookup) {
							EdbMapLookup	map = lm->d;

							if(TABLE_EID(TupleTable(rtuple))==edb_maplookup_table_eid(ec, map)) {
								l_mci = map->l_caption_ci;
								break;
							}
						}

						acap = edb_tuple_make_caption_func(ec, rtuple, mode, recurse-1, l_mci);
						if(acap) LinkAppend(lcaption, link_atom_new_spf((mode&CAP_nobracket) ? "%s" : "[%s]", acap->c_main));
						edb_caption_free(ec, acap);
					}
				} else if(tc->ci->typeInfo->typeID == EDB_TYPE_PICTURE) {
					LinkAppend(lcaption, link_atom_new_spf("\\P{%d}", TupleEID(tuple)));
				} else {
					if(ec->japanese) {
						if(DatumJapaneseIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumJapanese(d)));
						} else if(DatumEnglishIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumEnglish(d)));
						}
					} else {
						if(DatumEnglishIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumEnglish(d)));
						} else if(DatumJapaneseIsValid(d)) {
							LinkAppend(lcaption, link_atom_new_sdup((char *)DatumJapanese(d)));
						}
					}
				}
				if(tc->ci->alias) break;
			}
		}
	}
	if(!ec->japanese && (mode&CAP_noreverse)==0) LinkReverse(lcaption);
#if 0
#if 1
	/* hierarchical (e.g. organization) */
	if(edb_table_is_hierarchical(ec, TupleTable(tuple)) && !(mode&CAP_norecurse)) {
		EdbLinkFree(lcaption);
		lcaption = NULL;
	}
#endif
#endif
	if(XML_AN_DATE_FROM_DEFAULT < TupleDateFrom(tuple) ||  TupleDateTo(tuple) < XML_AN_DATE_TO_DEFAULT) {
		char	*s, *f, *t;
		int	d;

		if(XML_AN_DATE_FROM_DEFAULT < (d=TupleDateFrom(tuple))) {
			int day = DATE_D(d);
			if(day==0) {
				asprintf(&f, "%d年%d月初日", DATE_Y(d), DATE_M(d));
			} else if(day>=32) {
				asprintf(&f, "%d年%d月末日", DATE_Y(d), DATE_M(d));
			} else {
				asprintf(&f, "%d年%d月%d日", DATE_Y(d), DATE_M(d), day);
			}
		} else {
			f = strdup("");
		}
		if((d=TupleDateTo(tuple)) < XML_AN_DATE_TO_DEFAULT) {
			int day = DATE_D(d);
			if(day==0) {
				asprintf(&t, "%d年%d月初日", DATE_Y(d), DATE_M(d));
			} else if(day>=32) {
				asprintf(&t, "%d年%d月末日", DATE_Y(d), DATE_M(d));
			} else {
				asprintf(&t, "%d年%d月%d日", DATE_Y(d), DATE_M(d), day);
			}
		} else {
			t = strdup("");
		}
		LinkAppend(lcaption, link_atom_new_spf("%s〜%s", f, t));
		FREE(f);
		FREE(t);
	}
	if(TupleMapto(tuple)!=0 && !(mode&CAP_nomapto)) {
		EdbTupleInfo	mtuple = edb_tuple_get(ec, TupleMapto(tuple), 0);
		if(mtuple) {
			Caption	mcap = edb_tuple_make_caption_func(ec, mtuple, mode, recurse-1, NULL);
			if(mcap) LinkAppend(lcaption, link_atom_new_spf((mode&CAP_nobracket) ? "->%s %s" : "->%s[%s]", TupleName_ja(mtuple), mcap->c_main));
			edb_caption_free(ec, mcap);
		} else {
			LinkAppend(lcaption, link_atom_new_sdup("->[]"));
		}
	}
	if(mode&CAP_delimiter_comma) {
		str = str_concatenate_str(lcaption, ", ");
	} else {
		str = str_concatenate(lcaption, '/');
	}
	EdbLinkFree(lcaption);
	str_oneline(str);
	return (utf8_t *)str;
}

Caption
edb_tuple_make_caption(EdbContext ec, EdbTupleInfo tuple, int mode)
{
	int	recurse;
	Caption	cap;

	if(tuple==NULL) {
		cap = edb_caption_new(ec);
		if(cap==NULL) return NULL;
		cap->c_main = (utf8_t *)strdup("(no caption)");
		cap->c_main_english = (utf8_t *)strdup("(no caption)");
		cap->c_main_japanese = (utf8_t *)strdup("(見出しなし)");
		cap->c_main_pronounce = (utf8_t *)strdup("みだしなし");
		cap->c_sub = (utf8_t *)strdup("(no caption)");
		return cap;
	}

	if(mode&CAP_norecurse) {
		recurse = MAX_CAPTION_RECURSE;
	} else {
		recurse = MAX_CAPTION_RECURSE;
	}
	cap = edb_tuple_make_caption_func(ec, tuple, mode, recurse, NULL);
	if(cap && (mode&CAP_long)) {
		FREE(cap->c_sub);
		cap->c_sub = edb_tuple_make_longcaption_func(ec, tuple, mode, recurse+1);
	}
	if(!STRisValid(cap->c_main)) {
		if(cap->c_main) FREE(cap->c_main);
		asprintf((char **)&cap->c_main, "(no caption; EID=%d)", TupleEID(tuple));
	}
	if(!STRisValid(cap->c_main_english) && !STRisValid(cap->c_main_japanese)) {
		if(cap->c_main_english) FREE(cap->c_main_english);
		asprintf((char **)&cap->c_main_english, "(no caption; EID=%d)", TupleEID(tuple));
		if(cap->c_main_japanese) FREE(cap->c_main_japanese);
		asprintf((char **)&cap->c_main_japanese, "(見出しなし; EID=%d)", TupleEID(tuple));
	}
	return cap;
}

Caption
edb_tuple_make_specified_caption(EdbContext ec, EdbTupleInfo tuple, int mode, Link	l_ci)
{
	int	recurse;
	Caption	cap;

	if(tuple==NULL) {
		cap = edb_caption_new(ec);
		if(cap==NULL) return NULL;
		cap->c_main = (utf8_t *)strdup("(no caption)");
		cap->c_main_english = (utf8_t *)strdup("(no caption)");
		cap->c_main_japanese = (utf8_t *)strdup("(見出しなし)");
		cap->c_main_pronounce = (utf8_t *)strdup("みだしなし");
		cap->c_sub = (utf8_t *)strdup("(no caption)");
		return cap;
	}

	if(mode&CAP_norecurse) {
		recurse = MAX_CAPTION_RECURSE;
	} else {
		recurse = MAX_CAPTION_RECURSE;
	}
	cap = edb_tuple_make_caption_func(ec, tuple, mode, recurse, l_ci);
	if(cap && (mode&CAP_long)) {
		FREE(cap->c_sub);
		cap->c_sub = edb_tuple_make_longcaption_func(ec, tuple, mode, recurse+1);
	}
	if(!STRisValid(cap->c_main)) {
		if(cap->c_main) FREE(cap->c_main);
		asprintf((char **)&cap->c_main, "(no caption; EID=%d)", TupleEID(tuple));
	}
	if(!STRisValid(cap->c_main_english) && !STRisValid(cap->c_main_japanese)) {
		if(cap->c_main_english) FREE(cap->c_main_english);
		asprintf((char **)&cap->c_main_english, "(no caption; EID=%d)", TupleEID(tuple));
		if(cap->c_main_japanese) FREE(cap->c_main_japanese);
		asprintf((char **)&cap->c_main_japanese, "(見出しなし; EID=%d)", TupleEID(tuple));
	}
	return cap;
}

int
edb_tuple_expand_check(EdbContext ec, EdbTupleInfo tuple)
{
	Link	l, l_index;
	EdbTupleInfo	rtuple;

	if(tuple==NULL) return FALSE;
	if(!tuple->expand_checked) {
		l_index = edb_eid_get_tuple_indexes(ec, EDB_INDEX_GET_map, TupleEID(tuple), 0);
		tuple->expand_map = NULL;
		ForLink(l, l_index) {
			EdbBaseInfo	eb = &((EdbIndexInfo)l->d)->base;

			if(!BaseAvail(eb)) continue;
			rtuple = edb_tuple_get(ec, BaseEID(eb), 0);
			if(rtuple) {		// for parallel processing
				Link	ll;

				ForLink(ll, tuple->expand_map) {
					if(ll->d==rtuple) break;
				}
				if(!ll) LinkAppend(tuple->expand_map, link_atom_new(rtuple, NULL));
			}
		}
		EdbLinkFree(l_index);
		tuple->expand_checked = TRUE;
	}
	return (tuple->expand_map!=NULL);
}

static
void
edb_tuple_expand_sub(EdbContext ec, EdbCatalogue ca, EdbTupleInfo tuple)
{
	Link	l;
	EdbTupleInfo	rtuple;

	if(tuple==NULL) return;
	if(edb_catalogue_lookup(ca, TupleEID(tuple))) return;
	edb_catalogue_append(ca, TupleEID(tuple), tuple);

	edb_tuple_expand_check(ec, tuple);
	ForLink(l, tuple->expand_map) edb_tuple_expand_sub(ec, ca, (EdbTupleInfo)l->d);
	if(EIDisVALID(TupleMapto(tuple))) {
		rtuple = edb_tuple_get(ec, TupleMapto(tuple), 0);
		if(rtuple) edb_tuple_expand_sub(ec, ca, rtuple);
	}
}

EdbCatalogue
edb_tuple_expand(EdbContext ec, EdbTupleInfo tuple)
{

	if(tuple==NULL) return NULL;
	if(tuple->ca_expand) return tuple->ca_expand;
	tuple->ca_expand = edb_catalogue_new();
	edb_tuple_expand_sub(ec, tuple->ca_expand, tuple);
	return tuple->ca_expand;
}

static
int
edb_tuple_hierarchyparent_func(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbTupleCBData cbd)
{
	EdbDatum	*parent = (EdbDatum *)cbd->user_data;

	if(tc->ci->hierarchyparent && tc->datum) {
		EdbDatum datum;

		ForTCDatum(datum, tc) {
			if(DatumEIDisValid(datum) && edb_datum_at_present(ec, datum)) *parent = datum;
		}
	}
	return TRUE;
}

static
int
edb_tuple_hierarchyparent_anytime_func(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbTupleCBData cbd)
{
	EdbDatum	*parent = (EdbDatum *)cbd->user_data;

	if(tc->ci->hierarchyparent && tc->datum) {
		EdbDatum datum;

		ForTCDatum(datum, tc) {
			if(DatumEIDisValid(datum)) *parent = datum;
		}
	}
	return TRUE;
}

eid_t
edb_tuple_hierarchyparent(EdbContext ec, EdbTupleInfo tuple)
{
	EdbDatum	parent = NULL;

	if(!edb_table_is_hierarchical(ec, TupleTable(tuple))) return 0;
	edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_hierarchyparent_func, &parent);
	if(parent) return DatumEID(parent);
	return 0;
}

static
eid_t
edb_tuple_hierarchyparent_anytime(EdbContext ec, EdbTupleInfo tuple)
{
	EdbDatum	parent = NULL;

	if(!edb_table_is_hierarchical(ec, TupleTable(tuple))) return 0;
	edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_hierarchyparent_func, &parent);
	if(parent) return DatumEID(parent);
	edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_hierarchyparent_anytime_func, &parent);
	if(parent) return DatumEID(parent);
	return 0;
}

EdbDatum
edb_tuple_hierarchyparent_datum(EdbContext ec, EdbTupleInfo tuple)
{
	EdbDatum	parent = NULL;

	if(!edb_table_is_hierarchical(ec, TupleTable(tuple))) return 0;
	edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_hierarchyparent_func, &parent);
	return parent;
}

int
edb_tuple_has_history(EdbContext ec, EdbTupleInfo tuple)
{

	if(tuple->number_of_histories==0 || !ec->readonly) {
		utf8_t	*query;

		asprintf((char **)&query, "SELECT COUNT(*) FROM %s WHERE %s=%d",
			EDB_XMLDB_TABLE,
			EDB_XMLDB_COLUMN_EID,
			TupleEID(tuple));
		tuple->number_of_histories = edb_pgsql_query_int(ec, query);
		FREE(query);
	}
	return (tuple->number_of_histories>1);
}

int
edb_tuple_print_additional_information(EdbContext ec, EdbTupleInfo tuple, Doc doc)
{
	int	(*print_func) EdbTuplePrintAdditionalInformation_P;
	int	status = FALSE;

	if(edb_table_module_load(ec, TupleTable(tuple)) && (print_func = TupleTable(tuple)->tuple_print_additional_information)) {
		doc_puts(doc, "<div class=\"contents\" style=\"font-size:smaller\">\n");
		status = (*print_func)(ec, tuple, doc);
		doc_puts(doc, "</div>\n");
	}
	return status;
}

static
int
edb_tuple_collect_conducted_columns_func(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbDatum datum, EdbTupleCBData cbd)
{
	register EdbTupleInfo	ptuple = (EdbTupleInfo)cbd->user_data;

	if(datum==NULL) return TRUE;
	if(!edb_datum_at_present(ec, datum)) return TRUE;
	if(!DatumEnglishIsValid(datum)) return TRUE;
	if(ColumnConductorUsage(tc->ci)==0) return TRUE;
	switch(ColumnConductorUsage(tc->ci)) {
	case USAGE_REQUIRED:		LinkAppend(ptuple->l_required_cxn, link_atom_new_sdup((char *)DatumEnglish(datum)));		break;
	case USAGE_PREFERRED:	LinkAppend(ptuple->l_preferred_cxn, link_atom_new_sdup((char *)DatumEnglish(datum)));	break;
	case USAGE_OPTIONAL:		LinkAppend(ptuple->l_optional_cxn, link_atom_new_sdup((char *)DatumEnglish(datum)));		break;
	case USAGE_UNUSED:		LinkAppend(ptuple->l_unused_cxn, link_atom_new_sdup((char *)DatumEnglish(datum)));		break;
	}

	return TRUE;
}

static
int
edb_tuple_check_conductor_func(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbTupleCBData cbd)
{

	if(tc->ci->conductor && tc->datum) {
		EdbDatum datum;

		ForTCDatum(datum, tc) {
			if(DatumEIDisValid(datum) && edb_datum_at_present(ec, datum)) {
				EdbTupleInfo ctuple = edb_tuple_get(ec, DatumEID(datum), 0);
				if(ctuple) {
					edb_tuple_TCD_callback(ec, ctuple, NULL, edb_tuple_collect_conducted_columns_func, tuple);
					tuple->conductor_found = TRUE;
				}
			}
		}
	}
	return TRUE;
}

static
void
edb_tuple_check_conductor(EdbContext ec, EdbTupleInfo tuple)
{

	if(!edb_table_has_conductor(ec, TupleTable(tuple))) {
		tuple->conductor_checked = TRUE;
		return;
	}
	if(tuple->conductor_checked) return;
	edb_tuple_TC_callback(ec, tuple, NULL, edb_tuple_check_conductor_func, NULL);
	if(tuple->l_required_cxn) tuple->l_required_cxn = str_uniq(tuple->l_required_cxn);
	if(tuple->l_preferred_cxn) tuple->l_preferred_cxn = str_uniq(tuple->l_preferred_cxn);
	if(tuple->l_optional_cxn) tuple->l_optional_cxn = str_uniq(tuple->l_optional_cxn);
	if(tuple->l_unused_cxn) tuple->l_unused_cxn = str_uniq(tuple->l_unused_cxn);
	tuple->conductor_checked = TRUE;
}

int
edb_tuple_has_conductor(EdbContext ec, EdbTupleInfo tuple)
{

	if(!edb_table_has_conductor(ec, TupleTable(tuple))) return FALSE;
	edb_tuple_check_conductor(ec, tuple);
	return tuple->conductor_found;
}

int
edb_tuple_column_usage(EdbContext ec, EdbTupleInfo tuple, EdbColumnInfo ci)
{

	edb_tuple_check_conductor(ec, tuple);
	if(ci->conductor) return USAGE_REQUIRED;
	if(!tuple->conductor_found) return ColumnUsage(ci);
	if(str_match(tuple->l_required_cxn, (char *)ColumnXN(ci))) return USAGE_REQUIRED;
	if(str_match(tuple->l_preferred_cxn, (char *)ColumnXN(ci))) return USAGE_PREFERRED;
	if(str_match(tuple->l_optional_cxn, (char *)ColumnXN(ci))) return USAGE_OPTIONAL;
	if(str_match(tuple->l_unused_cxn, (char *)ColumnXN(ci))) return USAGE_UNUSED;
	return ColumnUsage(ci);
}

int
edb_tuple_named_column_usage(EdbContext ec, EdbTupleInfo tuple, utf8_t *xn)
{
	EdbColumnInfo ci = edb_table_seek_column(ec, TupleTable(tuple), xn);

	if(ci==NULL) return USAGE_UNUSED;
	return edb_tuple_column_usage(ec, tuple, ci);
}

static
int
edb_tuple_get_minimum_date_TCD(EdbContext ec, EdbTupleInfo tuple, int layer, EdbTupleColumn tc, EdbDatum datum, EdbTupleCBData cbd)
{
	int	*date = (int *)cbd->user_data;

	if(tc->ci->typeInfo->typeID == EDB_TYPE_DATE && datum && DatumEnglishIsValid(datum)) {
		utf8_t *s = DatumEnglish(datum);
		int d = ATOI((char *)s);
		if(*date==0 || d<*date) *date = d;
	}
	return TRUE;
}

int
edb_tuple_get_minimum_date(EdbContext ec, EdbTupleInfo tuple)
{
	int date = 0;

	edb_tuple_TCD_callback(ec, tuple, NULL, edb_tuple_get_minimum_date_TCD, &date);
	return date;
}

