/*
	$Id: unicode.c,v 1.7 2008/08/09 09:21:37 alex Exp $
	unicodel.c--- unicode
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: unicode.c,v 1.7 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	"edb/unicode.h"

/*
	UTF8 <-> UCS4 converstion
*/

int
convUTF8toUCS4(utf8_t *utf, ucs4_t *ucs)
{
	register utf8_t	c;
	register ucs4_t	ch = 0;
	register int	n = 0;	/* number of characters will be placed in UCS4 */
	register int	iter = 0;

	while((c = *utf++)) {
		if(iter>0) {
			ch <<= 6;
			ch |= c & 0x3F;
			--iter;
			if(iter==0) {
				if(ucs) *ucs++ = ch;
				++n;
			}
		} else if((c & 0xFE) == 0xFC) {
			ch = c & 0x01;
			iter = 5;
		} else if((c & 0xFC) == 0xF8) {
			ch = c & 0x03;
			iter = 4;
		} else if((c & 0xF8) == 0xF0) {
			ch = c & 0x07;
			iter = 3;
		} else if((c & 0xF0) == 0xE0) {
			ch = c & 0x0F;
			iter = 2;
		} else if((c & 0xE0) == 0xC0) {
			ch = c & 0x1F;
			iter = 1;
		} else if((c & 0x80) == 0x80) {
			iter = 0;
		} else {
			if(ucs) *ucs++ = c;
			++n;
			iter = 0;
		}
	}
	if(ucs) *ucs = UCS4_NULL;
	return n;
}

#define	ucs4_bits(c)	(0x80 | ((c) & 0x3F))
#define	ucs4_put(c)	if(utf) {*utf++=(c); } n++;
#define	ucs4_putbits(c)	ucs4_put(ucs4_bits(c))

int
convUCS4toUTF8(ucs4_t *ucs, utf8_t *utf)
{
	register ucs4_t	c;
	register int	n = 0;	/* number of characters will be placed in UTF8 */

	while((c = *ucs++)!=UCS4_NULL) {
		if(c < 0x80) {
			ucs4_put(c);
		} else if(c < 0x800) {
			ucs4_put(0xC0 | (c >>  6));
			ucs4_putbits(c);
		} else if(c < 0x10000) {
			ucs4_put(0xE0 | (c >> 12));
			ucs4_putbits(c >>  6);
			ucs4_putbits(c);
		} else if(c < 0x200000) {
			ucs4_put(0xF0 | (c >> 18));
			ucs4_putbits(c >> 12);
			ucs4_putbits(c >>  6);
			ucs4_putbits(c);
		} else if(c < 0x400000) {
			ucs4_put(0xF8 | (c >> 24));
			ucs4_putbits(c >> 18);
			ucs4_putbits(c >> 12);
			ucs4_putbits(c >>  6);
			ucs4_putbits(c);
		} else if(c < 0x80000000) {
			ucs4_put(0xFC | (c >> 30));
			ucs4_putbits(c >> 24);
			ucs4_putbits(c >> 18);
			ucs4_putbits(c >> 12);
			ucs4_putbits(c >>  6);
			ucs4_putbits(c);
		} else {
			/* Error, cannot convert */
		}
	}
	if(utf) *utf = '\0';
	return n;
}

utf8_t	*
ucs4_to_utf8(ucs4_t *u4, utf8_t *u8)
{

	convUCS4toUTF8(u4, u8);
	return u8;
}

ucs4_t	*
utf8_to_ucs4(utf8_t *u8, ucs4_t *u4)
{

	convUTF8toUCS4(u8, u4);
	return u4;
}

/*
	string validation
*/

int
utf8string_is_valid(utf8_t *u)
{
	register utf8_t	c;
	register int	iter = 0;

	while((c = *u++)) {
		if(iter>0) {
			if((c & 0xC0) != 0x80) return FALSE;
			--iter;
		} else if((c & 0xFE) == 0xFC) {
			iter = 5;
		} else if((c & 0xFC) == 0xF8) {
			iter = 4;
		} else if((c & 0xF8) == 0xF0) {
			iter = 3;
		} else if((c & 0xF0) == 0xE0) {
			iter = 2;
		} else if((c & 0xE0) == 0xC0) {
			iter = 1;
		} else if((c & 0x80) == 0x80) {
			return FALSE;
		} else {
			iter = 0;
		}
	}
	if(iter>0) return FALSE;
	return TRUE;
}

/*
	string functions
*/

ucs4_t *
ucs4_strcat(
	register ucs4_t	*s1,
	register ucs4_t	*s2)
{
	register ucs4_t	*ret = s1;

	while(*s1++) ;
	s1--;
	while((*s1++ = *s2++)) ;
	return ret;
}

ucs4_t *
ucs4_strncat(
	register ucs4_t	*s1,
	register ucs4_t	*s2,
	register int	n)
{
	register ucs4_t	*ret = s1;

	while(*s1++) ;
	s1--;
	while(n-- > 0 && (*s1++ = *s2++)) ;
	if(n < 0) *s1 = 0;
	return ret;
}

int
ucs4_strcmp(
	register ucs4_t	*s1,
	register ucs4_t	*s2)
{

	while(*s1 && *s1 == *s2) s1++, s2++;
	return ((int)*s1 - (int)*s2);
}

int
ucs4_strncmp(
	register ucs4_t	*s1,
	register ucs4_t	*s2,
	register int n)
{

	while(n-- > 0 && *s1 && *s1 == *s2) s1++, s2++;
	if(n < 0) return 0;
	return ((int)*s1 - (int)*s2);
}

int
ucs4_strcasecmp(
	register ucs4_t	*s1,
	register ucs4_t	*s2)
{
	register ucs4_t c1, c2;

	while(*s1) {
		c1 = *s1;
		c2 = *s2;
		if(ucs4_isupper(c1)) c1 = ucs4_tolower(c1);
		if(ucs4_isupper(c2)) c2 = ucs4_tolower(c2);
		if(c1 != c2) break;
		s1++, s2++;
	}
	c1 = *s1;
	c2 = *s2;
	if(ucs4_isupper(c1)) c1 = ucs4_tolower(c1);
	if(ucs4_isupper(c2)) c2 = ucs4_tolower(c2);
	return ((int)c1 - (int)c2);
}

int
ucs4_strncasecmp(
	register ucs4_t	*s1,
	register ucs4_t	*s2,
	register int	n)
{
	register ucs4_t c1, c2;

	while(n-- > 0 && *s1) {
		c1 = *s1;
		c2 = *s2;
		if(ucs4_isupper(c1)) c1 = ucs4_tolower(c1);
		if(ucs4_isupper(c2)) c2 = ucs4_tolower(c2);
		if(c1 != c2) break;
		s1++, s2++;
	}
	if(n < 0) return 0;
	c1 = *s1;
	c2 = *s2;
	if(ucs4_isupper(c1)) c1 = ucs4_tolower(c1);
	if(ucs4_isupper(c2)) c2 = ucs4_tolower(c2);
	return ((int)c1 - (int)c2);
}

ucs4_t *
ucs4_strcpy(
	register ucs4_t	*s1,
	register ucs4_t	*s2)
{
	register ucs4_t	*ret = s1;

	while((*s1++ = *s2++)) ;
	return ret;
}

ucs4_t *
ucs4_strncpy(
	register ucs4_t	*s1,
	register ucs4_t	*s2,
	register int	n)
{
	register ucs4_t	*ret = s1;

	while(n-- > 0 && (*s1++ = *s2++)) ;
	while(n-- > 0) *s1++ = 0;
	return ret;
}

int
ucs4_strlen(register ucs4_t *s)
{
	register int	n = 0;

	while(*s++) n++;
	return n;
}

ucs4_t *
ucs4_index(
	register ucs4_t	*s,
	register ucs4_t	c)
{
	register int	x;

	while((x = *s++)) {
		if (x == c)
			return s - 1;
	}
	return 0;
}

ucs4_t *
ucs4_caseindex(
	register ucs4_t	*s,
	register ucs4_t	c)
{
	register int	x;

	if(ucs4_isupper(c)) c = ucs4_tolower(c);
	while((x = *s++)) {
		if(ucs4_isupper(x)) {
			if (ucs4_tolower(x) == c)
				return s - 1;
		} else {
			if (x == c)
				return s - 1;
		}
	}
	return 0;
}

ucs4_t *
ucs4_rindex(
	register ucs4_t	*s,
	register ucs4_t	c)
{
	register ucs4_t	*r = 0;
	register int	x;

	while((x = *s++)) {
		if(x == c) r = s - 1;
	}
	return r;
}

ucs4_t *
ucs4_strstr(
	register ucs4_t	*s,
	register ucs4_t	*t)
{
	register ucs4_t *p;
	register int n;

	n = ucs4_strlen(t);
	if(n == 0) return s;

	p = s;
	while(p) {
		p = ucs4_index(p, t[0]);
		if(!p) break;
		if(ucs4_strncmp(p, t, n) == 0) return p;
		++p;
	}
	return NULL;
}

ucs4_t *
ucs4_strcasestr(
	register ucs4_t	*s,
	register ucs4_t	*t)
{
	register ucs4_t	*p;
	register int	n;

	n = ucs4_strlen(t);
	if(n == 0) return s;

	p = s;
	while(p) {
		p = ucs4_caseindex(p, t[0]);
		if(!p) break;
		if(ucs4_strncasecmp(p, t, n) == 0) return p;
		++p;
	}
	return NULL;
}

ucs4_t *
ucs4_strdup(register ucs4_t *s)
{
	register ucs4_t *wbuf;

	if(s == NULL) return NULL;
	wbuf = (ucs4_t *)malloc((ucs4_strlen(s)+1) * sizeof(ucs4_t));
	if(wbuf == NULL) return NULL;
	ucs4_strcpy(wbuf, s);
	return wbuf;
}

ucs4_t *
ucs4_strndup(
	register ucs4_t	*s,
	register int	n)
{
	register ucs4_t *wbuf;

	if(s == NULL) return NULL;
	wbuf = (ucs4_t *)malloc((n+1) * sizeof(ucs4_t));
	if(wbuf == NULL) return NULL;
	ucs4_strncpy(wbuf, s, n);
	wbuf[n] = UCS4_NULL;
	return wbuf;
}

ucs4_t *
ucs4_strmerge(
	register ucs4_t	*s1,
	register ucs4_t	*s2)
{
	register int len;
	register ucs4_t *wbuf;

	len = ucs4_strlen(s1) + ucs4_strlen(s2);
	wbuf = (ucs4_t *)malloc((len+1) * sizeof(ucs4_t));
	if(wbuf == NULL) return NULL;
	ucs4_strcpy(wbuf, s1);
	ucs4_strcat(wbuf, s2);
	return wbuf;
}

ucs4_t *
ucs4_strrev(register ucs4_t *s)
{
	register int	len;
	register ucs4_t	*s1, *s2;

	len = ucs4_strlen(s);
	if(len<=1) return s;
	for(s1=s, s2=&s[len-1]; s1<s2; ++s1, --s2) {
		register ucs4_t	c;

		c = *s1;
		*s1 = *s2;
		*s2 = c;
	}
	return s;
}

void
ucs4_str_tokatakana(register ucs4_t *s)
{

	while(*s) {
		*s = ucs4_tokatakana(*s);
		++s;
	}
}

Link
ucs4_str_split(
	register ucs4_t	*str,
	register ucs4_t	c)
{
	register Link	lstr;
	register ucs4_t	*s, *p;

	s = str;
	lstr = NULL;
	while(*s) {
		p = ucs4_index(s, c);
		if(p != NULL) {
			LinkAppendTop(link_atom_new(ucs4_strndup(s, p-s), free), lstr);
			s = ++p;
		} else {
			LinkAppendTop(link_atom_new(ucs4_strdup(s), free), lstr);
			break;
		}
	}
	return link_reverse(lstr);
}

static	ucs4_t	ucs4_str_null_string[1] = {UCS4_NULL};

ucs4_t *
ucs4_str_concatenate(
	register Link	lstr,
	register ucs4_t	c)
{
	register int	len;
	register Link	l;
	register ucs4_t	*s, *str;

	if(lstr==NULL) return ucs4_strdup(ucs4_str_null_string);
	len = 0;
	for(l=lstr; l; l=l->n) len += ucs4_strlen((ucs4_t *)l->d) + 1;
	str = s = (ucs4_t *)malloc(sizeof(ucs4_t)*len);
	*s = UCS4_NULL;
	for(l=lstr; l; l=l->n) {
		ucs4_strcpy(s, (ucs4_t *)l->d);
		while(*s) ++s;
		if(l->n) {
			if(c) *s++ = c;
			*s = UCS4_NULL;
		}
	}
	return str;
}

Link
ucs4_str_split_bychar(register ucs4_t *str)
{
	register Link	l;
	register ucs4_t	*s;
	ucs4_t		 ustr[2];

	ustr[1] = UCS4_NULL;
	l = NULL;
	for(s=str; *s; ++s) {
		ustr[0] = *s;
		LinkAppendTop(link_atom_new(ucs4_strdup(ustr), free), l);
	}
	return link_reverse(l);
}

void
ucs4_str_capitalize1(register ucs4_t *s)
{

	for( ; *s; ++s) {
		if(ucs4_islower(*s)) *s = ucs4_toupper(*s) ;
	}
}

void
ucs4_str_capitalize(
	register ucs4_t *dst,
	register ucs4_t *src)
{

	while(*src) {
		if(ucs4_islower(*src)) {
			*dst++ = ucs4_toupper(*src++) ;
		} else {
			*dst++ = *src++;
		}
	}
	*dst = UCS4_NULL;
}

void
ucs4_str_minuscule1(register ucs4_t *s)
{

	for( ; *s; ++s) {
		if(ucs4_isupper(*s)) *s = ucs4_tolower(*s) ;
	}
}

void
ucs4_str_minuscule(
	register ucs4_t *dst,
	register ucs4_t *src)
{

	while(*src) {
		if(ucs4_isupper(*src)) {
			*dst++ = ucs4_tolower(*src) ;
			++src;
		} else {
			*dst++ = *src++;
		}
	}
	*dst = UCS4_NULL;
}

int
ucs4_str_isascii(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!ucs4_isascii(c)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_isalpha(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!ucs4_isalpha(c)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_isalnum(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!ucs4_isalnum(c)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_iswalnum(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!ucs4_iswalnum(c)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_isupper(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!ucs4_isupper(c)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_islower(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!ucs4_islower(c)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_isdigit(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!ucs4_isdigit(c)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_iskana(register ucs4_t *s)
{
	register ucs4_t	c;

	while((c=*s++)!=UCS4_NULL) {
		if(!(ucs4_ishiragana(c) || ucs4_iskatakana(c) || c==UCS4_SPACE)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_ishiragana(register ucs4_t *s)
{
	register ucs4_t	wc;

	while((wc=*s++)) {
		if(!(ucs4_ishiragana(wc) || wc==UCS4_SPACE)) return FALSE;
	}
	return TRUE;
}

int
ucs4_str_iskatakana(register ucs4_t *s)
{
	register ucs4_t	wc;

	while((wc=*s++)) {
		if(!(ucs4_iskatakana(wc) || wc==UCS4_SPACE)) return FALSE;
	}
	return TRUE;
}

void
ucs4_str_oneline(register ucs4_t *s)
{
	register ucs4_t	*dst, *src, c, *ul;
	register int	space;

	space = TRUE;
	ul = s;
	for(src = dst = s; (c=*src); ++src) {
		/* TAB，NL，空白(全角)，CR → 空白(半角) */
		if(c == UCS4_SPACE  || c == UCS4_WSPACE || c == UCS4_CR || c == UCS4_TAB || c == UCS4_NL ) {
			if(!space) {
				*dst++ = UCS4_SPACE;
				space = TRUE;
			}
		} else {
			*dst++ = c;
			ul = dst;
			space = FALSE;
		}
	}
	*ul = UCS4_NULL;
}

void
ucs4_str_standard(register ucs4_t *s)
{
	register int	i, len;
	register ucs4_t	*dst, *src, c;
	register int	space;

	/* 全角の空白文字と CR を半角の空白に変換する．*/
	for(src = s; *src; ++src) {
		c = *src;
		if(c == UCS4_WSPACE || c == UCS4_CR) *src = UCS4_SPACE;
	}

	/* 初めの空白文字と改行文字を削除する．*/
	dst = src = s;
	while( (*src == UCS4_TAB) || (*src == UCS4_SPACE) || (*src == UCS4_NL)) ++src;
	while((*dst++ = *src++)) ;

	/* 末尾の空白文字と改行文字を削除する．*/
	len = ucs4_strlen(s);
	for(i=len-1; i>=0; --i) {
		c = s[i];
		if(c == UCS4_TAB || c == UCS4_SPACE || c == UCS4_NL) {
			s[i] = UCS4_NULL;
		} else {
			break;
		}
	}

	/* 連続した空白を１つにする．*/
	space = FALSE;
	dst = src = s;
	while((c = *dst++ = *src++)) {
		if(space && (c == UCS4_SPACE)) {
			--dst;
		} else {
			space = (c == UCS4_SPACE);
		}
	}
#if 0
	*dst = UCS4_NULL;
#endif
}

void
ucs4_str_onephrase(register ucs4_t *s)
{
	register ucs4_t *c, *d;

	ucs4_str_oneline(s);

	/* 半角空白文字を消去する．*/
	for(c = d = s; (*d = *c); ++c) {
		if(*c != UCS4_SPACE) ++d;
	}
}

static
void
ucs4_str_token_to_string(register ucs4_t *str)
{
	register ucs4_t	*s, *d;

	for(d=s=str; *s; ++s) {
		if(s[0] == '\'') {
			if(s!=str && s[1]=='\'') {
				*d++ = '\'';
				++s;
			}
		} else {
			*d++ = s[0];
		}
	}
	*d = UCS4_NULL;
}

Link
ucs4_str_split_bytoken(register ucs4_t *str)
{
	register Link	l, lstr;
	register ucs4_t *s, *ss;
	register int quot, escape;

	str = UCS4_STRDUP(str);
	ucs4_str_oneline(str);
	ss = s = str;
	quot = FALSE;
	escape = FALSE;
	lstr = NULL;
	for(ss=s=str; *s; ++s) {
		register ucs4_t	c = *s;

		if(quot) {
			if(c == '\'') {
				quot = FALSE;
			} else {
				continue;
			}
		} else if(c=='\'') {
			quot = TRUE;
			continue;
		}
		if(c==UCS4_SPACE) {
			LinkAppendTop(link_atom_new(ucs4_strndup(ss, s-ss), free), lstr);
			ss = s+1;
		}
	}
	LinkAppendTop(link_atom_new(ucs4_strdup(ss), free), lstr);
	LinkReverse(lstr);
	for(l=lstr; l; l=l->n) ucs4_str_token_to_string((ucs4_t *)l->d);
	return lstr;
}


size_t
utf8_strlen(utf8_t *str)
{
	utf8_t *s;

	for(s=str; *s; ++s) ;
	return (s-str);
}

utf8_t *
utf8_strdup(utf8_t *str)
{
	size_t	len;
	utf8_t	*co;

	len = utf8_strlen(str) + 1;
	if((co=malloc(len)) == NULL) return (NULL);
	memcpy(co, str, len);
	return (co);
}

utf8_t *
utf8_strcpy(utf8_t *t, utf8_t *f)
{
	utf8_t *s = t;

	for( ; (*t = *f); ++f, ++t);
	return (s);
}

utf8_t *
utf8_strncpy(utf8_t *dst, utf8_t *src, size_t n)
{
	if (n != 0) {
		utf8_t *d = dst;
		utf8_t *s = src;

		do {
			if ((*d++ = *s++) == 0) {
				/* NUL pad the remaining n-1 bytes */
				while(--n != 0) *d++ = 0;
				break;
			}
		} while (--n != 0);
	}
	return (dst);
}

int
utf8_strcmp(utf8_t *s1, utf8_t *s2)
{

	while(*s1 == *s2++) if(*s1++ == 0) return (0);
	return (*(utf8_t *)s1 - *(utf8_t *)(s2 - 1));
}

int
utf8_strncmp(utf8_t *s1, utf8_t *s2, size_t n)
{

	if (n == 0) return (0);
	do {
		if (*s1 != *s2++) return (*(utf8_t *)s1 - *(utf8_t *)(s2 - 1));
		if (*s1++ == 0) break;
	} while (--n != 0);
	return (0);
}

int
utf8_strcasecmp(utf8_t *s1, utf8_t *s2)
{
	utf8_t *us1 = (utf8_t *)s1, *us2 = (utf8_t *)s2;

	while(tolower(*us1) == tolower(*us2++)) if(*us1++ == '\0') return (0);
	return (tolower(*us1) - tolower(*--us2));
}

int
utf8_strncasecmp(utf8_t *s1, utf8_t *s2, size_t n)
{
	if (n != 0) {
		utf8_t *us1 = (utf8_t *)s1, *us2 = (utf8_t *)s2;

		do {
			if (tolower(*us1) != tolower(*us2++)) return (tolower(*us1) - tolower(*--us2));
			if (*us1++ == '\0') break;
		} while (--n != 0);
	}
	return (0);
}

