/*
	$Id: xmldb.c,v 1.6 2011/04/18 09:45:28 alex Exp $
	xmldb.c--- input/output of XMLDB
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: xmldb.c,v 1.6 2011/04/18 09:45:28 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/xml.h"

eid_t
edb_xmldb_eid_max(EdbContext ec)
{

#if 0
	return edb_pgsql_query_int(ec, (utf8_t *)"SELECT MAX(" EDB_XMLDB_COLUMN_EID ") FROM " EDB_XMLDB_TABLE);
#else
	return edb_pgsql_query_int(ec, (utf8_t *)"SELECT " EDB_XMLDB_COLUMN_EID
		" FROM " EDB_XMLDB_TABLE
		" ORDER BY " EDB_XMLDB_COLUMN_EID " DESC"
		" LIMIT 1 OFFSET 0");
#endif
}

eoid_t
edb_xmldb_eoid_max(EdbContext ec)
{

#if 0
	return edb_pgsql_query_int(ec, (utf8_t *)"SELECT MAX(" EDB_XMLDB_COLUMN_EOID ") FROM " EDB_XMLDB_TABLE);
#else
	return edb_pgsql_query_int(ec, (utf8_t *)"SELECT " EDB_XMLDB_COLUMN_EOID
		" FROM " EDB_XMLDB_TABLE
		" ORDER BY " EDB_XMLDB_COLUMN_EOID " DESC"
		" LIMIT 1 OFFSET 0");
#endif
}

eoid_t
edb_xmldb_eoid_of_tuple_max(EdbContext ec)
{

	return edb_pgsql_query_int(ec, (utf8_t *)"SELECT " EDB_XMLDB_COLUMN_EOID
		" FROM " EDB_XMLDB_TABLE
		" WHERE " EDB_XMLDB_COLUMN_TEID "!=0"
		" ORDER BY " EDB_XMLDB_COLUMN_EOID " DESC"
		" LIMIT 1 OFFSET 0");
}

EdbXmldbInfo
edb_xmldb_new(EdbContext ec)
{
	EdbXmldbInfo	xi;

	if(_edb_delayed_free_xmldb) {
		++edb_delayed_free_recycle_count;
		EdbDelayedFreePop(_edb_delayed_free_xmldb, xi);
	} else {
		xi = Edb_malloc(sizeof(struct EDB_XMLDB_INFO));
	}
	if(xi==NULL) return NULL;
	bzero(xi, sizeof(struct EDB_XMLDB_INFO));
	return xi;
}

void
edb_xmldb_free(EdbContext ec, EdbXmldbInfo xi)
{

	if(xi==NULL) return;
	FREE(XmldbXML(xi));
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_xmldb, xi);
	} else {
		Edb_free(xi);
	}
}

EdbXmldbInfo
edb_xmldb_dup(EdbContext ec, EdbXmldbInfo xi)
{
	EdbXmldbInfo	new;

	new = edb_xmldb_new(ec);
	if(new==NULL) return NULL;
	bcopy(xi, new, sizeof(struct EDB_XMLDB_INFO));
	XmldbXML(new) = (utf8_t *)strdup_null((char *)XmldbXML(new));
	return new;
}

int
edb_xmldb_compare(EdbContext ec, EdbXmldbInfo xi1, EdbXmldbInfo xi2)
{
	register int	st;

	if(xi1==xi2) return 0;	/* including NULL */
	if(xi1==NULL) return -1;
	if(xi2==NULL) return 1;
	if((st = XmldbEID(xi1) - XmldbEID(xi2))) return st;
	if((st = XmldbEOID(xi1) - XmldbEOID(xi2))) return st;
	if((st = XmldbTEID(xi1) - XmldbTEID(xi2))) return st;
	if(XmldbXML(xi1)==XmldbXML(xi2)) return 0;	/* including NULL */
	if(XmldbXML(xi1)==NULL) return -1;
	if(XmldbXML(xi2)==NULL) return 1;
	return strcmp((char *)XmldbXML(xi1), (char *)XmldbXML(xi2));
}

#define	XMLDB_COLUMNs			\
	EDB_XMLDB_COLUMN_EID ","		\
	EDB_XMLDB_COLUMN_EOID ","	\
	EDB_XMLDB_COLUMN_TEID ","	\
	EDB_XMLDB_COLUMN_XML

#define	XMLDB_COLUMNs_NUM		4	/* number of columns in XMLDB_COLUMNs */

static
void
edb_xmldb_get_lvalues(EdbContext ec, EdbXmldbInfo xi, Link l_values)
{

	XmldbEID(xi) = ATOI(l_values->d);	l_values = l_values->n;
	XmldbEOID(xi) = ATOI(l_values->d);	l_values = l_values->n;
	XmldbTEID(xi) = ATOI(l_values->d);	l_values = l_values->n;
	XmldbXML(xi) = (utf8_t *)(l_values->d ? strdup(l_values->d) : NULL);
}

EdbXmldbInfo
edb_xmldb_get(EdbContext ec, eid_t eid, eoid_t eoid)
{
	utf8_t *query, *select;
	EdbXmldbInfo	xi;
	Link	l, l_res;
	int	nfields;

	xi = edb_xmldb_new(ec);
	if(xi==NULL) return NULL;

#ifdef EDB_READER
	select = (utf8_t *)"SELECT " XMLDB_COLUMNs " FROM " EDB_XMLDB_TABLE " AS t1 WHERE ";
	if(eoid) {
		asprintf((char **)&query, "%s t1." EDB_XMLDB_COLUMN_EOID "=%d", select, eoid);
	} else {
		asprintf((char **)&query, "%s t1." EDB_XMLDB_COLUMN_EID "=%d", select, eid);
	}
#else /* !EDB_READER */
	select = (utf8_t *)"SELECT " XMLDB_COLUMNs " FROM " EDB_XMLDB_TABLE " AS t1 WHERE ";
	if(eoid) {
		asprintf((char **)&query, "%s t1." EDB_XMLDB_COLUMN_EOID "=%d", select, eoid);
	} else {
		utf8_t	*latest;

		latest = edb_create_latest_phrase(ec, (utf8_t *)EDB_XMLDB_TABLE, (utf8_t *)"t1", (utf8_t *)"t2", NULL, eid);
		asprintf((char **)&query, "%s t1." EDB_XMLDB_COLUMN_EID "=%d AND %s", select, eid, latest);
		FREE(latest);
	}
#endif /* EDB_READER */
	l_res = edb_pgsql_query_one_tuple(ec, query, &nfields);
	if(l_res==NULL || nfields!=XMLDB_COLUMNs_NUM) {
#if 0
		edb_trace(ec, "edb_xmldb_get: can't get xmldb (EID: %d, EOID:%d).\n", eid, eoid);
#endif
		edb_xmldb_free(ec, xi);
		xi = NULL;
	} else {
		l = l_res;
		edb_xmldb_get_lvalues(ec, xi, l_res);
		if(!STRisValid(XmldbXML(xi))) {
			edb_trace(ec, "edb_xmldb_get: can't get xml (EID: %d, EOID:%d).\n", eid, eoid);
			edb_xmldb_free(ec, xi);
			xi = NULL;
		}
	}
	EdbLinkFree(l_res);
	FREE(query);

	return xi;
}

Link
edb_xmldb_get_byTEID(EdbContext ec, eid_t teid, int all)
{
	utf8_t	*select, *query;
	int	ntuples, nfields;
	Link	l, l_res, l_xmldb;

	select = (utf8_t *)"SELECT " XMLDB_COLUMNs " FROM " EDB_XMLDB_TABLE " AS t1 WHERE ";
	if(all) {
		asprintf((char **)&query, "%s"
			EDB_XMLDB_COLUMN_TEID "=%d"
			" ORDER BY t1." EDB_XMLDB_COLUMN_EOID,
			select, teid);
	} else {
#ifdef EDB_READER
		asprintf((char **)&query, "%s" 
			EDB_XMLDB_COLUMN_TEID "=%d"
			" ORDER BY t1." EDB_XMLDB_COLUMN_EID,
			select, teid);
#else /* !EDB_READER */
		utf8_t	*latest;

		latest = edb_create_latest_phrase(ec, (utf8_t *)EDB_XMLDB_TABLE, (utf8_t *)"t1", (utf8_t *)"t2", NULL, 0);
		asprintf((char **)&query, "%s" 
			EDB_XMLDB_COLUMN_TEID "=%d"
			" AND %s ORDER BY t1." EDB_XMLDB_COLUMN_EID,
			select, teid, latest);
		FREE(latest);
#endif /* EDB_READER */
	}

	l_res = edb_pgsql_query(ec, query, &ntuples, &nfields);

	l_xmldb = NULL;
	if(l_res && nfields==XMLDB_COLUMNs_NUM) {
		ForLink(l, l_res) {
			EdbXmldbInfo	xi;

			xi = edb_xmldb_new(ec);
			edb_xmldb_get_lvalues(ec, xi, l->d);
			if(!STRisValid(XmldbXML(xi))) {
				edb_trace(ec, "edb_xmldb_get_byTEID: can't get xml (TEID: %d).\n", teid);
				edb_xmldb_free(ec, xi);
				xi = NULL;
			} else {
				LinkAppendTop(link_atom_new(xi, edb_xmldb_free), l_xmldb);
			}
		}
		LinkReverse(l_xmldb);
	}
	EdbLinkFree(l_res);
	FREE(query);
	return l_xmldb;
}

Link
edb_xmldb_get_byEID(EdbContext ec, eid_t eid)
{
	utf8_t	*select, *query;
	int	ntuples, nfields;
	Link	l, l_res, l_xmldb;

	select = (utf8_t *)"SELECT " XMLDB_COLUMNs " FROM " EDB_XMLDB_TABLE " AS t1 WHERE ";
	asprintf((char **)&query, "%s"
		EDB_XMLDB_COLUMN_EID "=%d"
		" ORDER BY " EDB_XMLDB_COLUMN_EOID,
		select, eid);

	l_res = edb_pgsql_query(ec, query, &ntuples, &nfields);

	l_xmldb = NULL;
	if(l_res && nfields==XMLDB_COLUMNs_NUM) {
		ForLink(l, l_res) {
			EdbXmldbInfo	xi;

			xi = edb_xmldb_new(ec);
			edb_xmldb_get_lvalues(ec, xi, l->d);
			if(!STRisValid(XmldbXML(xi))) {
				edb_trace(ec, "edb_xmldb_get_byEID: can't get xml (EID: %d).\n", eid);
				edb_xmldb_free(ec, xi);
				xi = NULL;
			} else {
				LinkAppendTop(link_atom_new(xi, edb_xmldb_free), l_xmldb);
			}
		}
		LinkReverse(l_xmldb);
	}
	EdbLinkFree(l_res);
	FREE(query);
	return l_xmldb;
}

